"""
Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
Copyright (C) 2020 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt

----------------------------------------------------------------------------------------

Usage:
    net (add|del) bridge (bridge|<interface>) (dhcp-snoop|dhcp-snoop6) vlan <number-range-list>
    net (add|del) bridge (bridge|<interface>) (dhcp-snoop|dhcp-snoop6) vlan <number-range-list> trust <interface>
    net add bridge (bridge|<interface>) (dhcp-snoop|dhcp-snoop6) vlan <number-range-list> rate-limit <0-1000>
    net del bridge (bridge|<interface>) (dhcp-snoop|dhcp-snoop6) vlan <number-range-list> rate-limit

    net del bridge (bridge|<interface>) (dhcp-snoop|dhcp-snoop6)

    # Disable all dhcp-snoop functionality
    net del dhcp-snoop all

    net show (dhcp-snoop|dhcp-snoop6) table [json]

Options:
    dhcp-snoop          : DHCP snooping for IPv4
    dhcp-snoop6         : DHCP snooping for IPv6
    trust               : Trusted interfaces for DHCP snooping
    rate-limit          : DHCP packets per second (default: 100)
"""
import logging
import os
import json
from copy import deepcopy
from itertools import groupby

from tabulate import tabulate

from nclu import (
    ifname_expand_glob,
    ifnames_to_glob,
    glob_to_numbers,
    numbers_to_glob,
    WORKING_DIRECTORY,
)
from nclu.plugins import (
    use_exceptions,
    get_config_files_helper,
    get_pending_diffs,
    make_scratchpads,
    persist_configurations,
    remove_scratch_files_and_backups,
    run_commands,
)


log = logging.getLogger("netd")

BIN_SYSTEMCTL = "/bin/systemctl"
DHCP_SNOOP_SERVICE = "dhcpsnoop.service"
SWITCHD_SERVICE = "switchd.service"
DEL_ALL_COMMANDS = [
    (BIN_SYSTEMCTL, "disable", DHCP_SNOOP_SERVICE),
    (BIN_SYSTEMCTL, "reset-failed", DHCP_SNOOP_SERVICE),
    (BIN_SYSTEMCTL, "stop", DHCP_SNOOP_SERVICE),
]


DHCP_SNOOP_CONFIG = "/etc/dhcpsnoop/dhcp_snoop.json"
DHCP_SNOOP_SCRATCHPAD = os.path.join(
    WORKING_DIRECTORY, "dhcp_snoop", "dhcp_snoop.json"
)
DHCP_SNOOP_DEFAULT = """\
{
    "bridge": []
}
"""

SWITCHD_CONFIG = "/etc/cumulus/switchd.d/50dhcp_snoop.conf"
SWITCHD_SCRATCHPAD = os.path.join(
    WORKING_DIRECTORY, "dhcp_snoop", "50dhcp_snoop.conf"
)


SWITCHD_CONF_KEY_DHCP_TO_CPU = "dhcp_snoop.dhcp_to_cpu"


SWITCHD_DEFAULT = """\
{} = FALSE
""".format(SWITCHD_CONF_KEY_DHCP_TO_CPU)


SWITCHD_CONTENTS_ENABLED = """\
{} = TRUE
""".format(SWITCHD_CONF_KEY_DHCP_TO_CPU)


SWITCHD_NODE_DHCP_CPU_ENABLE = "/cumulus/switchd/config/dhcp_snoop/dhcp_to_cpu"


DHCP_SNOOP_V4_TABLE = "/etc/dhcpsnoop/dhcp_table.json"
DHCP_SNOOP_V6_TABLE = "/etc/dhcpsnoop/dhcpv6_table.json"


DEFAULT_RATE_LIMIT = 100


####################################
# EXPECTED NCLU PLUGIN FUNCTIONS
####################################
# Override keyword definitions created by other plugins/wrappers
per_command_helptext = {
    "net del": {
        # Normally "dhcp-snoop" is specifying just the IPv4 stuff. Here, it
        # means everything.
        "dhcp-snoop": "DHCP snooping",
    },
    "net del dhcp-snoop": {
        "all": "All DHCP snooping configuration",
    },
    "net show (dhcp-snoop|dhcp-snoop6)": {
        "table": "DHCP snooping database",
    },
}


def commit_pending(verbose, net_del_all):
    """
    Copy scratchpad files over configuration files, call necessary restart commands, etc.

    Return a 2-tuple of (bool, dict).  The boolean indicates if the commit succeeded.
    The dictionary is as described in nclu/plugins/__init__.py:execute_modular.

    net_del_all - A boolean; if True, the user is committing "net del all"
    """
    pending_dhcp_snoop = get_pending_diffs(
        [(DHCP_SNOOP_CONFIG, DHCP_SNOOP_SCRATCHPAD)],
        False,
    )
    pending_switchd = get_pending_diffs(
        [(SWITCHD_CONFIG, SWITCHD_SCRATCHPAD)],
        False,
    )

    persist_configurations(
        [
            (DHCP_SNOOP_CONFIG, DHCP_SNOOP_SCRATCHPAD),
            (SWITCHD_CONFIG, SWITCHD_SCRATCHPAD),
        ]
    )

    commands = []
    if net_del_all:
        # Turn off dhcp-snooping on `del all`
        commands = DEL_ALL_COMMANDS
    elif pending_dhcp_snoop != "":
        # For everything else, use the normal restart commands.
        commands = get_restart_commands(DHCP_SNOOP_CONFIG)

    if net_del_all:
        # Always turn this off on del all
        write_switchd_node_dhcp_to_cpu(False)
    elif pending_switchd != "":
        # Otherwise, get the pending config from the scratchpad
        with open(SWITCHD_SCRATCHPAD) as f:
            switchd_contents = f.read()
        should_enable = parse_switchd_conf_for_dhcp_to_cpu(switchd_contents)
        write_switchd_node_dhcp_to_cpu(should_enable)

        # If we're disabling the feature, shift it over to del all commands.
        if not should_enable:
            # HACK: override whatever commands we had before
            commands = DEL_ALL_COMMANDS

    success = True
    if commands:
        success, _, messages = run_commands(commands, verbose)

    if success:
        abort_pending()
        return (True, {})
    else:
        return (False, messages)


def show_config_commands(_ifupdown2_wrapper):
    """
    net show configuration commands
    """
    try:
        with open(DHCP_SNOOP_CONFIG) as f:
            config = json.load(f)
    except IOError:  # EARLY RETURN
        # Probably missing the config file, so report no commands.
        return ([], [])

    # Start by crawling the config and coercing the data structure into
    # something easier to generate commands from. This is necessary because we
    # have to sort/group commands that can be expressed with globs. The general
    # structure of the tree is thus:
    #   {
    #       bridge_id: {
    #           ip_version: {
    #               vlan_id: (rate_limit, trusted_interface)
    #           }
    #       }
    #   }
    # All the identifiers in that example correspond with their values in the
    # config file, except for `trusted_interface`, which is a frozenset of
    # strings instead of a list of strings.
    command_tree = {}
    for bridge in config.get("bridge", []):
        bridge_id = bridge.get("bridge_id")
        command_tree[bridge_id] = {}
        for vlan in bridge.get("vlan", []):
            vlan_id = vlan.get("vlan_id")
            is_snooping = vlan.get("snooping")
            # EARLY CONTINUE
            if not is_snooping:
                # No snooping? No parsing.
                continue

            ip_version = vlan.get("ip_version")
            rate_limit = vlan.get("rate_limit")
            trusted_interfaces = vlan.get("trusted_interface", [])

            command_tree[bridge_id].setdefault(ip_version, {})
            # Each vlan entry is a 2-tuple of (rate-limit, trusted_interfaces)
            command_tree[bridge_id][ip_version][vlan_id] = (
                rate_limit,
                frozenset(trusted_interfaces),
            )

    # Key functions for sorting/grouping. To group correctly, the key for
    # grouping needs to be the same as the key for sorting.
    def rate_limit_keyfunc(vlantuple):
        return vlantuple[1][0]

    def trust_keyfunc(vlantuple):
        return vlantuple[1][1]

    # This will be used to determine which items need to be expressed without
    # the extra command tokens at the end.
    default_vlan = (DEFAULT_RATE_LIMIT, frozenset())

    # Walk the tree and create the list of commands.
    commands = []
    for bridge_id, snoops in command_tree.items():
        # NOTE: We don't attempt to glob bridge_ids here. Too bad!
        for ip_version, vlans in snoops.items():
            snoop_token = "dhcp-snoop"
            if ip_version == 6:
                snoop_token = "dhcp-snoop6"
            # Each item here is going to be
            # `(vlan_id, (rate_limit, trusted_interface))`
            vlans_and_args = list(vlans.items())
            ######################
            # Glob base commands
            base_command_args = filter(
                lambda vlantuple: vlantuple[1] == default_vlan,
                vlans_and_args
            )
            if len(base_command_args) > 0:
                vlan_ids = [vlan_id for vlan_id, _ in base_command_args]
                vlan_glob = numbers_to_glob(vlan_ids)
                commands.append(
                    "net add bridge {bridge_id} {snoop_token} vlan {vlan_glob}".format(
                        bridge_id=bridge_id,
                        snoop_token=snoop_token,
                        vlan_glob=vlan_glob,
                    )
                )
            ###################################
            # Group and glob "trust" commands
            trust_command_args = filter(
                lambda vlantuple: trust_keyfunc(vlantuple) != default_vlan[1],
                vlans_and_args,
            )
            grouped_by_trust = groupby(
                sorted(
                    trust_command_args,
                    key=trust_keyfunc
                ),
                key=trust_keyfunc,
            )
            for trusted_interfaces, grouped_vlantuples in grouped_by_trust:
                ifglob = ifnames_to_glob(list(trusted_interfaces))
                vlan_ids = [vlan_id for vlan_id, _ in grouped_vlantuples]
                vlan_glob = numbers_to_glob(vlan_ids)
                commands.append(
                    "net add bridge {bridge_id} {snoop_token} vlan {vlan_glob} trust {ifglob}".format(
                        bridge_id=bridge_id,
                        snoop_token=snoop_token,
                        vlan_glob=vlan_glob,
                        ifglob=ifglob,
                    )
                )
            ########################################
            # Group and glob "rate-limit" commands
            rate_limit_command_args = filter(
                lambda vlantuple: rate_limit_keyfunc(vlantuple) != default_vlan[0],
                vlans_and_args,
            )
            grouped_by_rate_limit = groupby(
                sorted(
                    rate_limit_command_args,
                    key=rate_limit_keyfunc,
                ),
                key=rate_limit_keyfunc,
            )
            for rate_limit, grouped_vlantuples in grouped_by_rate_limit:
                vlan_ids = [vlan_id for vlan_id, _ in grouped_vlantuples]
                vlan_glob = numbers_to_glob(vlan_ids)
                commands.append(
                    "net add bridge {bridge_id} {snoop_token} vlan {vlan_glob} rate-limit {rate_limit}".format(
                        bridge_id=bridge_id,
                        snoop_token=snoop_token,
                        vlan_glob=vlan_glob,
                        rate_limit=rate_limit,
                    )
                )
    return (commands, [])


def show_config_summary(summary):
    """
    net show configuration dhcp-snoop

    Populate the "summary" dictionary with information about the dhcp-snoop configuration.
    """
    # nope
    pass


@use_exceptions()
def execute_command(net_daemon):
    """
    Execute an NCLU nat command. Assumes that the NCLU core is only sending it
    valid nat commands.

    (Called by NCLU framework code.)

    Side effects:
        Writes files to NCLU scratchpad.

    Returns:
        A 2-tuple of (bool, dict), where bool represents whether or not the
        command succeeded, and dict contains output/error messages to display
        to user.
    """

    # Convenience variables
    tokens = net_daemon.nclu_parser.args
    argv_expanded = net_daemon.nclu_parser.argv_expanded
    assert "configuration" not in tokens, "\"net show configuration dhcp-snoop\" does not use this code path."

    # EARLY RETURN
    if "show" in tokens:
        # `net show ...` is a special snowflake, so we'll handle it and bail.
        table_file = (
            DHCP_SNOOP_V4_TABLE
            if "dhcp-snoop" in tokens
            else DHCP_SNOOP_V6_TABLE
        )
        try:
            with open(table_file) as f:
                table_contents = f.read()
        except IOError:
            # Use an empty JSON object when the file isn't found.
            table_contents = "{}"
        output = format_dhcp_snoop_table_json(table_contents, tokens)
        # Make sure we have a single trailing newline.
        return output.rstrip("\n") + "\n"
    elif argv_expanded == ["net", "del", "dhcp-snoop", "all"]:
        # `net del dhcp-snoop all` is an easy snowflake.
        return del_all()

    # HACK: We also point `net del bridge <interface>` here, so we can
    #       handle deleting all dhcp-snoop config related to that bridge. It's
    #       weird, but it lets us test this plugin's handling of those commands
    #       the same way we handle the other commands.
    is_dhcp_snoop_command = (
        "dhcp-snoop" in tokens
        or "dhcp-snoop6" in tokens
    )
    # Create the relevant scratchpads (if necessary) and track which ones were
    # created for this command. If the command creates scratchfiles, but
    # doesn't actually change anything, we'll clean up the scratchpad at the
    # end.
    did_create_dhcp_snoop_scratchpad = False
    did_create_switchd_scratchpad = False
    if not os.path.isfile(DHCP_SNOOP_SCRATCHPAD):
        did_create_dhcp_snoop_scratchpad = True
        make_scratchpads([
            (
                DHCP_SNOOP_CONFIG,
                DHCP_SNOOP_SCRATCHPAD,
                DHCP_SNOOP_DEFAULT,
            ),
        ])
    # HACK: Only mess with switchd scratchpad if it's actually a dhcp-snoop
    #       command.
    if is_dhcp_snoop_command and not os.path.isfile(SWITCHD_SCRATCHPAD):
        did_create_switchd_scratchpad = True
        make_scratchpads([
            (
                SWITCHD_CONFIG,
                SWITCHD_SCRATCHPAD,
                SWITCHD_DEFAULT,
            ),
        ])

    # Extract command inputs from the parser tokens.
    should_del = "del" in tokens
    if "trust" in tokens:
        # "trust" is weird, since there's two "<interface>" tokens (or one
        # <interface> and one "bridge".)

        # Is <interface> a string?
        if tokens["<interface>"] == str(tokens["<interface>"]):
            # Then we've got a "bridge bridge".
            token_bridge = tokens["bridge"]
            token_trust_iface = tokens["<interface>"]
        else:
            # <interface> as a list means the first token is the bridge and the
            # second is the trust interface.
            token_bridge = tokens["<interface>"][0]
            token_trust_iface = tokens["<interface>"][1]
    else:
        # HACK: if <interface> doesn't exist, then we've got a "bridge bridge".
        token_bridge = tokens.get("<interface>") or tokens["bridge"]
        token_trust_iface = None

    snoop_token = tokens.get("dhcp-snoop") or tokens.get("dhcp-snoop6")
    ip_version = {
        "dhcp-snoop": 4,
        "dhcp-snoop6": 6,
        # HACK: this can be `None` if it's a plain
        #       `net del bridge (bridge|<interface)`
        #       (ie, it's not actually a dhcp-snoop command.)
        None: None,
    }.get(snoop_token)

    # NOTE: vlan should only be optional on del commands
    token_vlan = tokens.get("<number-range-list>")
    token_rate_limit = tokens.get("<0-1000>")
    # `net del ... rate-limit`
    if should_del and "rate-limit" in tokens:
        # HACK: Set the rate-limit to default if we're deleting it, so the
        #       mutation function knows what to do.
        token_rate_limit = DEFAULT_RATE_LIMIT

    # Expand the glob tokens.
    unglobbed_bridge_ids = ifname_expand_glob(token_bridge)
    # WEIRD PATTERN ALERT! The vlan_ids and trust_ifaces are optional
    # arguments. When present, each unglobs to a list of tokens. When not
    # present, they have the value `None`, which cannot be unglobbed. In order
    # to loop over `None` the same way we handle the list of unglobbed tokens,
    # we convert that to `[None]`. Weird? Sure. Cleaner than putting each inner
    # loop inside an if/else block? Very much so.
    unglobbed_vlan_ids = (
        # Mutate config once for each vlan covered by glob.
        glob_to_numbers(str(token_vlan))
        if token_vlan is not None
        # Or mutate it once with `None` as the vlan_id
        else [None]
    )
    unglobbed_trust_ifaces = (
        # Mutate config once for each interface covered by glob.
        ifname_expand_glob(token_trust_iface)
        if token_trust_iface is not None
        # Or mutate it once with `None` as the trust_iface
        else [None]
    )

    #########################################
    # Update the dhcp_snoop.json scratchpad
    with open(DHCP_SNOOP_SCRATCHPAD) as f:
        original_scratch_json = json.load(f)
    # Mutate a copy of the scratchpad, so we can see later if we actually made
    # any changes to the config.
    scratch_json = deepcopy(original_scratch_json)

    # Ha ha, nested for loops go brrrrrrrrrrrr...
    # (but seriously, if there's no globbing, this will only run once)
    for bridge_id in unglobbed_bridge_ids:
        for vlan_id in unglobbed_vlan_ids:
            for trust_iface in unglobbed_trust_ifaces:
                mutate_dhcp_snoop_config_json(
                    scratch_json,
                    should_del,
                    bridge_id,
                    vlan_id,
                    ip_version,
                    trust_iface,
                    token_rate_limit,
                )

    if scratch_json != original_scratch_json:
        # Write the updated config to the scratchpad.
        with open(DHCP_SNOOP_SCRATCHPAD, "w") as f:
            json.dump(
                scratch_json,
                f,
                indent=4,
                sort_keys=True,
                # HACK: use "," as separator to prevent trailing whitespace
                separators=(",", ": "),
            )
            # Add the trailing newline to the file.
            f.write("\n")
    elif did_create_dhcp_snoop_scratchpad:
        # If we created a scratchpad but didn't change it, clean it up.
        os.remove(DHCP_SNOOP_SCRATCHPAD)

    # EARLY RETURN
    if not is_dhcp_snoop_command:
        # Don't touch the switchd file for commands that aren't related to
        # dhcp-snooping.
        return

    #########################################
    # Update the switchd conf scratchpad.
    with open(SWITCHD_SCRATCHPAD) as f:
        original_switchd_conf = f.read()

    # Always enable on update.
    switchd_conf = update_switchd_conf_dhcp_to_cpu_value(
        original_switchd_conf, True
    )

    if switchd_conf != original_switchd_conf:
        with open(SWITCHD_SCRATCHPAD, "w") as f:
            f.write(switchd_conf)
    elif did_create_switchd_scratchpad:
        # If we created a scratchpad but didn't change anything, clean it up.
        os.remove(SWITCHD_SCRATCHPAD)


def del_all():
    """
    net del all

    Revert all configuration files associated with this plug-in to their defaults.
    """
    with open(DHCP_SNOOP_SCRATCHPAD, "w") as f:
        f.write(DHCP_SNOOP_DEFAULT)
    with open(SWITCHD_SCRATCHPAD, "w") as f:
        f.write(SWITCHD_DEFAULT)


def get_config_files():
    """
    Return a string with the content of all existing, non-empty configuration files managed with this plug-in.
    """
    return get_config_files_helper(get_managed_files())


def get_pending(use_colors):
    """ Return a string of the pending changes in all managed files. """
    return get_pending_diffs(
        [
            (DHCP_SNOOP_CONFIG, DHCP_SNOOP_SCRATCHPAD),
            (SWITCHD_CONFIG, SWITCHD_SCRATCHPAD),
        ],
        use_colors
    )


def abort_pending():
    remove_scratch_files_and_backups([
        DHCP_SNOOP_SCRATCHPAD,
        SWITCHD_SCRATCHPAD,
    ])


def get_managed_files():
    """ Return a list of managed files. """

    return [DHCP_SNOOP_CONFIG, SWITCHD_CONFIG]


def get_restart_commands(filename=None):
    """
    Return a list of commands that should be called for the given file after a rollback.
    Some plug-ins use this in commit_pending as well.
    """
    if filename == DHCP_SNOOP_CONFIG:
        # Always enable/start the snoop daemon whenever its config file is
        # changed, EVEN IF WE'RE ROLLING BACK TO A STATE WHERE DHCP SNOOP WAS
        # DISABLED. For normal commits, we look at the `SWITCHD_CONFIG` to see
        # whether or not the snoop daemon should be disabled. During rollback,
        # we can only look at one file at a time, so we can't reliably
        # determine if this is a `del all` state. Too bad!
        return [
            (BIN_SYSTEMCTL, "enable", DHCP_SNOOP_SERVICE),
            # Make sure systemd attempts a start, even if the service is in a
            # failed state.
            (BIN_SYSTEMCTL, "reset-failed", DHCP_SNOOP_SERVICE),
            # We don't need to "restart" the dhcpsnoop service, since it uses
            # inotify events to detect changes in the config file. Instead, we
            # do a "start" to make sure the service is running without stopping
            # it if it's already running..
            (BIN_SYSTEMCTL, "start", DHCP_SNOOP_SERVICE),
        ]
    elif filename == SWITCHD_CONFIG:
        # We need to restart switchd completely on restart, since there's no
        # sane/reliable way to write just the FUSE node we need via
        # `check_output` commands. Too bad!
        return [
            (BIN_SYSTEMCTL, "restart", SWITCHD_SERVICE),
        ]
    else:
        return []


###########################
# NON-PLUGIN FUNCTIONS
###########################
DHCP_SNOOP_TABLE_COLUMNS = [
    "Port",
    "VLAN",
    "IP",
    "MAC",
    "Lease",
    "State",
    "Bridge",
]


def format_dhcp_snoop_table_json(contents, tokens):
    """
    Parse the contents of a DHCP snooping table file and format it as an NCLU
    table.

    Args:
        contents: The string contents of the dhcp_snoop table file.
        tokens: The network_docopt dict of parsed tokens.

    Returns:
        The string table output.
    """
    # EARLY RETURN
    if "json" in tokens:
        # Contents are already in JSON. Load it, and dump it for formatting
        # consistency.
        return json.dumps(
            json.loads(contents),
            indent=4,
            sort_keys=True,
            # HACK: use "," as separator to prevent trailing whitespace
            separators=(",", ": "),
        )
    # The top-level key is different depending on whether it's IPv4 or IPv6
    top_key = (
        "dhcp_table"
        if "dhcp-snoop" in tokens
        else "dhcpv6_table"
    )

    dhcp_table_dicts = json.loads(contents).get(top_key, [])

    # Munge the table to make sure we're getting the output in a consistent
    # order.
    dhcp_table = [
        [
            dict_row[key]
            for key in DHCP_SNOOP_TABLE_COLUMNS
        ]
        for dict_row in dhcp_table_dicts
    ]
    return tabulate(
        dhcp_table,
        headers=DHCP_SNOOP_TABLE_COLUMNS,
        disable_numparse=True,  # Don't align number columns differently.
    )


def mutate_dhcp_snoop_config_json(
    scratch_json,
    should_del,
    bridge_id,
    vlan_id,
    ip_version,
    trust_iface,
    rate_limit,
):
    """
    Mutate the dhcp_snoop.json config structure in place with the specified
    arguments. There's little-to-no validation of command inputs here, since
    network_docopt should have already handled that. (If there's a bug in
    network_docopt's input validation, then we certainly don't want to cover
    that up.)

    Args:
        scratch_json: The JSON-compatible dhcp_snoop config to update.
        should_del: True if we're deleting the configuration. Otherwise, we're
            adding the config
        bridge_id: The string bridge_id to edit DHCP snooping on.
        vlan_id: The integer vlan number to update DHCP snooping on. Set to
            `None` to delete all vlans with specified `ip_version` on bridge.
            (NOTE: `vlan_id` is required for all "add" commands.)
        ip_version: Integer representing whether we're editing the IPv4
            configuration or the IPv6 configuration for the vlan/bridge. Set to
            `None` to delete both IP versions.
        trust_iface: The string interface to add/del as trusted on the
            vlan/bridge. Set to `None` if updating "trust" is not desired.
        rate_limit: The integer rate limit to set for the snooping
            configuration. Set to `None` if updating "rate-limit" is not
            desired.

    Side effects:
        Mutates `scratch_json` argument in place.
    """
    bridges = scratch_json.setdefault("bridge", [])

    # Find the specified bridge, and note its index (in case we need to delete
    # the bridge later.)
    bridge_idx, bridge_config = next(
        (
            (idx, bridge)
            for idx, bridge in enumerate(bridges)
            if bridge.get("bridge_id") == bridge_id
        ),
        (-1, None)
    )
    if bridge_config is None:
        if should_del:  # EARLY RETURN
            # Can't delete something that's not there. Our work here is done.
            return
        # Create the bridge config and add it to the config.
        bridge_config = {
            "bridge_id": bridge_id,
            "vlan": [],
        }
        bridges.append(bridge_config)

    # HACK: "vlan" should always be present, but it's a user-editable file, so
    #       we have to handle shenanigans to some extent.
    vlans = bridge_config.setdefault("vlan", [])

    # EARLY RETURN
    if should_del and vlan_id is None:
        # SNOWFLAKE: `net del bridge <interface> [dhcp-snoop|dhcp-snoop6]`
        # Delete all vlans for specified ip version on specified bridge.
        if bridge_config is not None:
            vlans = [
                vlan
                for vlan in vlans
                if vlan.get("ip_version") != ip_version
            ]
            bridge_config["vlan"] = vlans
            # Delete the bridge if no vlans are left, or if we're supposed
            # delete both IP versions.
            if len(vlans) == 0 or ip_version is None:
                del bridges[bridge_idx]
        return

    # Grab the index too, so we can delete it later.
    vlan_idx, vlan_config = next(
        (
            (idx, vlan)
            for idx, vlan in list(enumerate(vlans))
            if (
                vlan.get("vlan_id") == vlan_id
                and vlan.get("ip_version") == ip_version
            )
        ),
        (-1, None)
    )
    if vlan_config is None:
        if should_del:  # EARLY RETURN
            # Nothing to do here if we're deleting
            return
        # Create/add the config if it doesn't exist yet.
        vlan_config = {
            "vlan_id": vlan_id,
            "snooping": 1,
            "rate_limit": DEFAULT_RATE_LIMIT,
            "ip_version": ip_version,
            "trusted_interface": [],
        }
        vlans.append(vlan_config)
    elif (
        # EARLY RETURN
        should_del
        # No trust_iface was provided
        and trust_iface is None
        # and we're not messing with rate-limit.
        and rate_limit != DEFAULT_RATE_LIMIT
    ):
        # We're deleting the vlan's dhcp-snoop configuration from the bridge.
        del vlans[vlan_idx]
        return

    if not should_del:
        # Make sure the snooping config is actually enabled.
        vlan_config["snooping"] = 1

    # `net (add|del) ... trust <interface>`
    if trust_iface is not None:
        # HACK: Again, this should always be present, but we don't want to
        #       crash if the user gets the config into a weird state.
        trusted_interfaces_config = vlan_config.setdefault("trusted_interface", [])
        if should_del:
            try:
                trusted_interfaces_config.remove(trust_iface)
            except ValueError:
                # Eh, it's fine if the interface didn't exist
                pass
        elif trust_iface not in trusted_interfaces_config:
            trusted_interfaces_config.append(trust_iface)

    # `net (add|del) ... rate-limit <0-1000>`
    if rate_limit is not None:
        vlan_config["rate_limit"] = rate_limit


def parse_switchd_conf_for_dhcp_to_cpu(contents):
    """
    Parse the switchd conf file and return the value of
    `dhcp_snoop.dhcp_to_cpu`.

    Args:
        contents: The string contents of the switchd conf file.

    Returns:
        A boolean representing whether or not the flag is enabled.
    """
    found_line = next((
        line.strip()
        for line in contents.splitlines()
        if line.strip().startswith(SWITCHD_CONF_KEY_DHCP_TO_CPU)
    ), None)
    if found_line is None:
        return False

    # Remove the key
    line = found_line[len(SWITCHD_CONF_KEY_DHCP_TO_CPU):]
    # HACK: Assume there's an equal sign
    key, _, value_and_comment = map(
        lambda s: s.strip(),
        line.partition("="),
    )
    value, _, _ = map(
        lambda s: s.strip(),
        value_and_comment.partition("#"),
    )
    return value == "TRUE"


def update_switchd_conf_dhcp_to_cpu_value(contents, value):
    """
    Update the contents of the switchd conf with a new value for
    `dhcp_snoop.dhcp_to_cpu`. Makes a reasonable effort to preserve whitespace
    and comments.

    Args:
        contents: The string contents of the switchd conf file.
        value: The new boolean value to update the contents with.

    Returns:
        The updates string contents of the switchd conf file.
    """
    lines = contents.splitlines()
    idx, found_line = next((
        (idx, line)
        for idx, line in enumerate(contents.splitlines())
        if line.strip().startswith(SWITCHD_CONF_KEY_DHCP_TO_CPU)
    ), (-1, None))

    value_as_string = str(value).upper()

    # EARLY RETURN
    if found_line is None:
        # Add the config line to the end of the file.
        lines.append(
            "{} = {}".format(
                SWITCHD_CONF_KEY_DHCP_TO_CPU,
                value_as_string,
            )
        )
        return "\n".join(lines) + "\n"

    # HACK: Assume there's an equal sign
    key, _, value_and_comment = found_line.partition("=")
    value, sep, comment = value_and_comment.partition("#")

    value_whitespace_right = value[len(value.rstrip()):]
    value_whitespace_left = value[:len(value) - len(value.lstrip())]

    new_line = "{}={}".format(
        key,
        (
            value_whitespace_left
            + value_as_string
            + value_whitespace_right
            + sep
            + comment
        ),
    )
    # Replace the found config line with the updated contents.
    lines[idx] = new_line

    return "\n".join(lines) + "\n"


def write_switchd_node_dhcp_to_cpu(enable):
    """
    Write a value to the dhcp_to_cpu switchd FUSE node, if it exists. If
    switchd isn't running (as is the case on VX) DHCP snooping will work
    without any special switchd shenanigans. The dhcp_to_cpu flag is needed on
    switchd-enabled systems to make DHCP traffic visible to the CPU (instead of
    handling it in the ASIC.)

    Args:
        enable: A boolean value representing whether we're enabling/disabling
            the feature.

    Side effects:
        Writes to a switchd FUSE node, if it exists.
    """
    # EARLY RETURN
    if not os.path.exists(SWITCHD_NODE_DHCP_CPU_ENABLE):
        # We don't want to create the file if it doesn't exist.
        return
    contents = ("TRUE" if enable else "FALSE") + "\n"
    with open(SWITCHD_NODE_DHCP_CPU_ENABLE, "w") as f:
        f.write(contents)
