from collections import OrderedDict
from nclu import (
    files_match,
    get_pending_add_del_commands,
    WORKING_DIRECTORY,
    CONFIG,
)
from nclu import service_is_active
from nclu.utils import check_output
from nclu.plugins import merge_messages
from subprocess import STDOUT, CalledProcessError
import os
import re
import shutil
import logging


log = logging.getLogger("netd")


USR_LIB_FRR_PIMD = "/usr/lib/frr/pimd"
FRR_RELOAD_PY = "/usr/lib/frr/frr-reload.py"
show_linux_command = CONFIG.show_linux_command

if os.path.exists("/run/frr/.qcompat"):
    SERVICE_NAME = "quagga.service"
    OWNER_NAME = 'quagga'
    SCRATCHPAD_DIRECTORY = os.path.join(WORKING_DIRECTORY, "quagga")
    ETC_FRR = "/etc/quagga/"
    ETC_FRR_DAEMONS = "/etc/quagga/daemons"
    ETC_FRR_CONF = "/etc/quagga/Quagga.conf"

else:
    SERVICE_NAME = "frr.service"
    OWNER_NAME = 'frr'
    SCRATCHPAD_DIRECTORY = os.path.join(WORKING_DIRECTORY, "frr")
    ETC_FRR = "/etc/frr/"
    ETC_FRR_DAEMONS = "/etc/frr/daemons"
    ETC_FRR_CONF = "/etc/frr/frr.conf"

ETC_FRR_DAEMONS_SCRATCHPAD = os.path.join(SCRATCHPAD_DIRECTORY, "daemons.scratchpad")
ETC_FRR_DAEMONS_BACKUP = ETC_FRR_DAEMONS_SCRATCHPAD + ".bak"
ETC_FRR_DAEMONS_BASELINE = ETC_FRR_DAEMONS_SCRATCHPAD + ".baseline"

ETC_FRR_CONF_SCRATCHPAD = SCRATCHPAD_DIRECTORY + "/frr.conf.scratchpad"
ETC_FRR_CONF_BACKUP = ETC_FRR_CONF_SCRATCHPAD + ".bak"
ETC_FRR_CONF_BASELINE = ETC_FRR_CONF_SCRATCHPAD + ".baseline"
ETC_FRR_CONF_DEFAULT = ETC_FRR_CONF_SCRATCHPAD + ".default"


def file_baselines_match():
    messages = {}
    # Verify that FRR's daemon and configuration files have not been modified.
    if not files_match(ETC_FRR_DAEMONS, ETC_FRR_DAEMONS_BASELINE):
        if not os.path.exists(ETC_FRR_DAEMONS_BASELINE):
            log.error(("{0} does not exist and therefore can't be compared to"
                       "{1}").format(ETC_FRR_DAEMONS_BASELINE, ETC_FRR_DAEMONS))
        elif not os.path.exists(ETC_FRR_DAEMONS):
            log.error(("{0} does not exist and therefore can't be compared to"
                       "{1}").format(ETC_FRR_DAEMONS, ETC_FRR_DAEMONS_BASELINE))
        messages["error"] = ["{0} was modified by another user.\n{1}".format(ETC_FRR_DAEMONS, get_pending_add_del_commands())]
        return (False, messages)

    if (os.path.exists(ETC_FRR_CONF) and os.path.exists(ETC_FRR_CONF_BASELINE) and
        os.path.getmtime(ETC_FRR_CONF) > os.path.getmtime(ETC_FRR_CONF_BASELINE)):
        messages["error"] = ["{0} was modified by another user.\n{1}".format(ETC_FRR_CONF, get_pending_add_del_commands())]
        return (False, messages)

    return (True, messages)


def frr_wr_mem():
    """
    Issue a "wr mem" to get ETC_FRR_CONF in a nice, clean format.
    """

    try:
        check_output(["/usr/bin/vtysh", "-c", "wr mem"], stderr=STDOUT)
        return (True, {})

    except CalledProcessError as e:
        log.exception(e)
        messages = {"error": ["\"wr mem\" failed due to \"{0}\".".format(e.output)]}
        log.error(messages["error"][0])
        return (False, messages)


def init_guts_add_del():
    """
    - create BASELINE files if needed
    - create SCRATCHPAD files if needed
    """

    messages = {}
    # We use BASELINE files later to determine if the original config file was modified by another user.
    if not os.path.isfile(ETC_FRR_CONF_BASELINE):
        result, wr_mem_messages = frr_wr_mem()

        if not result:
            assert "error" in wr_mem_messages, "Why did frr_wr_mem fail?"
            return (False, wr_mem_messages)

        assert "error" not in wr_mem_messages, "frr_wr_mem returned True.  There shouldn't be any error messages."
        merge_messages(messages, wr_mem_messages)

        # To create ETC_FRR_CONF_BASELINE we need to "mark" the config file with "end"
        # strings to ID the various contexts and sub-contexts.  This is done via "vtysh -m".
        try:
            output = check_output(["/usr/bin/vtysh", "-m", "-f", ETC_FRR_CONF], stderr=STDOUT).splitlines()
        except CalledProcessError as e:
            # vtysh will fail to mark the file if it contains an invalid
            # configuration. Since we just issued a "wr mem", if ETC_FRR_CONF
            # contains an invalid config it is due to a bug in FRR.
            msg = "'vtysh -m -f {0}' failed due to '{1}'.\nThe contents of {0} are:\n".format(ETC_FRR_CONF, e.output)
            with open(ETC_FRR_CONF) as f:
                msg += f.read()
            log.error(msg)

            messages["error"] = ["Could not mark {0} due to '{1}'.\nSee /var/log/netd.log for details.".format(ETC_FRR_CONF, e.output)]
            return (False, messages)

        with open(ETC_FRR_CONF_BASELINE, 'w') as f:
            for line in output:
                if line != '!':
                    f.write(line + '\n')

        shutil.copyfile(ETC_FRR_CONF_BASELINE, ETC_FRR_CONF_SCRATCHPAD)

    if not os.path.isfile(ETC_FRR_DAEMONS_BASELINE):
        shutil.copyfile(ETC_FRR_DAEMONS, ETC_FRR_DAEMONS_BASELINE)
        shutil.copyfile(ETC_FRR_DAEMONS, ETC_FRR_DAEMONS_SCRATCHPAD)

    result, files_modified_messages = file_baselines_match()
    merge_messages(messages, files_modified_messages)

    return (result, messages)


def init_guts():
    """
    - handle the transition from quagga to frr and vice versa
    - create the scratchpad directory
    - start frr if needed
    """

    messages = {}

    if os.path.exists("/run/frr/.qcompat"):
        if SERVICE_NAME == "frr":
            messages["error"] = ["frr was migrated to quagga, please run 'sudo systemctl restart netd'"]
            return (False, messages)
    else:
        if SERVICE_NAME == "quagga":
            messages["error"] = ["quagga was migrated to frr, please run 'sudo systemctl restart netd'"]
            return (False, messages)

    # Create the scratchpad directory
    if not os.path.isdir(SCRATCHPAD_DIRECTORY):
        os.makedirs(SCRATCHPAD_DIRECTORY, mode=WORKING_DIRECTORIES_PERMISSIONS)

    if not os.path.isdir(ETC_FRR):
        # TODO - Why not just create the directory?
        messages["error"] = ["The FRR plug-in could not initialize. {0} is missing.".format(ETC_FRR)]
        log.error(messages["error"][0])
        return (False, messages)

    # /etc/frr/daemons does not exist or is empty.  Create a default file.
    if not os.path.exists(ETC_FRR_DAEMONS) or os.path.getsize(ETC_FRR_DAEMONS) == 0:
        with open(ETC_FRR_DAEMONS, 'w') as f:
            f.write("zebra=yes\n")

    # /etc/frr/frr.conf does not exist or is empty.  Create a default file.
    if not os.path.isfile(ETC_FRR_CONF) or os.path.getsize(ETC_FRR_CONF) == 0:
        with open(ETC_FRR_CONF, 'w') as f:
            f.write("service integrated-vtysh-config\n")

    result = True

    if not service_is_active(SERVICE_NAME):
        daemon_list, daemon_index = get_daemons()

        # If zebra is not enabled, update the daemons file to enable zebra.
        zebra_index = daemon_index.get("zebra")
        if zebra_index:
            if daemon_list[zebra_index][2] != "yes":
                daemon_list[zebra_index][2] = "yes"

            with open(ETC_FRR_DAEMONS, 'w') as f:
                for line, key, val in daemon_list:
                    if key != '':
                        # we have a daemon or daemon option setting
                        f.write("{0}={1}\n".format(key, val))
                    else:
                        # just preserve the line
                        f.write(line)

        result, restart_messages = frr_restart(False)
        merge_messages(messages, restart_messages)

        if result:
            result, frr_wr_mem_messages = frr_wr_mem()
            merge_messages(messages, frr_wr_mem_messages)

    return (result, messages)


def get_daemons():
    """
    Return list of lists from ETC_FRR_DAEMONS_SCRATCHPAD or ETC_FRR_DAEMONS
    For comments, we add [line, '', ''], for key value lines we add
    [line, key, value].  We also return an daemon_index dict where the
    daemon name is the key and the value is the index into our list.
    """
    daemon_list = []
    daemon_index = {}
    if os.path.isfile(ETC_FRR_DAEMONS_SCRATCHPAD):
        filename = ETC_FRR_DAEMONS_SCRATCHPAD
    elif os.path.isfile(ETC_FRR_DAEMONS):
        filename = ETC_FRR_DAEMONS
    else:
        filename = None
    if filename is not None:
        with open(filename) as f:
            counter = 0
            for line in f:
                if re.findall(r"^\s*#.*", line) or line == '\n':
                    # a leading # indicates a comment, just save the line
                    # also save the line if it is just a newline
                    daemon_list.append([line, '', ''])
                elif '=' in line:
                    # we have a key = value so we
                    # create the tuple: for the key, we remove leading and
                    # traiiling white space  (original line, key, value)
                    key, val = re.findall(r"^\s*(\w+)\s*=(.*)", line.strip())[0]
                    daemon_list.append([line, key, val])
                    # we will use this later to change values
                    daemon_index[key] = counter
                else:
                    # We could just have whitespace or something without an
                    # equal sign so just save it
                    daemon_list.append([line, '', ''])
                counter += 1
    return (daemon_list, daemon_index)


ROUTER_ID_RE = r"""^(\+|\-)\s+ospf6? router-id (.*)"""

def is_ospf_router_id_changed(the_diff):
    """
    FRR needs to restart if an ospf router-id has changed. Detect
    the situations where that happened.
    """

    matches = dict()

    # find all occurrances of osfp router-id changes in the diff
    match = re.findall(ROUTER_ID_RE, the_diff, re.MULTILINE)

    # walk through all the router-id changes and tally the number
    # of '+' (adds) and '-' (del) of each router-id.
    for item in match:
        if item[0] == '+':
            matches[item[1]] = matches.get(item[1],0) + 1
        else:
            matches[item[1]] = matches.get(item[1],0) - 1

    # This seems odd but there is a case where a (many) add and del
    # of the router-id is in the same pending buffer essentially
    # leaving the router-id configuration unchanged.  We don't want
    # to do a restart on frr in that case as the change is benign.
    #
    # so this detects if any of the router-ids have non-zero match
    # values which means it's an add or delete and that should
    # cause frr to restart to pick up that change.
    return not all(x == 0 for x in matches.values())


def frr_restart_needed():
    """
    Return True if any of the following are True.  Otherwise, return False.
    - The "daemons" file was modified
    - frr is not running
    - The ospf/ospf6 router-id was modified
    """

    if os.path.isfile(ETC_FRR_DAEMONS_SCRATCHPAD) and not files_match(ETC_FRR_DAEMONS, ETC_FRR_DAEMONS_SCRATCHPAD):
        # The "daemons" file was modified.
        log.warning("FRR Restarting because daemons file changed")
        return True

    if not service_is_active(SERVICE_NAME):
        # frr is not running.
        log.warning("FRR Restarting because FRR is not running")
        return True

    if os.path.isfile(ETC_FRR_CONF) and os.path.isfile(ETC_FRR_CONF_SCRATCHPAD):
        try:
            check_output(["/usr/bin/diff", "-U", '3', ETC_FRR_CONF, ETC_FRR_CONF_SCRATCHPAD])
        except CalledProcessError as e:
            # There is a delta.
            if is_ospf_router_id_changed(e.output):
                log.warning("FRR Restarting because ospf router-ids changed")
                return True

    return False


def frr_restart(verbose):
    """
    Restart the FRR service.  Return a 2-tuple of (bool, str) indicating if the restart
    succeeded and any messages for the user.
    """

    assert isinstance(verbose, bool)
    messages = {}

    try:
        output = check_output(["/bin/systemctl", "enable", SERVICE_NAME])
        if verbose:
            messages["output"] = output

        command = ["/bin/systemctl", "reset-failed", SERVICE_NAME]
        try:
            output = check_output(command)
        except CalledProcessError as e:
            # If a service is already stopped and disabled, "systemctl reset-failed" will return non-zero.
            log.warning("\"{0}\" failed.  Is the service disabled?".format(' '.join(command)))
            output = e.output
        if verbose:
            messages["output"] += '\n' + output

        output = check_output(["/bin/systemctl", "restart", SERVICE_NAME])
        if verbose:
            messages["output"] += '\n' + output
            if not messages["output"].strip():
                # Don't output blank lines.
                del messages["output"]

    except CalledProcessError as e:
        log.exception(e)
        messages["error"] = ["Failed to restart {0} due to \"{1}\".".format(SERVICE_NAME, e.output)]
        log.error(messages["error"][0])
        return (False, messages)

    return (True, messages)


def frr_reload_exception_to_string(filename, e):
    if hasattr(e, 'output'):
        return "Failed to load {0} due to {1}\n\n{2} output\n=======================\n{3}\n".format(filename, e, FRR_RELOAD_PY, e.output)
    return "Failed to load {0} due to {1}".format(filename, e)
