"""
Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
Copyright (C) 2020 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt

----------------------------------------------------------------------------------------

Usage:
    ##############
    # Static NAT
    ##############
    # ICMP doesn't have ports
    net (add|del) nat static snat icmp <ipv4-original> [out-interface <interface>] translate <ipv4-translated>
    net (add|del) nat static dnat icmp <ipv4-original> [in-interface <interface>] translate <ipv4-translated>

    net (add|del) nat static snat (udp|tcp) <ipv4-original> [<number-port-original>] [out-interface <interface>] translate <ipv4-translated> [<number-port-translated>]
    net (add|del) nat static dnat (udp|tcp) <ipv4-original> [<number-port-original>] [in-interface <interface>] translate <ipv4-translated> [<number-port-translated>]

    ##############
    # Dynamic NAT
    ##############
    # ICMP doesn't have ports
    net (add|del) nat dynamic snat icmp source-ip <ipv4/prefixlen> [destination-ip <ipv4/prefixlen>] [out-interface <interface>] translate <text-ipv4-range> [<number-range>]
    net (add|del) nat dynamic dnat icmp [source-ip <ipv4/prefixlen>] destination-ip <ipv4/prefixlen> [in-interface <interface>] translate <text-ipv4-range> [<number-range>]
    # If the pool IP is a single adress needs to have a port range attached
    net (add|del) nat dynamic snat icmp source-ip <ipv4/prefixlen> [destination-ip <ipv4/prefixlen>] [out-interface <interface>] translate <ipv4> <number-range>
    net (add|del) nat dynamic dnat icmp [source-ip <ipv4/prefixlen>] destination-ip <ipv4/prefixlen> [in-interface <interface>] translate <ipv4> <number-range>

    # Same as ICMP commands, but with (udp|tcp) and source/destination ports
    net (add|del) nat dynamic snat (udp|tcp) source-ip <ipv4/prefixlen> [source-port <number-port-source>] [destination-ip <ipv4/prefixlen>] [destination-port <number-port-destination>] [out-interface <interface>] translate <text-ipv4-range> [<number-range>]
    net (add|del) nat dynamic dnat (udp|tcp) [source-ip <ipv4/prefixlen>] [source-port <number-port-source>] destination-ip <ipv4/prefixlen> [destination-port <number-port-destination>] [in-interface <interface>] translate <text-ipv4-range> [<number-range>]
    # If the pool IP is a single add then it needs to have a port range attached
    net (add|del) nat dynamic snat (udp|tcp) source-ip <ipv4/prefixlen> [source-port <number-port-source>] [destination-ip <ipv4/prefixlen>] [destination-port <number-port-destination>] [out-interface <interface>] translate <ipv4> <number-range>
    net (add|del) nat dynamic dnat (udp|tcp) [source-ip <ipv4/prefixlen>] [source-port <number-port-source>] destination-ip <ipv4/prefixlen> [destination-port <number-port-destination>] [in-interface <interface>] translate <ipv4> <number-range>


    ##############
    # del all
    ##############
    net del nat [static|dynamic] all
Options:
    nat                         : Network Address Translation
    static                      : Static NAT configuration
    snat                        : Source NAT
    dnat                        : Destination NAT
    source-ip                   : Source IP address
    destination-ip              : Destination IP address
    source-port                 : Source port number
    destination-port            : Destination port number
    <ipv4-original>             : IPv4 address to be matched by NAT rule
    <ipv4-translated>           : IPv4 address to give matched packets
    <number-port-original>      : Port number (0-65535) to be matched by NAT rule
    <number-port-translated>    : Port number (0-65535) to give matched packets
    <number-port-source>        : Port number (0-65535) to be matched by NAT rule
    <number-port-destination>   : Port number (0-65535) to be matched by NAT rule
    <text-ipv4-range>           : A range of IPv4 addresses to use for the NAT pool (example: 192.0.2.1-192.0.2.100)
    translate                   : Translation of matched packets
    in-interface                : Incoming interface to match
    out-interface               : Outgoing interface to match
    icmp                        : Internet Control Message Protocol
    tcp                         : Transmission Control Protocol
    udp                         : User Datagram Protocol
"""

import logging
import os
from collections import namedtuple
from network_docopt.match import match_ipv4_address

from nclu import WORKING_DIRECTORY
from nclu.plugins import (
    get_config_files_helper,
    get_pending_diffs,
    make_scratchpads,
    persist_configurations,
    remove_scratch_files_and_backups,
    run_commands,
    use_exceptions,
)

log = logging.getLogger("netd")

BIN_CL_ACLTOOL = "/usr/cumulus/bin/cl-acltool"

ACL_RULE_FILE = "/etc/cumulus/acl/policy.d/50nat.nclu.rules"

ACL_RULE_SCRATCHFILE = os.path.join(
    WORKING_DIRECTORY, "nat",
    ACL_RULE_FILE.lstrip("/")  # Strip leading slash so the join works
)


DEFAULT_ACL_CONTENT = """\
[iptables]
"""


class InvalidPoolError(Exception):
    """
    Raised when the user-provided NAT pool is not valid.
    """
    pass


class RuleParseError(Exception):
    """
    Raised when config_line_to_nclu_command fails to parse something.
    """
    pass


class CannotDelError(Exception):
    """
    Raised when we can't execute a `del` command due to the specified
    configuration not being present.
    """
    pass


####################################
# EXPECTED NCLU PLUGIN FUNCTIONS
####################################
# Override keyword definitions created by other plugins/wrappers
per_command_helptext = {
    "net (add|del) nat": {
        "static": "Static NAT configuration",
        "dynamic": "Dynamic NAT configuration",
    },
}


def commit_pending(verbose, net_del_all):
    """
    Copy scratchpad files over configuration files, call necessary restart commands, etc.

    Return a 2-tuple of (bool, dict).  The boolean indicates if the commit succeeded.
    The dictionary is as described in nclu/plugins/__init__.py:execute_modular.

    net_del_all - A boolean; if True, the user is committing "net del all"
    """

    pending = get_pending(use_colors=False)
    if pending == "":
        # Scratch may exist; clean it up.
        abort_pending()
        # Nothing else to do.
        return (True, {})

    persist_configurations([(ACL_RULE_FILE, ACL_RULE_SCRATCHFILE)])

    restart_commands = get_restart_commands(ACL_RULE_FILE)
    outcome, _, messages = run_commands(restart_commands, verbose)

    abort_pending()

    if outcome:
        return (True, {})
    else:
        return (False, messages)


def show_config_commands(ifupdown2_wrapper):
    """
    net show configuration commands
    """
    try:
        with open(ACL_RULE_FILE) as f:
            contents = f.read()
    except IOError:
        # It's not a problem if the file isn't there. Quietly eat the error.
        return ([], [])

    by_line = get_nclu_commands_by_line(contents)
    recognized = [entry[1] for entry in by_line]

    return (recognized, [])


def show_config_summary(summary):
    """
    net show configuration nat

    Populate the "summary" dictionary with information about the nat configuration.
    """
    # nope
    pass


@use_exceptions(Exception)
def execute_command(net_daemon):
    """
    Execute an NCLU nat command. Assumes that the NCLU core is only sending it
    valid nat commands.

    (Called by NCLU framework code.)

    Side effects:
        Writes files to NCLU scratchpad.

    Returns:
        A 2-tuple of (bool, dict), where bool represents whether or not the
        command succeeded, and dict contains output/error messages to display
        to user.
    """
    # Convenience variables
    tokens = net_daemon.nclu_parser.args
    argv_expanded = net_daemon.nclu_parser.argv_expanded
    assert "configuration" not in tokens, "\"net show configuration nat\" does not use this code path."
    assert "nat" in tokens, "Why are you here?"

    did_create_scratchfile = False
    if not os.path.isfile(ACL_RULE_SCRATCHFILE):
        did_create_scratchfile = True
        make_scratchpads([
            (ACL_RULE_FILE, ACL_RULE_SCRATCHFILE, DEFAULT_ACL_CONTENT)
        ])

    if argv_expanded == ["net", "del", "nat", "all"]:
        # This one's easy, so we'll handle it first. It's basically
        # `net del all` as far as this plugin is concerned, so we'll handle
        # it in exactly the same way.
        del_all()
    elif "add" in tokens:
        # Create a rule from the parsed add command
        if "static" in tokens:
            command = parse_to_static_nat_rule(net_daemon.nclu_parser)
            rule = static_nat_rule_to_config_line(command)
        elif "dynamic" in tokens:
            command = parse_to_dynamic_nat_rule(net_daemon.nclu_parser)
            rule = dynamic_nat_rule_to_config_line(command)

        # Update the config file with that rule
        with open(ACL_RULE_SCRATCHFILE) as f:
            current_scratch = f.read()
        # Only add it if it's not already there
        if rule.strip("\n") not in current_scratch.splitlines():
            updated_scratch = add_to_config_file(
                current_scratch,
                "iptables",
                rule,
            )

            with open(ACL_RULE_SCRATCHFILE, "w") as f:
                f.write(updated_scratch)
    elif "del" in tokens:
        with open(ACL_RULE_SCRATCHFILE) as f:
            pending_scratch = f.read()

        # Get a list of "add" commands we're going to delete from the pending
        # scratch file.
        if "all" in tokens:
            # "net del nat (static|dynamic) all"
            # Grab all the recognized nat commands
            by_line = get_nclu_commands_by_line(pending_scratch)
            all_commands = [entry[1] for entry in by_line]

            # We already handled "net del nat all", so the token we're
            # filtering on has to be one of these.
            delete_token = tokens.get("static") or tokens.get("dynamic")
            # Filter it down to the nat commands that contain the delete token.
            commands_to_delete = [
                cmd
                for cmd in all_commands
                if delete_token in cmd.split(" ")
            ]
        else:
            # It's a specific "del" command with a single "add" counterpart.
            # HACK: Put it in a list so we don't have to differentiate between
            #       the "del all" commands when updating the file.
            commands_to_delete = [
                # HACK: Convert the command to its "add" form, since that's how
                #       it's recognized by `del_from_config_file`.
                " ".join(["net", "add"] + argv_expanded[2:])
            ]

        try:
            # Remove each command from the pending config.
            for cmd in commands_to_delete:
                pending_scratch = del_from_config_file(pending_scratch, cmd)
        except CannotDelError as err:
            # We tried to delete something we couldn't. Let's clean up, tell
            # the user, and pretend like this never happened.
            # Remove the scratchfile if it was created for this command.
            if did_create_scratchfile:
                os.remove(ACL_RULE_SCRATCHFILE)
            # Bubble up the error to the user.
            raise err

        with open(ACL_RULE_SCRATCHFILE, "w") as f:
            f.write(pending_scratch)


def del_all():
    """
    net del all

    Revert all configuration files associated with this plug-in to their defaults.
    """
    if not os.path.isfile(ACL_RULE_SCRATCHFILE):
        make_scratchpads([
            (ACL_RULE_FILE, ACL_RULE_SCRATCHFILE, DEFAULT_ACL_CONTENT)
        ])
    with open(ACL_RULE_SCRATCHFILE, "w") as f:
        f.write(DEFAULT_ACL_CONTENT)


def get_config_files():
    """
    Return a string with the content of all existing, non-empty configuration files managed with this plug-in.
    """
    return get_config_files_helper(get_managed_files())


def get_pending(use_colors):
    """ Return a string of the pending changes in all managed files. """
    return get_pending_diffs(
        [(ACL_RULE_FILE, ACL_RULE_SCRATCHFILE)],
        use_colors
    )


def abort_pending():
    remove_scratch_files_and_backups([
        ACL_RULE_SCRATCHFILE,
    ])


def get_managed_files():
    """ Return a list of managed files. """
    return [
        ACL_RULE_FILE,
    ]


def get_restart_commands(filename=None):
    """
    Return a list of commands that should be called for the given file after a rollback.
    Some plug-ins use this in commit_pending as well.
    """
    if filename == ACL_RULE_FILE:
        return [(BIN_CL_ACLTOOL, "--install")]

    return []


###########################
# NON-PLUGIN FUNCTIONS
###########################
###########################
# Static NAT
###########################
StaticNatRule = namedtuple(
    "StaticNatRule",
    [
        "direction",  # "snat" or "dnat"
        "protocol",
        "original_ip",
        "original_port",
        "translated_ip",
        "translated_port",
        "in_interface",
        "out_interface",
    ],
)


def parse_to_static_nat_rule(parser):
    """
    Parse a command parsed by network_docopt into a StaticNatRule data
    structure.

    Args:
        parser: a network_docopt parser (e.g., `daemon.nclu_parser`)

    Returns:
        A filled StaticNatRule.
    """
    assert "static" in parser.args, "Should only attempt to parse static NAT commands"
    match = parser.match
    argv_expanded = parser.argv_expanded
    # Skip to first arg after "net add nat static"
    cur_idx = 4

    # Get a dict with all the expected keys and default values
    command_kwargs = {
        key: None  # default to None for everything
        for key in StaticNatRule._fields
    }

    # "snat" or "dnat"
    command_kwargs["direction"] = argv_expanded[cur_idx]
    cur_idx = cur_idx + 1

    command_kwargs["protocol"] = argv_expanded[cur_idx]
    cur_idx = cur_idx + 1

    command_kwargs["original_ip"] = argv_expanded[cur_idx]
    cur_idx = cur_idx + 1

    # Is the next token a port?
    if "port" in str(match[cur_idx]):
        command_kwargs["original_port"] = argv_expanded[cur_idx]
        # Move to the next token
        cur_idx = cur_idx + 1

    # Do we have an interface in the commmand?
    if "interface" in str(match[cur_idx]):
        # Grab the key
        key = argv_expanded[cur_idx].replace("-", "_")
        # Move to the value
        cur_idx = cur_idx + 1
        value = argv_expanded[cur_idx]
        command_kwargs[key] = value

        # Moving on
        cur_idx = cur_idx + 1

    # Next token is "translate"; skip to something we care about
    cur_idx = cur_idx + 1

    command_kwargs["translated_ip"] = argv_expanded[cur_idx]
    cur_idx = cur_idx + 1

    # Do we have a next token? Is it a port?
    if len(match) > cur_idx and "port" in str(match[cur_idx]):
        command_kwargs["translated_port"] = argv_expanded[cur_idx]
        # Move to the next token
        cur_idx = cur_idx + 1

    return StaticNatRule(**command_kwargs)


def static_nat_rule_to_config_line(command):
    """
    Generate an iptables rule from a StaticNatRule.

    Args:
        command: a StaticNatRule

    Returns:
        A string iptables rule.
    """
    cmd = command
    is_snat = cmd.direction == "snat"
    opts = {
        "-t": "nat",
        "-A": ("POSTROUTING"
               if is_snat
               else "PREROUTING"),
        "-s": cmd.original_ip if is_snat else None,
        "-d": cmd.original_ip if not is_snat else None,
        "-p": cmd.protocol,
        "--sport": cmd.original_port if is_snat else None,
        "--dport": cmd.original_port if not is_snat else None,
        "--in-interface": cmd.in_interface,
        "--out-interface": cmd.out_interface,
        "-j": cmd.direction.upper(),
        "--to-source": (
            cmd.translated_ip
            + (":" + cmd.translated_port
               if cmd.translated_port
               else "")
            if is_snat else None
        ),
        "--to-destination": (
            cmd.translated_ip
            + (":" + cmd.translated_port
               if cmd.translated_port
               else "")
            if not is_snat else None
        ),
    }
    order = [
        "-t",
        "-A",
        "-s",
        "-d",
        "-p",
        "--sport",
        "--dport",
        "--in-interface",
        "--out-interface",
        "-j",
        "--to-source",
        "--to-destination",
    ]
    tokens = []
    for opt in order:
        value = opts[opt]
        if value is not None:
            tokens.append(opt)
            tokens.append(value)
    return " ".join(tokens)


###########################
# Dynamic NAT
###########################
DynamicNatRule = namedtuple(
    "DynamicNatRule",
    [
        "direction",  # "snat" or "dnat"
        "protocol",
        "source_ip",
        "destination_ip",
        "source_port",
        "destination_port",
        "in_interface",
        "out_interface",
        "pool_ip",  # IPv4 address or range of addresses
        "pool_port_range",  # Range of ports
    ],
)


def parse_to_dynamic_nat_rule(parser):
    args = parser.args
    assert "dynamic" in args, "Should only attempt to parse dynamic NAT commands"

    # Get a dict with all the expected keys and default values
    rule_kwargs = {
        key: None  # default to None for everything
        for key in DynamicNatRule._fields
    }
    rule_kwargs["direction"] = args.get("snat") or args.get("dnat")
    rule_kwargs["protocol"] = args.get("tcp") or args.get("udp") or args.get("icmp")

    ipv4_prefixlen = args.get("<ipv4/prefixlen>")
    # Coerce to list
    if type(ipv4_prefixlen) is str:
        ipv4_prefixlen = [ipv4_prefixlen]

    if "source-ip" in args:
        # Get the first ipv4/prefixlen
        rule_kwargs["source_ip"] = ipv4_prefixlen[0]
    if "destination-ip" in args:
        # Get the last ipv4/prefixlen
        rule_kwargs["destination_ip"] = ipv4_prefixlen[-1]

    rule_kwargs["source_port"] = args.get("<number-port-source>")
    rule_kwargs["destination_port"] = args.get("<number-port-destination>")
    # Convert port numbers to strings if not None
    for key in ["source_port", "destination_port"]:
        if rule_kwargs[key] is not None:
            rule_kwargs[key] = str(rule_kwargs[key])

    pool_ip = args.get("<ipv4>")
    # Validate the pool ip
    if pool_ip is None:
        pool_ip = args.get("<text-ipv4-range>")
        validate_pool_ip_range(pool_ip)
    rule_kwargs["pool_ip"] = pool_ip
    rule_kwargs["pool_port_range"] = args.get("<number-range>")

    if "out-interface" in args:
        rule_kwargs["out_interface"] = args.get("<interface>")
    elif "in-interface" in args:
        rule_kwargs["in_interface"] = args.get("<interface>")

    return DynamicNatRule(**rule_kwargs)


def dynamic_nat_rule_to_config_line(rule):
    is_snat = rule.direction == "snat"
    # Create the config value for the pool
    pool = ":".join([
        val
        for val in [rule.pool_ip, rule.pool_port_range]
        if val is not None
    ])
    opts = {
        "-t": "nat",
        "-A": ("POSTROUTING"
               if is_snat
               else "PREROUTING"),
        "-s": rule.source_ip,
        "-d": rule.destination_ip,
        "--sport": rule.source_port,
        "--dport": rule.destination_port,
        "-p": rule.protocol,
        "--in-interface": rule.in_interface,
        "--out-interface": rule.out_interface,
        "-j": rule.direction.upper(),
        "--to-source": pool if is_snat else None,
        "--to-destination": pool if not is_snat else None,
    }
    order = [
        "-t",
        "-A",
        "-s",
        "-d",
        "--in-interface",
        "--out-interface",
        "-p",
        "--sport",
        "--dport",
        "-j",
        "--to-source",
        "--to-destination",
    ]
    tokens = []
    for opt in order:
        value = opts[opt]
        if value is not None:
            tokens.append(opt)
            tokens.append(value)
    return " ".join(tokens)


def validate_pool_ip_range(pool_ip_token):
    """
    Validate the NAT ip pool token input.

    Args:
        pool_ip_token: String IP range token to validate.

    Raises:
        InvalidPoolError when the given token is not in the expected format.
    """
    start_ip, _, end_ip = pool_ip_token.partition("-")
    if not (match_ipv4_address(start_ip)
            and match_ipv4_address(end_ip)):
        raise InvalidPoolError('NAT pool IP range must be two IPv4 addresses in the following format: "XX.XX.XX.XX-YY.YY.YY.YY"')
    # Do nothing on success
    return


###########################
# Config file handling
###########################
def _static_options_to_nclu_command(opts):
    # Get a dict with all the expected keys and default values
    command_kwargs = {
        key: None  # default to None for everything
        for key in StaticNatRule._fields
    }

    command_kwargs["direction"] = opts.get("-j").lower()
    command_kwargs["protocol"] = opts.get("-p")

    # TODO: verify that the command is correct
    if "-s" in opts:
        command_kwargs["original_ip"] = opts.get('-s')
        command_kwargs["original_port"] = opts.get('--sport')
        command_kwargs["translated_ip"], _, command_kwargs["translated_port"] = opts.get("--to-source").partition(":")
        command_kwargs["out_interface"] = opts.get("--out-interface")
    else:
        command_kwargs["original_ip"] = opts.get('-d')
        command_kwargs["original_port"] = opts.get('--dport')
        command_kwargs["translated_ip"], _, command_kwargs["translated_port"] = opts.get("--to-destination").partition(":")
        command_kwargs["in_interface"] = opts.get("--in-interface")

    command_kwargs["translated_port"] = command_kwargs["translated_port"] or None

    cmd = StaticNatRule(**command_kwargs)

    # Reconstruct the NCLU command
    nclu_command = (
        "net add nat static"
        + " {direction} {protocol} {original_ip}".format(**vars(cmd))
        + (
            " {original_port}".format(**vars(cmd))
            if cmd.original_port is not None
            else ""
        )
        + (
            " in-interface {in_interface}".format(**vars(cmd))
            if cmd.in_interface is not None
            else ""
        )
        + (
            " out-interface {out_interface}".format(**vars(cmd))
            if cmd.out_interface is not None
            else ""
        )
        + " translate {translated_ip}".format(**vars(cmd))
        + (
            " {translated_port}".format(**vars(cmd))
            if cmd.translated_port is not None
            else ""
        )
    )

    return nclu_command


def _dynamic_options_to_nclu_command(opts):
    # Get a dict with all the expected keys and default values
    rule_kwargs = {
        key: None  # default to None for everything
        for key in DynamicNatRule._fields
    }

    rule_kwargs["direction"] = opts.get("-j").lower()
    rule_kwargs["protocol"] = opts.get("-p")

    rule_kwargs["source_ip"] = opts.get('-s')
    rule_kwargs["destination_ip"] = opts.get('-d')

    rule_kwargs["source_port"] = opts.get("--sport")
    rule_kwargs["destination_port"] = opts.get("--dport")

    is_snat = opts["-j"] == "SNAT"
    if is_snat:
        rule_kwargs["pool_ip"], _, rule_kwargs["pool_port_range"] = opts.get("--to-source").partition(":")
        rule_kwargs["out_interface"] = opts.get("--out-interface")
    else:
        rule_kwargs["pool_ip"], _, rule_kwargs["pool_port_range"] = opts.get("--to-destination").partition(":")
        rule_kwargs["in_interface"] = opts.get("--in-interface")

    # This might be an empty string
    rule_kwargs["pool_port_range"] = rule_kwargs["pool_port_range"] or None

    rule = DynamicNatRule(**rule_kwargs)

    # Reconstruct the NCLU command
    nclu_command = (
        "net add nat dynamic"
        + " {direction} {protocol}".format(**vars(rule))
        + (
            " source-ip {source_ip}".format(**vars(rule))
            if rule.source_ip is not None
            else ""
        )
        + (
            " source-port {source_port}".format(**vars(rule))
            if rule.source_port is not None
            else ""
        )
        + (
            " destination-ip {destination_ip}".format(**vars(rule))
            if rule.destination_ip is not None
            else ""
        )
        + (
            " destination-port {destination_port}".format(**vars(rule))
            if rule.destination_port is not None
            else ""
        )
        + (
            " in-interface {in_interface}".format(**vars(rule))
            if rule.in_interface is not None
            else ""
        )
        + (
            " out-interface {out_interface}".format(**vars(rule))
            if rule.out_interface is not None
            else ""
        )
        + " translate {pool_ip}".format(**vars(rule))
        + (
            " {pool_port_range}".format(**vars(rule))
            if rule.pool_port_range is not None
            else ""
        )
    )

    return nclu_command


def config_line_to_nclu_command(config_line):
    """
    Parse a config line and return the `add` command that would have created
    that line.

    Args:
        config_line: The string config line to convert

    Returns:
        A string NCLU command

    Raises:
        RuleParseError when we detect that `config_line` is not something we
            should worry about parsing.
    """
    tokens = config_line.strip().split()
    # The syntax alternates between option and value. Put that into a dict
    opts = dict(zip(tokens[0::2], tokens[1::2]))

    required_opts = ['-t', '-A', '-p', '-j']
    for opt in required_opts:
        if opt not in opts:
            raise RuleParseError('Missing required option: "{}"'.format(opt))

    opt_translated = opts.get("--to-source") or opts.get("--to-destination")

    # A range means it's dynamic
    is_dynamic = "-" in opt_translated
    if is_dynamic:
        nclu_command = _dynamic_options_to_nclu_command(opts)
    else:
        nclu_command = _static_options_to_nclu_command(opts)
    return nclu_command


def config_file_sections(contents):
    """
    Split an ACL config file into sections.

    Args:
        contents: string contents of the config file

    Returns:
        A dictionary data structure. (See `nclu/tests/test_nat.py` for an
        example of what this would look like.)
    """
    lines = contents.splitlines()
    current_section = None
    result = {}
    for idx, line in enumerate(lines):
        stripped = line.strip()
        # Handle section changes
        if stripped.startswith("[") and stripped.endswith("]"):
            if current_section is not None:
                result[current_section]["last"] = idx - 1
            current_section = stripped[1:-1]
            # Init the section
            result[current_section] = {
                "first": idx,
                "contents": [],
            }
        elif current_section is None:
            continue
        elif len(stripped) != 0:
            result[current_section]["contents"].append({
                "line_number": idx,
                "text": line,
            })

    # Add ending of last section
    if current_section is not None:
        result[current_section]["last"] = len(lines) - 1
    return result


def add_to_config_file(contents, section, added_line):
    """
    Add a line to the given section of an ACL config file.

    Args:
        contents: string contents of the file
        section: string section to add the line to
        added_line: string line to be added

    Returns:
        The updated contents of the file as a string.
    """
    added_line = added_line.strip("\n")
    lines = contents.splitlines()
    sections = config_file_sections(contents)

    section_info = sections[section]
    insertion_idx = section_info["last"] + 1

    result_lines = lines[:insertion_idx] + [added_line] + lines[insertion_idx:]
    return "\n".join(result_lines) + "\n"


def get_nclu_commands_by_line(contents):
    """
    Find all NCLU commands in the file and associate them with their respective
    line numbers.

    Args:
        contents: The string contents of the file.

    Returns:
        A list of 2-tuples `(line_number, nclu_command)` where `line_number` is
        the line number of the matching rule in the config file, and
        `nclu_command` is the command defined by the rule.
    """
    recognized = []

    sections = config_file_sections(contents)
    iptables = sections["iptables"]
    for entry in iptables["contents"]:
        try:
            recognized.append((
                entry["line_number"],
                config_line_to_nclu_command(entry["text"])
            ))
        except RuleParseError:
            pass  # Eh, it's probably fine

    return recognized


def del_from_config_file(contents, command):
    """
    Remove all lines that implement the given command from the contents of a
    config file.

    Args:
        contents: string contents of the file
        command: string NCLU command to remove from file

    Returns:
        Updated string contents of the config file.

    Raises:
        CannotDelError when the command cannot be deleted from the file.
    """
    commands_by_line = get_nclu_commands_by_line(contents)

    indexes_to_remove = [
        line[0]
        for line in commands_by_line
        if line[1] == command
    ]
    if len(indexes_to_remove) == 0:
        raise CannotDelError(
            '"nat" configuration does not have "{}"'.format(
                " ".join(command.split(" ")[3:])
            )
        )

    content_lines = contents.splitlines()

    # Start from end, so we don't change indexes
    for idx in reversed(indexes_to_remove):
        del content_lines[idx]
    return "\n".join(content_lines) + "\n"
