"""
Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
Copyright (C) 2016-2020 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt

----------------------------------------------------------------------------------------

Usage:
    net (add|del) time ntp source <interface>
    net add time ntp server <text> [iburst]
    net del time ntp server <text>
    net show time ntp servers

Options:
    iburst  : When the server is unreachable, send a burst of eight packets instead of the usual one
    ntp     : Network Time Protocol
    server  : Server
    servers : Servers
    source  : Source

"""

import logging
import os
import json
import re
from collections import OrderedDict
from nclu import files_match, ifname_expand_glob, WORKING_DIRECTORY
from nclu.NetDaemon import NetDaemon
from nclu.plugins import (
    get_config_files_helper,
    get_pending_diffs,
    make_scratchpads,
    merge_messages,
    persist_configurations,
    remove_scratch_files_and_backups,
    run_commands,
)
from nclu import (
    service_is_active,
    service_is_enabled,
)
from subprocess import CalledProcessError
from nclu.utils import check_output
from nclu.wrappers.ifupdown2 import ifname_in_etc_network_interfaces

NTP_CONF = "/etc/ntp.conf"
NTP_CONF_SCRATCHPAD = os.path.join(WORKING_DIRECTORY, "ntp", os.path.basename(NTP_CONF))


log = logging.getLogger("netd")


def commit_pending(verbose, net_del_all, changes):
    """
    Copy scratchpad files over configuration files, call necessary restart commands, etc.

    Return a 2-tuple of (bool, dict).  The boolean indicates if the commit succeeded.
    The dictionary is as described in nclu/plugins/__init__.py:execute_modular.

    net_del_all - A boolean; if True, the user is committing "net del all"
    """

    # If ntp isn't enabled to begin with, no reason to enable it in the mgmt
    # vrf
    restart_commands = []
    if '+iface mgmt' in changes and service_is_enabled('ntp.service'):
        restart_commands = [['/bin/systemctl', 'stop', 'ntp.service'],
                            ['/bin/systemctl', 'disable', 'ntp.service'],
                            ['/bin/systemctl', 'enable', 'ntp@mgmt.service'],
                            ['/bin/systemctl', 'start', 'ntp@mgmt.service']]
    # If ntp@mgmt isn't enabled, don't enable it globally
    elif '-iface mgmt' in changes and service_is_enabled('ntp@mgmt.service'):
        restart_commands = [['/bin/systemctl', 'stop', 'ntp@mgmt.service'],
                            ['/bin/systemctl', 'disable', 'ntp@mgmt.service'],
                            ['/bin/systemctl', 'enable', 'ntp.service'],
                            ['/bin/systemctl', 'start', 'ntp.service']]

    elif (not os.path.isfile(NTP_CONF_SCRATCHPAD) or files_match(NTP_CONF, NTP_CONF_SCRATCHPAD)):
        # There are no changes to commit.
        return (True, {})

    persist_configurations([(NTP_CONF, NTP_CONF_SCRATCHPAD)])

    if net_del_all and not service_is_enabled("ntp.service") \
       and not service_is_enabled("ntp@mgmt.service"):
        log.info("ntp@mgmt.service and ntp.service are disabled and will not be enabled.")

        if service_is_active("ntp@mgmt.service"):
            restart_commands = [["/bin/systemctl", "reset-failed", "ntp@mgmt.service"],
                                ["/bin/systemctl", "restart", "ntp@mgmt.service"]]
        elif service_is_active("ntp.service"):
            restart_commands = [["/bin/systemctl", "reset-failed", "ntp.service"],
                                ["/bin/systemctl", "restart", "ntp.service"]]
        else:
            log.info("ntp@mgmt.service and ntp.service are inactive and will not be started.")
            restart_commands = []
    elif not restart_commands:
        # Enable/reset/restart ntp.service.
        restart_commands = get_restart_commands(NTP_CONF)

    outcome, _, messages = run_commands(restart_commands, verbose)

    if outcome:
        # The commit succeeded.
        remove_scratch_files_and_backups([NTP_CONF_SCRATCHPAD])

    else:
        # The commit failed.  Revert the configuration.
        if os.path.isfile(NTP_CONF_SCRATCHPAD + ".bak"):
            persist_configurations([(NTP_CONF, NTP_CONF_SCRATCHPAD + ".bak")], False)
            _, _, revert_messages = run_commands(restart_commands, verbose)
            merge_messages(messages, revert_messages)

    return (outcome, messages)


def show_config_commands(ifupdown2_wrapper):
    """
    net show configuration commands
    """

    summary = {}
    show_config_summary(summary)

    supported = []

    if "ntp" in summary["time"]:
        if "servers" in summary["time"]["ntp"]:
            supported.extend(["net add time ntp server {0}".format(server) for server in summary["time"]["ntp"]["servers"]])

        if "source" in summary["time"]["ntp"]:
            supported.append("net add time ntp source {0}".format(summary["time"]["ntp"]["source"]))

    return (supported, [])


def show_config_summary(summary):
    """ net show configuration """

    assert isinstance(summary, dict)

    if "time" not in summary:
        summary["time"] = OrderedDict()

    servers, source = get_servers_and_source(NTP_CONF)
    if servers or source is not None:
        summary["time"]["ntp"] = OrderedDict()

        if servers:
            summary["time"]["ntp"]["servers"] = servers

        if source is not None:
            summary["time"]["ntp"]["source"] = source


def execute_command(net_daemon):
    """
    Given an instance of NetDaemon that has already run an input command through NetworkDocopt,
    perform the appropriate actions, and return the standard 2-tuple.
    """

    assert isinstance(net_daemon, NetDaemon)

    # Convenience variable
    tokens = net_daemon.nclu_parser.args
    assert "configuration" not in tokens, "\"net show configuration\" does not use this code path."
    assert "time" in tokens, "Why are you here?"

    if "add" in tokens or "del" in tokens:
        if "ntp" in tokens:
            if "server" in tokens or "source" in tokens:
                return execute_command_ntp(net_daemon)

    elif "show" in tokens:
        messages = {}

        if "ntp" in tokens:
            # net show time ntp servers

            # The ntpq command gets information from a server via L3 traffic.  If VRF is enabled
            # and ntpq is invoked from outside the VRF, ntpq will return the empty string.
            ntpq_command = "/usr/bin/ntpq -p"

            # Try to find a command that works rather or not VRF is enabled.
            command = "/bin/systemctl show -p MainPID ntp*"
            try:
                systemctl_output = check_output(command.split())

                # Get the first non-zero PID, if any, from the output.
                pid = re.search(r"""MainPID=(?P<pid>[1-9]\d*)""", systemctl_output)

                if pid is not None:
                    command = "/usr/bin/vrf task identify {0}".format(pid.group("pid"))
                    vrf_output = check_output(command.split())

                    # This command should work rather or not VRF is enabled.
                    ntpq_command = "/usr/bin/vrf task exec {0} ntpq -p".format(vrf_output)
            except CalledProcessError as e:
                # Hope for the best, which is that VRF is not enabled.
                messages["warning"] = ["If VRF is enabled, this output might be incomplete or incorrect."]
                log.warning("\"{0}\" failed: {1}.".format(command, e))

            # Call the command determined above.
            try:
                ntpq_output = check_output(ntpq_command.split())
                messages["output"] = ntpq_output.strip()
                return (True, messages)
            except CalledProcessError as e:
                log.exception(e)
                messages["error"] = ["The call to \"{0}\" failed.  See the log for details.".format(ntpq_command)]
                return (False, messages)

    return (False, {"error": ["NCLU doesn't know how to process this command."]})


def execute_command_ntp(net_daemon):
    """
    Return the standard 2-tuple for the following commands:

    net add time ntp server <text> [iburst]
    net del time ntp server <text>
    net (add|del) time ntp source <interface>
    """

    # Convenience variable
    tokens = net_daemon.nclu_parser.args

    if not os.path.isfile(NTP_CONF_SCRATCHPAD):
        make_scratchpads([(NTP_CONF, NTP_CONF_SCRATCHPAD, NTP_CONF_DEFAULT)])

    with open(NTP_CONF_SCRATCHPAD, "r+") as f:
        ntp_conf = f.read()

        if "server" in tokens:
            server_re = re.compile(r"""^\s*server\s+{0}(?P<iburst>\s+iburst)?\s*$""".format(re.escape(tokens["<text>"])), re.MULTILINE)
            server_line = server_re.search(ntp_conf)

            if "add" in tokens:
                if server_line is None:
                    # The server is not configured.  Append it.
                    f.write("server {0}{1}\n".format(tokens["<text>"], " iburst" if ("iburst" in tokens) else ''))
                    return (True, {})

                # The server is already configured.  Test for a no-op command.
                if ("iburst" in tokens and server_line.group("iburst") is not None) or \
                        ("iburst" not in tokens and server_line.group("iburst") is None):
                    log.info("ntp.py:execute_command_ntp: {0} already has \"{1}\".".format(NTP_CONF_SCRATCHPAD, server_line.group(0)))
                    return (True, {"info": ["\"{0}\" is already configured.".format(server_line.group(0))]})

                # The existing server line needs to be modified.
                ntp_conf = server_re.sub("server " + tokens["<text>"] + (" iburst" if ("iburst" in tokens) else '') + '\n', ntp_conf)

            elif "del" in tokens:
                if server_line is None:
                    log.info("ntp.py:execute_command_ntp: {0} does not have server \"{1}\".".format(NTP_CONF_SCRATCHPAD, tokens["<text>"]))
                    return (True, {"info": ["\"{0}\" is not configured.".format(tokens["<text>"])]})

                # The existing server line needs to be removed.
                ntp_conf = server_re.sub('', ntp_conf)

        elif "source" in tokens:
            assert "<interface>" in tokens and isinstance(tokens["<interface>"], basestring)
            if len(ifname_expand_glob(tokens["<interface>"])) > 1:
                return (False, {"error": ["This command does not accept a glob.  There is only one source."]})

            source_re = re.compile(r"""^\s*interface\s+listen\s+(?P<interface>\S+)\s*$""", re.MULTILINE)
            source_line = source_re.search(ntp_conf)

            if "add" in tokens:
                if source_line is None:
                    # The source is not configured.  Append it.
                    f.write("interface listen {0}\n".format(tokens["<interface>"]))
                    return (True, {})

                # The source is already configured.  Test for a no-op command.
                if tokens["<interface>"] == source_line.group("interface"):
                    log.info("ntp.py: {0} already has \"{1}\".".format(NTP_CONF_SCRATCHPAD, tokens["<interface>"]))
                    return (True, {"info": ["\"{0}\" is already configured.".format(tokens["<interface>"])]})

                # The existing source needs to be replaced.
                ntp_conf = source_re.sub("interface listen {0}\n".format(tokens["<interface>"]), ntp_conf)

            if "del" in tokens:
                if source_line is None:
                    log.info("ntp.py:execute_command_ntp: {0} does not have source \"{1}\".".format(NTP_CONF_SCRATCHPAD, tokens["<interface>"]))
                    return (True, {"info": ["\"{0}\" is not configured.".format(tokens["<interface>"])]})

                # The existing source needs to be removed.
                ntp_conf = source_re.sub('', ntp_conf)

        else:
            return (False, {"error": ["NCLU doesn't know how to process this command."]})

        # Update the scratchpad with the staged changes.
        f.seek(0)
        f.write(ntp_conf)
        f.truncate()

    return (True, {})


def del_all():
    """
    net del all

    Revert all configuration files associated with this plug-in to their defaults.
    """

    # Reset NTP_CONF.
    with open(NTP_CONF_SCRATCHPAD, 'w') as f:
        f.write(NTP_CONF_DEFAULT)


def get_config_files():
    """
    Return a string with the content of all existing, non-empty configuration files managed with this plug-in.
    """

    return get_config_files_helper(get_managed_files())


def get_pending(use_colors):
    """ Return a string of the pending changes in all managed files. """

    return get_pending_diffs([(NTP_CONF, NTP_CONF_SCRATCHPAD)], use_colors)


def abort_pending():
    remove_scratch_files_and_backups([NTP_CONF_SCRATCHPAD])


def get_managed_files():
    """ Return a list of managed files. """

    return [NTP_CONF]


def get_restart_commands(filename):
    """
    Return a list of commands that should be called for the given file after a rollback.
    Some plug-ins use this in commit_pending as well.
    """

    assert filename in get_managed_files(), "filename = {0}".format(filename)

    log.debug("mgmt interface in /etc/network/interfaces: {}".format(
        ifname_in_etc_network_interfaces('mgmt')))
    if filename == NTP_CONF:
        if not ifname_in_etc_network_interfaces('mgmt'):
            return [("/bin/systemctl", "stop", "ntp@mgmt.service"),
                    ("/bin/systemctl", "disable", "ntp@mgmt.service"),
                    ("/bin/systemctl", "enable", "ntp.service"),
                    ("/bin/systemctl", "reset-failed", "ntp.service"),
                    ("/bin/systemctl", "restart", "ntp.service")]
        else:
            return [("/bin/systemctl", "stop", "ntp.service"),
                    ("/bin/systemctl", "disable", "ntp.service"),
                    ("/bin/systemctl", "enable", "ntp@mgmt.service"),
                    ("/bin/systemctl", "reset-failed", "ntp@mgmt.service"),
                    ("/bin/systemctl", "restart", "ntp@mgmt.service")]

    raise RuntimeError("Unable to retreive restart commands for \"{0}\".".format(filename))


def get_servers_and_source(configuration):
    """
    Return the "server" and "source" components of the given configuration file.
    "server" lines include "iburst", if present.  The return type is (list, str or None).
    """

    assert configuration in (NTP_CONF, NTP_CONF_SCRATCHPAD)

    if not os.path.isfile(configuration):
        log.warning("ntp.py:get_servers_and_source: {0} is not a regular file.".format(configuration))
        return ([], None)

    servers = []
    source = None

    with open(configuration) as f:
        for line in f:
            line = line.strip()
            if line.startswith("server "):
                servers.append(line[7:])
            elif line.startswith("interface listen "):
                if source is not None:
                    # There should only be one source.  The file might have been manually misconfigured.
                    log.warning("{0} contains multiple \"source\" lines.".format(configuration))
                source = line[17:]

    return (servers, source)


NTP_CONF_DEFAULT = """\
# /etc/ntp.conf, configuration for ntpd; see ntp.conf(5) for help

driftfile /var/lib/ntp/ntp.drift

# Enable this if you want statistics to be logged.
#statsdir /var/log/ntpstats/

statistics loopstats peerstats clockstats
filegen loopstats file loopstats type day enable
filegen peerstats file peerstats type day enable
filegen clockstats file clockstats type day enable


# You do need to talk to an NTP server or two (or three).
#pool ntp.your-provider.example
# OR
#server ntp.your-provider.example

# pool.ntp.org maps to about 1000 low-stratum NTP servers.  Your server will
# pick a different set every time it starts up.  Please consider joining the
# pool: <http://www.pool.ntp.org/join.html>
server 0.cumulusnetworks.pool.ntp.org iburst
server 1.cumulusnetworks.pool.ntp.org iburst
server 2.cumulusnetworks.pool.ntp.org iburst
server 3.cumulusnetworks.pool.ntp.org iburst

# Access control configuration; see /usr/share/doc/ntp-doc/html/accopt.html for
# details.  The web page <http://support.ntp.org/bin/view/Support/AccessRestrictions>
# might also be helpful.
#
# Note that "restrict" applies to both servers and clients, so a configuration
# that might be intended to block requests from certain clients could also end
# up blocking replies from your own upstream servers.

# By default, exchange time with everybody, but don't allow configuration.
restrict -4 default kod notrap nomodify nopeer noquery limited
restrict -6 default kod notrap nomodify nopeer noquery limited

# Local users may interrogate the ntp server more closely.
restrict 127.0.0.1
restrict ::1

# Needed for adding pool entries
restrict source notrap nomodify noquery

# Clients from this (example!) subnet have unlimited access, but only if
# cryptographically authenticated.
#restrict 192.168.123.0 mask 255.255.255.0 notrust


# If you want to provide time to your local subnet, change the next line.
# (Again, the address is an example only.)
#broadcast 192.168.123.255

# If you want to listen to time broadcasts on your local subnet, de-comment the
# next lines.  Please do this only if you trust everybody on the network!
#disable auth
#broadcastclient

# Specify interfaces, don't listen on switch ports

interface listen eth0
"""
