"""
Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
Copyright (C) 2020 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt

----------------------------------------------------------------------------------------

Usage:

    net add port-mirror session <0-7> (ingress|egress) span src-port <interface-src-port> dst-port <interface-dst-port>
    net add port-mirror session <0-7> (ingress|egress) erspan src-port <interface-src-port> src-ip <ipv4-src-ip> dst-ip <ipv4-dst-ip>

    net del port-mirror session (<0-7>|all)

    net show port-mirror session (<0-7>|all) [json]

Options:
    port-mirror         : port-mirror
    session             : port-mirror session
    <0-7>               : port-mirror session number
    ingress             : ingress direction
    egress              : egress direction
    span                : SPAN
    erspan              : ERSPAN
    src-port            : Source port (swps or bond interfaces)
    dst-port            : Destination port (swp only)
    src-ip              : Source IP in erspan packet
    dst-ip              : Destination IP in erspan packet
"""
import logging
import os
import json
import collections
import itertools

from tabulate import tabulate

from nclu import (
    ifname_expand_glob,
    ifnames_to_glob,
    glob_to_numbers,
    numbers_to_glob,
    WORKING_DIRECTORY,
)

from nclu.plugins import (
    use_exceptions,
    get_config_files_helper,
    get_pending_diffs,
    make_scratchpads,
    persist_configurations,
    remove_scratch_files_and_backups,
    run_commands,
)

from nclu.utils import(
    get_platform_info,
)

from configobj import ConfigObj
from network_docopt import sort_for_humans

log = logging.getLogger("netd")

PORT_MIRROR_CONFIG = "/etc/cumulus/switchd.d/port-mirror.conf"
PORT_MIRROR_SCRATCHPAD = os.path.join(
    WORKING_DIRECTORY, "port_mirror", "port-mirror.conf"
)

# port mirror gets applied through this switchdctl command
PORT_MIRROR_COMMAND = "/usr/lib/cumulus/switchdctl"

# this is the full port mirror reload command
PORT_MIRROR_RELOAD = [PORT_MIRROR_COMMAND, "--load", PORT_MIRROR_CONFIG, "-prefix", "mirror"]

PORT_MIRROR_CONFIG_DEFAULT = """
# Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
# Copyright (C) 2020 Cumulus Networks, Inc. All rights reserved
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt
#
# [session_n]
# session-id = n
# mirror.session.n.direction = (ingress | egress)
# mirror.session.n.src = <swpx, bond>
# mirror.session.n.dest = (swpx | <src-ip> <dst-ip>)
# mirror.session.n.type = (span | erspan | none)
#
# Default is all sessions off
# mirror.session.all.type = none
mirror.session.all.type = none
"""

####################################
# EXPECTED NCLU PLUGIN FUNCTIONS
####################################
# Override keyword definitions created by other plugins/wrappers
per_command_helptext = {}


def reorder_session(lines, session):
    """
    Reorder the session keys.  Handle wrongly formatted session stanzas
    by ignoring them.

    The port-mirror.conf has ordering requirements on the keys.  Sort and
    rearrange the file to put the keys in each section in this order:
    [session_nn]
    session-id = nn
    mirror.session.nn.direction =
    mirror.session.nn.src =
    mirror.session.nn.dest =
    mirror.session.nn.type =

    NOTE: A deleted session will only have a type = none.  direction, src, dest
    will not be present
    """
    rewrite = []

    try:
        stanza_line_num = session['stanza'][0]

        # stanza, then session-id
        rewrite.append(session['stanza'])
        rewrite.append(session['session-id'])

        # session type of none is a deleted session and only has session-id
        # and type.  Other sessions have direction, src, dest
        if 'none' != session['type'][1].split('=')[1].strip():
            rewrite.append(session['direction'])
            rewrite.append(session['src'])
            rewrite.append(session['dest'])

        # type goes last
        rewrite.append(session['type'])

        for idx, item in enumerate(rewrite):
            # Rewrite the session stanza with keys in the correct order.
            lines[stanza_line_num + idx] = item[1]

    except Exception as e:
        log.exception(e)
        log.warning("Session {0} in port_mirror.conf is formatted incorrectly. Removing it".format(stanza_line_num))
        return


def sort_port_mirror_file(pending_file):
    # do nothing if the file does not exist.
    if not os.path.isfile(pending_file):
        return

    lines = []
    with open(pending_file, 'r') as f:
        lines = f.readlines()

        session = None

        # Create a list that has a tuple for each line in the file.  The tuple contains:
        # - index to the line in the file port-mirror.conf file
        # - the line from the file.
        # - the next line in the file (None if there is no next line i.e. EOF)
        #
        # This will allow the code to walk through the file and find the [session_] stanzas.
        # Each session stanza needs to be processed and sorted.
        for item in [i for i in itertools.izip_longest(range(len(lines)), lines, lines[1:])]:

            # Get the next line from the file.
            line = item[1].strip()

            if session:
                # in a session stanza, find the session-id, direction, src, dest, and type keys
                # and save those in the session dict.  Note: The session-id line is formatted
                # differently than the rest of them.
                if 'session-id' in line:
                    # The format of the session-id is session-id = n
                    session['session-id'] = item
                else:
                    # This begs for a regular expression but ... this works
                    # Format is mirror.session.n.<keyword> = <value>
                    suffix = line.split('=')[0].strip().split('.')[3]
                    session[suffix] = item

                    # if the following line in the file is the next session or
                    # None (means we are at EOF) then we have captured the entire
                    # session stanza.  Time to reorder it.
                    if item[2] is None or item[2].startswith('[session_'):
                        reorder_session(lines, session)
                        session = None

            elif line.startswith('[session_'):
                # else not in a session stanza.  If we find a net session stanza
                # start processing that.

                # create a session dict and save the 'stanza' as the first item.
                session = dict()
                session['stanza'] = item

    # write the sorted file
    with open(pending_file, 'w') as f:
        f.writelines(lines)


def commit_pending(verbose, net_del_all):
    """
    Copy scratchpad files over configuration files, call necessary restart commands, etc.

    Return a 2-tuple of (bool, dict).  The boolean indicates if the commit succeeded.
    The dictionary is as described in nclu/plugins/__init__.py:execute_modular.

    net_del_all - A boolean; if True, the user is committing "net del all"
    """

    sort_port_mirror_file(PORT_MIRROR_SCRATCHPAD)

    pending_port_mirror = get_pending_diffs(
        [(PORT_MIRROR_CONFIG, PORT_MIRROR_SCRATCHPAD)],
        False,
    )

    # copy the scratchpad over the config file to prep for running the
    # commands to apply the configuration.
    persist_configurations([(PORT_MIRROR_CONFIG, PORT_MIRROR_SCRATCHPAD)])

    commands = []

    if net_del_all:
        # Reload the port-mirror config on `del all`
        commands = get_restart_commands()
    elif pending_port_mirror != "":
        # For everything else, reload port-mirror config if pending changes
        commands = get_restart_commands()

    success = True
    if commands:
        success, _, messages = run_commands(commands, verbose)

    if success:
        abort_pending()
        return (True, {})
    else:
        # If the command fails, restore the config file to the backup to keep the changes
        # in 'pending'.
        if os.path.isfile(PORT_MIRROR_SCRATCHPAD + ".bak"):
            persist_configurations([(PORT_MIRROR_CONFIG, PORT_MIRROR_SCRATCHPAD + ".bak")], make_backup=False)
        messages["error"] = ["Unable to apply port mirror configuration.  See switchd.log for details"]
        return (False, messages)


def show_config_commands(_ifupdown2_wrapper):
    """
    net show configuration commands
    """

    command_tree = {}

    # Walk the tree and create the list of commands.
    commands = []

    iniobj = create_ini_object(PORT_MIRROR_CONFIG)

    # walk through the ini file and construct a 'net add' command for each section
    for section in iniobj.sections:
        args = dict(iniobj[section].items())

        # form every command from the session sections in the ini file.  Be careful of
        # keys that are not there
        try:

            session_id = args['session-id']
            if 'none' == args['mirror.session.{0}.type'.format(session_id)]:
                continue

            command = "net add port-mirror session {session_id} {direction} {pm_type} src-port {src_ports} ".format(
                session_id=session_id,
                direction=args['mirror.session.{0}.direction'.format(session_id)],
                pm_type=args['mirror.session.{0}.type'.format(session_id)],
                src_ports=args['mirror.session.{0}.src'.format(session_id)]
            )

            if 'erspan' == args['mirror.session.{0}.type'.format(session_id)]:
                dst = args['mirror.session.{0}.dest'.format(session_id)].split(' ')
                command_suffix = "src-ip {src_ip} dst-ip {dst_ip}".format(
                    src_ip=dst[0], dst_ip=dst[1])
            else:
                command_suffix = "dst-port {dst_ports}".format(
                    dst_ports=args['mirror.session.{0}.dest'.format(session_id)]
                )
        except Exception as e:
            # if something goes wrong finding all the keys, etc. for this command, skip it.  This
            # could happen if someone mucks with the config file for example and deletes something
            # important.
            log.exception(e)
            log.warning("Missing keys in {0}".format(PORT_MIRROR_CONFIG))
            continue

        # append this 'net add port-mirror...' command list
        commands.append(command + command_suffix)

    return (commands, [])


def show_config_summary(summary):
    """
    net show configuration port-mirror

    Populate the "summary" dictionary with information about the port_mirror configuration.
    """
    # nope
    pass


@use_exceptions(Exception)
def execute_command(net_daemon):
    """
    Execute an NCLU port-mirror command. Assumes that the NCLU core is only sending it
    valid commands.

    (Called by NCLU framework code.)

    Side effects:
        Writes files to NCLU scratchpad.

    Returns:
        A 2-tuple of (bool, dict), where bool represents whether or not the
        command succeeded, and dict contains output/error messages to display
        to user.
    """

    # Convenience variables
    tokens = net_daemon.nclu_parser.args
    argv_expanded = net_daemon.nclu_parser.argv_expanded
    assert "configuration" not in tokens, "\"net show configuration mirror\" does not use this code path."

    mode = next((
        mode_token
        for mode_token in ["add", "del", "show"]
        if mode_token in tokens
    ))

    if not os.path.isfile(PORT_MIRROR_CONFIG):
        with open(PORT_MIRROR_CONFIG, "w") as f:
            f.write(PORT_MIRROR_CONFIG_DEFAULT)

    # Figure out what the command is from the inputs arguments...

    if "show" == mode:
        session_id = tokens.get("<0-7>", "all")
        # `net show ...` is a special snowflake, so we'll handle it and bail.
        if "json" in tokens:
            return format_port_mirror_table_json(session_id)
        else:
            return format_port_mirror_table(session_id)

    elif argv_expanded == ["net", "del", "port-mirror", "session", "all"]:
        # 'net del port-mirror all' is a special snowflake.

        # EARLY RETURN
        return del_all()

    # Create the relevant scratchpads (if necessary) and track which ones were
    # created for this command. If the command creates scratchfiles, but
    # doesn't actually change anything, we'll clean up the scratchpad at the
    # end.
    did_create_port_mirror_scratchpad = False
    if not os.path.isfile(PORT_MIRROR_SCRATCHPAD):
        did_create_port_mirror_scratchpad = True
        make_scratchpads([
            (
                PORT_MIRROR_CONFIG,
                PORT_MIRROR_SCRATCHPAD,
                PORT_MIRROR_CONFIG_DEFAULT,
            ),
        ])

    iniobj = create_ini_object(PORT_MIRROR_SCRATCHPAD)

    if "add" == mode:
        session_id = tokens["<0-7>"]

        # Back end does not 'replay' sessions well so adding a session to
        # an existing session (which modifies the existing session) will fail
        # in the backend but not well.  So not allowing adds on existing sessions.
        # The session has to be deleted first.
        session_n = "session_{0}".format(session_id)
        session = create_ini_object(PORT_MIRROR_CONFIG).get(session_n)
        if session is not None:
            session_type = session.get('mirror.session.{0}.type'.format(session_id))
            if session_type != 'none':
                if did_create_port_mirror_scratchpad:
                    os.remove(PORT_MIRROR_SCRATCHPAD)
                raise Exception("Modifying sessions is not supported. Delete session {0} first".format(session_n))

        direction = tokens.get("ingress") or tokens.get("egress")
        pm_type = tokens.get("span") or tokens.get("erspan")

        src_ports = tokens["<interface-src-port>"]

        if "span" == pm_type:
            dst_ports = tokens["<interface-dst-port>"]
            write_section(iniobj, session_id, direction, pm_type, src_ports, dst_ports)

        elif "erspan" == pm_type:
            src_ip = tokens["<ipv4-src-ip>"]
            dst_ip = tokens["<ipv4-dst-ip>"]

            write_section(iniobj, session_id, direction, pm_type, src_ports, src_ip=src_ip, dst_ip=dst_ip)
    elif "del" == mode:
        # expecting 'net del port-mirror session'

        session_id = tokens["<0-7>"]

        # when a session is deleted, set the 'type' to none
        session_n = 'session_{0}'.format(session_id)

        try:
            # change the 'type' to none and remove the other keys from this session.
            session_values = {'mirror.session.{0}.type'.format(session_id) : 'none'}
            iniobj[session_n].update(session_values)

            iniobj[session_n].pop('mirror.session.{0}.direction'.format(session_id))
            iniobj[session_n].pop('mirror.session.{0}.src'.format(session_id))
            iniobj[session_n].pop('mirror.session.{0}.dest'.format(session_id))

        except KeyError:
            if did_create_port_mirror_scratchpad:
                os.remove(PORT_MIRROR_SCRATCHPAD)
            raise Exception("Session {0} does not exist".format(session_id))

        iniobj.write()


def del_all():
    """
    net del all

    Revert all configuration files associated with this plug-in to their defaults.
    """
    with open(PORT_MIRROR_SCRATCHPAD, "w") as f:
        f.write(PORT_MIRROR_CONFIG_DEFAULT)


def get_config_files():
    """
    Return a string with the content of all existing, non-empty configuration files managed with this plug-in.
    """
    return get_config_files_helper(get_managed_files())


def get_pending(use_colors):
    """ Return a string of the pending changes in all managed files. """
    return get_pending_diffs(
        [
            (PORT_MIRROR_CONFIG, PORT_MIRROR_SCRATCHPAD),
        ],
        use_colors
    )


def abort_pending():
    remove_scratch_files_and_backups([
        PORT_MIRROR_SCRATCHPAD,
    ])


def get_managed_files():
    """ Return a list of managed files. """

    return [PORT_MIRROR_CONFIG]


def get_restart_commands(filename=None):
    """
    Return a list of commands that should be called for the given file after a rollback.
    Some plug-ins use this in commit_pending as well.
    """
    port_mirror_commands = []

    platform_info = get_platform_info()

    # Get the platform info.  If on a Vx ('software' platform) we can't support any
    # switchdctl commands so return empty command list.
    hw_sw = platform_info.get('platform')
    if hw_sw is None or hw_sw == 'software':
        return port_mirror_commands

    # otherwise on a hardware platform return the command to reload port-mirror config
    # if switchdctl is present.
    if os.path.isfile(PORT_MIRROR_COMMAND):
        port_mirror_commands.append(PORT_MIRROR_RELOAD)

    return port_mirror_commands


###########################
# NON-PLUGIN FUNCTIONS
###########################
def create_ini_object(from_file):
    try:
        return(ConfigObj(
            infile=from_file,
            write_empty_values=True,
            list_values=False))
    except Exception as e:
        log.exception(e)
        raise Exception("Unable to parse {0}".format(from_file))


def write_section(iniobj, session_id, direction, pm_type, src_ports, dst_ports=None, src_ip=None, dst_ip=None):

    session_n = 'session_{0}'.format(session_id)

    # common key:value pairs for all session types.
    session_values = {
        'session-id' : session_id,
        'mirror.session.{0}.direction'.format(session_id) : direction,
        'mirror.session.{0}.type'.format(session_id) : pm_type,
        'mirror.session.{0}.src'.format(session_id) : src_ports,
    }

    # the destination keys are different for span versus erspan
    if 'span' == pm_type:
        assert dst_ports is not None
        session_values['mirror.session.{0}.dest'.format(session_id)] = dst_ports
    else:
        assert src_ip is not None and dst_ip is not None
        session_values['mirror.session.{0}.dest'.format(session_id)] = '{0} {1}'.format(src_ip, dst_ip)

    # Make sure the global mirror.session.all.type = none is removed when any entry is added if it exists.
    try:
        iniobj.pop('mirror.session.all.type')
    except KeyError:
        pass

    # Add the new section to the conf file.
    iniobj[session_n] = session_values

    # write out the scratch file
    iniobj.write()


PORT_MIRROR_TABLE_COLUMNS = [
    "session-id",
    "direction",
    "type",
    "src",
    "dest",
]


def get_pm_display_data(iniobj, session_num):
    pm_data = collections.OrderedDict()

    try:
        # walk through the sections to get the port mirror data.  Skip any sessions
        # that are marked with type 'none' as these have been deleted.
        for section in sort_for_humans(iniobj.sections):
            section_dict = dict(iniobj[section].items())
            session_id = section_dict['session-id']
            if 'all' == session_num or session_id == str(session_num):
                if section_dict['mirror.session.{0}.type'.format(section_dict['session-id'])] != 'none':
                    pm_data[section] = dict(iniobj[section].items())
    except Exception as e:
        log.exception(e)
        # EARLY RETURN
        return {}

    return pm_data


def format_port_mirror_table(session_num):

    iniobj = create_ini_object(PORT_MIRROR_CONFIG)

    port_mirror_table = []

    try:
        pm_data = get_pm_display_data(iniobj, session_num)

        for section in pm_data:
            # convenience.  save the session_id in a local variable
            session_id = pm_data[section].get('session-id')

            # build a session entry table for each session.
            session_table = []
            session_table.append(session_id)
            for column in PORT_MIRROR_TABLE_COLUMNS:
                # convenience. build the string for the mirror key with the 'column' keyword
                mirror_key = 'mirror.session.{0}.{1}'.format(session_id,column)
                if pm_data[section].get(mirror_key) is not None:
                    session_table.append(pm_data[section].get(mirror_key))

            # add the session table to the parent table
            port_mirror_table.append(session_table)

    except KeyError:
        raise Exception("Unable to find port mirror data for session {0} from {1}".
                        format(session_id, PORT_MIRROR_CONFIG))

    table = tabulate(
        port_mirror_table,
        headers=PORT_MIRROR_TABLE_COLUMNS,
        numalign="right"
    )

    return table


def format_port_mirror_table_json(session_num):

    iniobj = create_ini_object(PORT_MIRROR_CONFIG)

    try:
        pm_data = get_pm_display_data(iniobj, session_num)

        return json.dumps(
            pm_data,
            indent=4,
            sort_keys=True,
            # HACK: use "," as separator to prevent trailing whitespace
            separators=(",", ": "),
        )
    except KeyError:
        raise Exception("Unable to find port mirror data for session {0} from {1}".
                        format(session_id, PORT_MIRROR_CONFIG))
