"""
Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
Copyright (C) 2019 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt

----------------------------------------------------------------------------------------

Usage:
    net show port-security [<interface>] [json]

    net (add|del) interface <interface> port-security
    net (add|del) interface <interface> port-security sticky-mac
    net (add|del) interface <interface> port-security sticky-mac aging
    net add interface <interface> port-security sticky-mac timeout <0-3600>
    net del interface <interface> port-security sticky-mac timeout
    net (add|del) interface <interface> port-security allowed-mac <mac>
    net del interface <interface> port-security allowed-mac
    net add interface <interface> port-security mac-limit <0-512>
    net del interface <interface> port-security mac-limit
    net add interface <interface> port-security violation (shutdown|restrict)
    net add interface <interface> port-security violation timeout <0-3600>
    net del interface <interface> port-security violation timeout

Options:
    <interface> : A swp interface
    <mac> : A MAC address
    port-security : Port security
    sticky-mac : Sticky MAC
    aging: Sticky MAC aging
    allowed-mac : Allowed MAC addresses
    violation : Action to take on port security violation
    shutdown : Shut down the port
    restrict : Drop unknown packets
    timeout : Port security timeout
    mac-limit : Limit the number of MACs on this port
"""

import logging
import os
import re
from json import dumps, loads
from glob import glob
from nclu import (
    ifname_expand_glob,
    ifnames_to_glob,
    WORKING_DIRECTORY,
)
from nclu.plugins import (
    get_config_files_helper,
    persist_configurations,
    get_pending_diffs,
    make_scratchpads,
    remove_scratch_files_and_backups,
    use_exceptions,
)
from network_docopt import sort_for_humans
from tabulate import tabulate
from collections import namedtuple


log = logging.getLogger("netd")


SWITCHD_CONF_FILE = "/etc/cumulus/switchd.d/port_security.conf"

SWITCHD_CONFIG_DIR = "/cumulus/switchd/config"
SWITCHD_INTERFACES_DIR = SWITCHD_CONFIG_DIR + "/interface"

PORT_SECURITY_DUMP_FILE = SWITCHD_CONFIG_DIR + "/port_security_dump"
PORT_SECURITY_MAC_DUMP_FILE = SWITCHD_CONFIG_DIR + "/port_security_mac_dump"

TMPL_PORT_SECURITY_DIR = SWITCHD_INTERFACES_DIR + "/{}/port_security"

TMPL_PORT_SECURITY_ENABLE_FILE = TMPL_PORT_SECURITY_DIR + "/enable"
TMPL_PORT_SECURITY_STICKY_MAC_FILE = TMPL_PORT_SECURITY_DIR + "/sticky_mac"
TMPL_PORT_SECURITY_STICKY_AGING_FILE = TMPL_PORT_SECURITY_DIR + "/sticky_aging"
TMPL_PORT_SECURITY_STICKY_TIMEOUT_FILE = TMPL_PORT_SECURITY_DIR + "/sticky_timeout"
TMPL_PORT_SECURITY_ALLOWED_MAC_FILE = TMPL_PORT_SECURITY_DIR + "/static_mac"
TMPL_PORT_SECURITY_MAC_LIMIT_FILE = TMPL_PORT_SECURITY_DIR + "/mac_limit"
TMPL_PORT_SECURITY_VIOLATION_MODE_FILE = (
    TMPL_PORT_SECURITY_DIR + "/violation_mode"
)
TMPL_PORT_SECURITY_VIOLATION_TIMEOUT_FILE = (
    TMPL_PORT_SECURITY_DIR + "/violation_timeout"
)


PATH_TEMPLATES = [
    TMPL_PORT_SECURITY_ENABLE_FILE,
    TMPL_PORT_SECURITY_STICKY_MAC_FILE,
    TMPL_PORT_SECURITY_ALLOWED_MAC_FILE,
    TMPL_PORT_SECURITY_MAC_LIMIT_FILE,
    TMPL_PORT_SECURITY_VIOLATION_MODE_FILE,
    TMPL_PORT_SECURITY_VIOLATION_TIMEOUT_FILE,
]


# Default file contents
DEFAULT_PORT_SECURITY_ENABLE = "0\n"
DEFAULT_PORT_SECURITY_STICKY_MAC = "0\n"
DEFAULT_PORT_SECURITY_STICKY_AGING = "0\n"
DEFAULT_PORT_SECURITY_STICKY_TIMEOUT = "1440\n"
DEFAULT_PORT_SECURITY_ALLOWED_MAC = "\n"
DEFAULT_PORT_SECURITY_MAC_LIMIT = "10\n"
DEFAULT_PORT_SECURITY_VIOLATION_MODE = "1\n"
DEFAULT_PORT_SECURITY_VIOLATION_TIMEOUT = "1800\n"


# Map the violation mode code to its pretty string counterpart
MAP_CODE_TO_VIOLATION_MODE = {
    "0": "shutdown",
    "1": "restrict",
}


class InvalidInterfaceError(Exception):
    """
    Error raised when a user tries to access an interface that can't have port
    security configured on it.

    Args:
        interfaces: The list of invalid interfaces.
    """
    def __init__(self, interfaces):
        self.interfaces = interfaces
        self.glob = ifnames_to_glob(interfaces)

    def __str__(self):
        return "{} not found in {}".format(
            self.glob,
            SWITCHD_INTERFACES_DIR,
        )


################################
# Expected NCLU plugin functions
################################
@use_exceptions(Exception)
def execute_command(net_daemon):
    """
    Execute an NCLU port-security command. Assumes that the NCLU core is only
    sending it valid port-security commands.

    (Called by NCLU framework code.)

    Side effects:
        Writes files to NCLU scratchpad.

    Returns:
        A 2-tuple of (bool, dict), where bool represents whether or not the
        command succeeded, and dict contains output/error messages to display
        to user.
    """
    tokens = net_daemon.nclu_parser.argv_expanded[1:]
    # Placeholder for iface name argument. This will be replaced with each of
    # the expanded iface_names
    _iface_placeholder = object()

    # The function we'll be using to handle the command
    command_handler = None
    # The arguments to pass into the command handler, as
    # `command_handler(*command_args)`
    command_args = []
    # The list of interfaces expanded from the glob
    iface_names = None

    # Should we enable port security on this port in addition to the action
    # specified by the command?
    needs_enable = False
    # Should we reset all the port sec parameters on this port to the defaults?
    needs_reset_defaults = False
    needs_enable_sticky_mac = False
    needs_reset_sticky_mac_defaults = False
    # Should we show the output?
    should_show = False

    # Parse the command and assign appropriate values to `command_handler`,
    # `command_args`, and `iface_names`.
    mode = tokens[0]
    if mode == "add" or mode == "del":
        is_add = mode == "add"

        # These are the tokens we expect so far:
        # `net (add|del) interface <interface> port-security`
        assert tokens[1] == "interface"
        iface_token = tokens[2]
        assert tokens[3] == "port-security"

        iface_names = ifname_expand_glob(iface_token)
        # Check that the specified interfaces are actually valid for port
        # security
        valid_ifaces = list_supported_ports()
        invalid_ifaces = set(iface_names) - set(valid_ifaces)
        if invalid_ifaces != set():
            raise InvalidInterfaceError(list(invalid_ifaces))

        port_sec_subcommand = tokens[4] if len(tokens) > 4 else None
        if is_add and port_sec_subcommand is not None:
            needs_enable = True
        if not is_add and port_sec_subcommand is None:
            needs_reset_defaults = True
        if port_sec_subcommand == "sticky-mac":
            if is_add and len(tokens) > 5:
                needs_enable_sticky_mac = True
            if not is_add and len(tokens) == 5:
                needs_reset_sticky_mac_defaults = True
        # Assign the handler/args based off `port_sec_subcommand`.
        command_handler, command_args = {
            None: lambda: (
                stage_port_security_enable,
                [
                    is_add,
                    _iface_placeholder
                ],
            ),
            "sticky-mac": lambda: (
                (
                    # `sticky-mac`
                    stage_port_security_sticky_mac,
                    [
                        is_add,
                        _iface_placeholder
                    ],
                ) if len(tokens) == 5 else (
                    # `sticky-mac aging`
                    stage_port_security_sticky_aging,
                    [
                        is_add,
                        _iface_placeholder
                    ],
                ) if tokens[5] == "aging" else (
                    # `sticky-mac timeout <timeout>`
                    stage_port_security_sticky_timeout,
                    [
                        is_add,
                        _iface_placeholder,
                        tokens[6] if len(tokens) > 6 else None
                    ],
                )
            ),
            "allowed-mac": lambda: (
                stage_port_security_allowed_mac,
                [
                    is_add,
                    _iface_placeholder,
                    tokens[5] if len(tokens) > 5 else None,
                ],
            ),
            "mac-limit": lambda: (
                stage_port_security_mac_limit,
                [
                    is_add,
                    _iface_placeholder,
                    tokens[5] if len(tokens) > 5 else None,
                ],
            ),
            "violation": lambda: (
                (
                    # `violation timeout [<number>]`
                    stage_port_security_violation_timeout,
                    [
                        is_add,
                        _iface_placeholder,
                        tokens[6] if len(tokens) > 6 else None,
                    ]
                ) if tokens[5] == "timeout" else (
                    # `violation (shutdown|restrict)`
                    stage_port_security_violation_mode,
                    [
                        is_add,
                        _iface_placeholder,
                        tokens[5],
                    ]
                )
            )
        }[port_sec_subcommand]()
    elif mode == "show":
        should_show = True
        # These are the tokens we expect so far:
        # `net show port-security`
        assert tokens[1] == "port-security"
        is_json = tokens[-1] == "json"

        if is_json:
            # Remove the "json" token to simplify selection of the handler
            tokens = tokens[:-1]

        if len(tokens) == 3:
            # Get iface_names
            iface_token = tokens[2]
            iface_names = ifname_expand_glob(iface_token)
            # Check that the specified interfaces are actually valid for port
            # security
            valid_ifaces = list_supported_ports()
            invalid_ifaces = set(iface_names) - set(valid_ifaces)
            if invalid_ifaces != set():
                raise InvalidInterfaceError(
                    interfaces=list(invalid_ifaces)
                )

        # Assign handler/args based off the number of tokens.
        command_handler, command_args = {
            2: (
                show_port_security_summary,
                [
                    is_json,
                ],
            ),
            3: (
                show_port_security_interface,
                [
                    _iface_placeholder,
                    is_json,
                ]
            ),
        }[len(tokens)]

    # Call the command_handler and assign the results as a string or None.
    result = None
    if iface_names is None:
        result = command_handler(*command_args)
    else:
        # Grab output for all globbed interfaces and concatenate into result
        # string.
        index = command_args.index(_iface_placeholder)
        outputs = []
        for iface in iface_names:
            command_args[index] = iface
            output = command_handler(*command_args)
            if needs_enable:
                stage_port_security_enable(True, iface)
            elif needs_reset_defaults:
                # Do a localized "net del all" for this port
                stage_port_security_sticky_mac(False, iface)
                stage_port_security_sticky_aging(False, iface)
                stage_port_security_sticky_timeout(False, iface)
                stage_port_security_allowed_mac(False, iface)
                stage_port_security_mac_limit(False, iface)
                stage_port_security_violation_mode(False, iface)
                stage_port_security_violation_timeout(False, iface)
            if needs_enable_sticky_mac:
                stage_port_security_sticky_mac(True, iface)
            elif needs_reset_sticky_mac_defaults:
                stage_port_security_sticky_aging(False, iface)
                stage_port_security_sticky_timeout(False, iface)
            if output is not None:
                outputs.append(str(output))
        # Only set the result if we have something to show.
        if should_show and outputs != []:
            result = "\n".join(outputs)

    return result


def get_pending(use_colors):
    """
    Get the diffs representing pending configuration file changes.

    (Called by NCLU framework code.)

    Args:
        use_colors: Whether or not to display the output with colors

    Returns:
        The diffs as a string.
    """
    pending_files = list_pending_files()
    return get_pending_diffs(pending_files, use_colors)


def abort_pending():
    """
    Remove all pending scratch files managed by port_security plugin.

    (Called by NCLU framework code.)

    Side effects:
        Deletes all scratchpad files managed by port_security plugin.
    """
    scratchpad_files = [
        pending[1]
        for pending in list_pending_files()
    ]
    remove_scratch_files_and_backups(scratchpad_files)


def del_all():
    """
    Stage defaults for all files managed by port_security plugin.

    (Called by NCLU framework code.)

    Side effects:
        Writes default values to scratchpad for all files managed by
        port_security plugin.
    """
    switchd_ports = list_supported_ports()
    for ifname in switchd_ports:
        stage_port_security_enable(False, ifname)


def get_restart_commands(filename=None):
    """
    Return a list of commands that should be called for the given file after a
    rollback.  Some plug-ins use this in commit_pending as well.

    (Called by NCLU framework code.)

    Args:
        filename: String filename of a file that's been updated.

    Returns:
        A list of command tuples.
    """
    if filename == SWITCHD_CONF_FILE:
        return [("/bin/systemctl", "restart", "switchd.service")]
    return []


def get_managed_files():
    """
    Get the list of files managed by port_security plugin. This is used by
    rollback. All FUSE files under /cumulus/switchd are omitted from rollback,
    so they're omitted here as well.

    (Called by NCLU framework code.)

    Returns:
        A list of string file paths.
    """
    return [SWITCHD_CONF_FILE]


def get_config_files():
    """
    Get the contents of all existing, non-empty files.

    (Called by NCLU framework code.)

    Returns:
        A string with the content of all existing, non-empty configuration
        files managed with this plug-in.
    """
    return get_config_files_helper(get_managed_files())


def show_config_commands(ifupdown2_wrapper):
    """
    net show configuration commands

    (Called by NCLU framework code.)

    Args:
        ifupdown2_wrapper: Apparently expected by the calling framework code
            for some reason. Not used by port_security plugin though.

    Returns:
        A 2-tuple `(commands, unsupported)` where `commands` list of NCLU
        commands as strings, and `unsupported` is a list of unsupported
        commands.
    """
    # Aggregate the ifnames that each command is being called on, so we can
    # globify them.
    command_template_to_ifnames = {}
    # Add commands for every place where the current value is not the default
    for ifname in list_supported_ports():
        details = loads(show_port_security_interface(ifname, True))

        enable = details["enable"]
        # EARLY CONTINUE
        if not enable:
            # We should never list port-security commands when it's disabled.
            continue
        mac_limit = details["mac_limit"]
        sticky_mac = details["sticky_mac"]
        sticky_aging = details["sticky_aging"]
        sticky_timeout = details["sticky_timeout"]
        static_mac = [
            mac[DUMP_KEY_MAC_ADDRESS]
            for mac in details["mac_addresses"]
            if mac[DUMP_KEY_TYPE] == "Static"
        ]
        violation_mode = details["violation_mode"]
        violation_timeout = details["violation_timeout"]
        if enable:
            command_template = "net add interface {} port-security"
            command_template_to_ifnames.setdefault(
                command_template,
                [],
            ).append(ifname)
        if str(mac_limit) != DEFAULT_PORT_SECURITY_MAC_LIMIT.strip():
            command_template = (
                "net add interface {} port-security mac-limit "
                + str(mac_limit)
            )
            command_template_to_ifnames.setdefault(
                command_template,
                []
            ).append(ifname)
        if sticky_mac:
            command_template = "net add interface {} port-security sticky-mac"
            command_template_to_ifnames.setdefault(
                command_template,
                [],
            ).append(ifname)
        if sticky_aging:
            command_template = "net add interface {} port-security sticky-mac aging"
            command_template_to_ifnames.setdefault(
                command_template,
                [],
            ).append(ifname)
        if str(sticky_timeout) != DEFAULT_PORT_SECURITY_STICKY_TIMEOUT.strip():
            command_template = (
                "net add interface {} port-security sticky-mac timeout "
                + str(sticky_timeout)
            )
            command_template_to_ifnames.setdefault(
                command_template,
                [],
            ).append(ifname)
        if static_mac != []:
            for mac in static_mac:
                command_template = (
                    "net add interface {} port-security allowed-mac "
                    + mac
                )
                command_template_to_ifnames.setdefault(
                    command_template,
                    []
                ).append(ifname)
        if violation_mode != MAP_CODE_TO_VIOLATION_MODE[DEFAULT_PORT_SECURITY_VIOLATION_MODE.strip()]:
            command_template = (
                "net add interface {} port-security violation "
                + violation_mode
            )
            command_template_to_ifnames.setdefault(
                command_template,
                []
            ).append(ifname)
        if str(violation_timeout) != DEFAULT_PORT_SECURITY_VIOLATION_TIMEOUT.strip():
            command_template = (
                "net add interface {} port-security violation timeout "
                + str(violation_timeout)
            )
            command_template_to_ifnames.setdefault(
                command_template,
                []
            ).append(ifname)

    commands = [
        tmpl.format(ifnames_to_glob(ifnames))
        for tmpl, ifnames in sorted(command_template_to_ifnames.items())
    ]

    return (commands, [])


def show_config_summary(summary):
    """
    Expected by framework code for `net show config <module-name>` and
    `net show config`. The port_security plugin isn't really messing with that,
    so this function is a noop.

    (Called by NCLU framework code.)
    """
    pass


def commit_pending(verbose, net_del_all):
    """
    Copy scratchpad files over configuration files, call necessary restart
    commands, etc.

    (Called by NCLU framework code.)

    Args:
        net_del_all: A boolean; if True, the user is committing "net del all"
    Returns:
        A 2-tuple of (bool, dict).  The boolean indicates if the commit
        succeeded.  The dictionary is as described in
        nclu/plugins/__init__.py:execute_modular.
    """
    pending = get_pending(use_colors=False)
    if pending == "":
        # Nothing to do
        return (True, {})
    pending_files = list_pending_files()
    # Ensure that the pending_files' dst_paths exist.
    # NOTE: This is probably not necessary anymore, now that
    #       /etc/cumulus/switchd.d/ is in CL builds. Removing this code is not
    #       worth the risk to me right now.
    for pending in pending_files:
        path = pending[0]
        dirname = os.path.dirname(path)
        try:
            os.makedirs(dirname)
        except Exception:
            # Not a problem if it already exists
            pass

    persist_configurations(pending_files)

    # Apply the updated contents of the switchd conf file to the switchd FUSE
    # nodes.
    updated_conf = SWITCHD_CONF_FILE
    if os.path.exists(updated_conf):
        with open(updated_conf) as f:
            new_contents = f.read()
        apply_switchd_conf(new_contents)
    # Clean up the scratch files
    abort_pending()
    return (True, {})


################
# Utility things
################
def list_pending_files():
    """
    Find the files in scratchpad and return them alongside the paths to their
    counterparts on the main filesystem.

    Only returns results that have an existing scratch file.

    Returns:
        A list of 2-tuples with (actual_path, scratchpad_path)
    """
    files = get_managed_files()

    pending_files = []
    for actual_path in files:
        scratchpad_path = WORKING_DIRECTORY + actual_path
        # Only include it in results if it's got a scratch file.
        if os.path.exists(scratchpad_path):
            pending_files.append((actual_path, scratchpad_path))

    return pending_files


def list_supported_ports():
    """
    Find the ports that we can configure port security on.

    Returns:
        A list of string interface names.
    """
    try:
        supported_ports = [
            iface
            for iface in os.listdir(SWITCHD_INTERFACES_DIR)
            # Filter out interfaces without port-security files
            if os.path.exists(
                os.path.join(
                    SWITCHD_INTERFACES_DIR, iface, "port_security/enable"
                )
            )
        ]
    except OSError:
        # If SWITCHD_INTERFACES_DIR isn't accessible or present, then we'll
        # just treat it as if there are no interfaces that support port
        # security.
        return []
    return sort_for_humans(supported_ports)


def write_file(path, contents):
    """
    Convenience wrapper for `open`. (This is mostly useful to make the test
    code easier to mock.)

    Args:
        path: String path of file.
        contents: String contents to write to file.

    Side effects:
        Writes a file.
    """
    with open(path, "w") as f:
        f.write(contents)


#################################
# Handlers for add|del commands
#################################
def _convert_dst_path_to_switchd_conf_key(dst_path):
    # Strip the part of the path we don't care about and replace the slashes
    # with dots.
    first_index = len("/cumulus/switchd/config/")
    return dst_path[first_index:].replace("/", ".")


def _stage_switchd_conf_update(dst_path, new_value):
    """
    Stage the update for the switchd configuration file for persistence.

    Args:
        dst_path (str): The path to the switchd config file that's being
            updated.
        new_value (str): The new value to persist in switchd config.

    Side effects:
        Writes pending file for `/etc/cumulus/switchd.d/port_security.conf` to
        scratchpad.
    """
    key = _convert_dst_path_to_switchd_conf_key(dst_path)
    overrides = {
        key: new_value
    }
    scratch_switchd_conf = WORKING_DIRECTORY + SWITCHD_CONF_FILE
    if not os.path.isfile(scratch_switchd_conf):
        make_scratchpads(
            [(SWITCHD_CONF_FILE, scratch_switchd_conf, "")]
        )

    with open(scratch_switchd_conf) as f:
        existing = f.read()

    updated = edit_switchd_conf(existing, overrides)

    write_file(scratch_switchd_conf, updated)


def stage_port_security_enable(is_add, interface):
    """
    Stage configuration files for
    `net (add|del) interface <interface> port-security`

    Args:
        is_add: True if `net add`, False if `net del`
        interface: The string name of a single interface to configure.

    Side effects:
        Writes configuration files to scratchpad.
    """
    dst_path = TMPL_PORT_SECURITY_ENABLE_FILE.format(interface)
    if is_add:
        new_config = "1\n"
    else:
        new_config = DEFAULT_PORT_SECURITY_ENABLE
    _stage_switchd_conf_update(dst_path, new_config.strip("\n"))
    return None


def stage_port_security_sticky_mac(is_add, interface):
    """
    Stage configuration files for
    `net (add|del) interface <interface> port-security sticky-mac`

    Args:
        is_add: True if `net add`, False if `net del`
        interface: The string name of a single interface to configure.

    Side effects:
        Writes configuration files to scratchpad.
    """
    dst_path = TMPL_PORT_SECURITY_STICKY_MAC_FILE.format(interface)
    if is_add:
        new_config = "1\n"
    else:
        new_config = DEFAULT_PORT_SECURITY_STICKY_MAC
    _stage_switchd_conf_update(dst_path, new_config.strip("\n"))
    return None


def stage_port_security_sticky_aging(is_add, interface):
    """
    Stage configuration files for
    `net (add|del) interface <interface> port-security sticky-mac aging`

    Args:
        is_add: True if `net add`, False if `net del`
        interface: The string name of a single interface to configure.

    Side effects:
        Writes configuration files to scratchpad.
    """
    dst_path = TMPL_PORT_SECURITY_STICKY_AGING_FILE.format(interface)
    if is_add:
        new_config = "1\n"
    else:
        new_config = DEFAULT_PORT_SECURITY_STICKY_AGING
    _stage_switchd_conf_update(dst_path, new_config.strip("\n"))
    return None


def stage_port_security_sticky_timeout(is_add, interface, timeout=None):
    """
    Stage configuration files for
    `net (add|del) interface <interface> port-security sticky-mac timeout <timeout>`

    Args:
        is_add: True if `net add`, False if `net del`
        interface: The string name of a single interface to configure.
        timeout: The string timeout to configure or `None` when deleting.

    Side effects:
        Writes configuration files to scratchpad.
    """
    dst_path = TMPL_PORT_SECURITY_STICKY_TIMEOUT_FILE.format(interface)
    if is_add:
        new_config = timeout + "\n"
    else:
        new_config = DEFAULT_PORT_SECURITY_STICKY_TIMEOUT
    _stage_switchd_conf_update(dst_path, new_config.strip("\n"))
    return None


def stage_port_security_allowed_mac(is_add, interface, mac=None):
    """
    Stage configuration files for
    `net (add|del) interface <interface> port-security allowed-mac <mac>`
    and
    `net del interface <interface> port-security allowed-mac`

    Args:
        is_add: True if `net add`, False if `net del`
        interface: The string name of a single interface to configure.
        mac: optional - The string MAC address to add/del. If mac is None, then
            `del` deletes all MACs.

    Side effects:
        Writes configuration files to scratchpad.
    """
    dst_path = TMPL_PORT_SECURITY_ALLOWED_MAC_FILE.format(interface)
    scratch_switchd_conf = WORKING_DIRECTORY + SWITCHD_CONF_FILE
    macs = None
    # Check if we already have a conf file we're working on.
    if os.path.isfile(scratch_switchd_conf):
        # Read what we've got so far in the conf file.
        with open(scratch_switchd_conf) as f:
            existing = f.read()
        key = _convert_dst_path_to_switchd_conf_key(dst_path)
        parsed = parse_switchd_conf(existing)
        # Use the parsed macs if applicable.
        macs = parsed.data.get(key, None)
        if macs is not None:
            macs = macs.value.strip("\n").split(" ")
            # Filter out empty strings
            macs = filter(len, macs)

    if macs is None:
        # Default to what's running, not what's in port_security.conf. The macs
        # might have been added elsewhere, and we don't want to lose them.
        macs = get_static_macs().get(interface, [])

    if is_add:
        macs.append(mac)
        new_config = " ".join(macs) + "\n"
    else:
        if mac is not None:
            macs.remove(mac)
            new_config = " ".join(macs) + "\n"
        else:
            new_config = "\n"
    _stage_switchd_conf_update(dst_path, new_config.strip("\n"))
    return None


def stage_port_security_mac_limit(is_add, interface, mac_limit=None):
    """
    Stage configuration files for
    `net add interface <interface> port-security mac-limit <mac_limit>`
    and
    `net del interface <interface> port-security mac-limit`

    Args:
        is_add: True if `net add`, False if `net del`
        interface: The string name of a single interface to configure.
        mac_limit: optional - The string MAC limit to configure or `None` if
            we're deleting the config.

    Side effects:
        Writes configuration files to scratchpad.
    """
    dst_path = TMPL_PORT_SECURITY_MAC_LIMIT_FILE.format(interface)
    if is_add:
        new_config = mac_limit + "\n"
    else:
        new_config = DEFAULT_PORT_SECURITY_MAC_LIMIT
    _stage_switchd_conf_update(dst_path, new_config.strip("\n"))
    return None


def stage_port_security_violation_mode(is_add, interface, violation_mode=None):
    """
    Stage configuration files for
    `net add interface <interface> port-security violation (shutdown|restrict)`

    Args:
        is_add: True if `net add`, False if `net del`
        interface: The string name of a single interface to configure.
        violation_mode: The string violation mode to configure.

    Side effects:
        Writes configuration files to scratchpad.
    """
    dst_path = TMPL_PORT_SECURITY_VIOLATION_MODE_FILE.format(interface)
    if is_add:
        if violation_mode == "shutdown":
            new_config = "0\n"
        elif violation_mode == "restrict":
            new_config = "1\n"
    else:
        new_config = DEFAULT_PORT_SECURITY_VIOLATION_MODE
    _stage_switchd_conf_update(dst_path, new_config.strip("\n"))
    return None


def stage_port_security_violation_timeout(is_add, interface, timeout=None):
    """
    Stage configuration files for
    `net add interface <interface> port-security timeout <timeout>`
    and
    `net del interface <interface> port-security timeout`

    Args:
        is_add: True if `net add`, False if `net del`
        interface: The string name of a single interface to configure.
        timeout: The string timeout to configure or `None` when deleting.

    Side effects:
        Writes configuration files to scratchpad.
    """
    dst_path = TMPL_PORT_SECURITY_VIOLATION_TIMEOUT_FILE.format(interface)
    if is_add:
        new_config = timeout + "\n"
    else:
        new_config = DEFAULT_PORT_SECURITY_VIOLATION_TIMEOUT
    _stage_switchd_conf_update(dst_path, new_config.strip("\n"))
    return None


#################################
# Handlers for show commands
#################################
_summary_keys = [
    "enable",
    "mac_limit",
    "sticky_mac",
    "sticky_aging",
    "sticky_timeout",
    "violation_mode",
    "violation_timeout",
]


_detail_keys = [
    "enable",
    "mac_limit",
    "sticky_mac",
    "sticky_aging",
    "sticky_timeout",
    "violation_mode",
    "violation_timeout",
    "mac_addresses",
]


_keys_to_pretty_headers = {
    "interface": "Interface",
    "enable": "Port security",
    "mac_limit": "MAC limit",
    "sticky_mac": "Sticky MAC",
    "sticky_aging": "Sticky MAC aging",
    "sticky_timeout": "Sticky MAC timeout",
    "violation_mode": "Violation mode",
    "violation_timeout": "Timeout",
    "mac_addresses": "MAC addresses",
}


def _translate_raw_value(key, value):
    """
    Translate the string value from a config file into the proper data type
    based off its key.

    Args:
        key: The string key associated with the value.
        value: The string value from the config file.

    Returns:
        The translated value.
    """
    if key in ["enable", "sticky_mac", "sticky_aging"]:
        value = {
            "0": False,
            "1": True,
        }[value]
    elif key in ["mac_limit", "violation_timeout", "sticky_timeout"]:
        value = int(value)
    elif key in ["violation_mode"]:
        value = MAP_CODE_TO_VIOLATION_MODE[value]
    return value


def _translate_value_to_pretty(key, value, is_enabled):
    """
    Translate a value into its pretty string based off its key.

    Args:
        key: The string key associated with the value.
        value: The value as returned by `_translate_raw_value`
        is_enabled: True if port security is enabled for the interface in
            question, otherwise False

    Returns:
        The pretty string version of the value.
    """
    # EARLY RETURN
    if not is_enabled and key in ["sticky_mac", "mac_limit", "mac_addresses",
                                  "sticky_aging", "sticky_timeout"]:
        # These aren't applicable when it's not enabled
        return "NA"
    elif value is None:
        return None

    if key in ["enable", "sticky_mac", "sticky_aging"]:
        pretty = {
            True: "ENABLED",
            False: "Disabled",
        }[value]
    elif key in ["violation_mode"]:
        pretty = value.capitalize()
    else:
        pretty = str(value)
    return pretty


def _format_table_summary(data):
    """
    Format the summary data as a string table.

    Args:
        data: The summary data.

    Returns:
        A string table to display to the user.
    """
    headers = ["interface"] + _summary_keys
    rows = []
    ordered = sort_for_humans(list(data.keys()))
    for iface in ordered:
        config = data[iface]
        current_row = []
        current_row.append(iface)
        is_enabled = config["enable"]
        for key in _summary_keys:
            value = _translate_value_to_pretty(
                key,
                config.get(key),
                is_enabled
            )
            current_row.append(value)
        rows.append(current_row)

    headers = [
        _keys_to_pretty_headers[key]
        for key in headers
    ]
    return tabulate(rows, headers=headers, numalign="left")


def show_port_security_summary(is_json):
    """
    Generate output for
    `net show port-security [json]`

    Args:
        is_json: True if user requested JSON, otherwise False.

    Returns:
        The string to be rendered to users.
    """
    glob_pattern = TMPL_PORT_SECURITY_DIR.format("*") + "/*"
    paths = glob(glob_pattern)
    result = {}
    for path in paths:
        start_idx = glob_pattern.index("*")
        starts_with_iface = path[start_idx:]
        iface = starts_with_iface.split("/", 1)[0]
        details = result.setdefault(iface, {})

        # Use the last segment of path as the keyword
        keyword = path.rsplit("/", 1)[1]
        if keyword in _summary_keys:
            with open(path) as f:
                contents = f.read()
            value = contents.strip("\n")
            details[keyword] = _translate_raw_value(keyword, value)

    if is_json:
        return dumps(result, indent=2, sort_keys=True)
    else:
        return _format_table_summary(result)


def _format_table_detail(data):
    """
    Format the detailed port security data about a specific interface as a
    table.

    Args:
        data: The detailed data about a specific interface.

    Returns:
        The string table.
    """
    is_enabled = data["enable"]
    order = ["interface"] + _detail_keys
    rows = [
        [
            _keys_to_pretty_headers[key],
            _translate_value_to_pretty(key, data[key], is_enabled)
        ]
        for key in order
        if key != "mac_addresses"  # We'll handle this one separately
    ]
    base_details = tabulate(rows, tablefmt="plain")
    mac_order = DUMP_KEYS
    mac_rows = [
        [
            row[key]
            for key in mac_order
        ]
        for row in data["mac_addresses"]
    ]
    mac_details = tabulate(mac_rows, tablefmt="plain")

    result = [
        base_details,
        "MAC addresses",
    ]
    if len(mac_rows) > 0:
        result.append(mac_details)

    return "\n".join(result)


def show_port_security_interface(interface, is_json):
    """
    Generate output for
    `net show port-security <interface> [json]`

    Args:
        interface: The string interface name.
        is_json: True if user requested JSON, otherwise False.

    Returns:
        The string to be rendered to users.
    """
    base_path = TMPL_PORT_SECURITY_DIR.format(interface)
    result = {
        "interface": interface,
    }
    for key in _detail_keys:
        # EARLY CONTINUE
        if key == "mac_addresses":
            # static_mac doesn't show up in the same FUSE file we write it to.
            # Instead of reading it like we read the others, we need to take a
            # different approach.
            macs = get_port_security_mac_dump().get(interface, [])
            result[key] = macs
            continue
        path = base_path + "/" + key
        with open(path) as f:
            contents = f.read()
        value = contents.strip("\n")
        result[key] = _translate_raw_value(key, value)

    if is_json:
        return dumps(result, indent=2)
    else:
        return _format_table_detail(result)


###############################
# Parse port_security_mac_dump
###############################
DUMP_KEY_MAC_ADDRESS = "MAC Address"
DUMP_KEY_TYPE = "TYPE"
DUMP_KEY_STATUS = "STATUS"

DUMP_KEYS = [
    DUMP_KEY_MAC_ADDRESS,
    DUMP_KEY_TYPE,
    DUMP_KEY_STATUS,
]


REGEX_SECTION_DIVIDER = r"^Interface "
REGEX_TABLE_LINE = r"^=+$"


def parse_port_security_mac_dump(content):
    sections = re.split(REGEX_SECTION_DIVIDER, content, flags=re.MULTILINE)
    split_sections = [
        [
            line
            for line in section.splitlines()
            # Drop the table lines. They're just noise.
            if re.match(REGEX_TABLE_LINE, line) is None
        ]
        for section in sections
        if len(section) != 0
    ]
    result = {}
    for section in split_sections:
        # EARLY CONTINUE
        if len(section) == 2:
            # Length 2 means there are no items
            continue
        # Because of how we split the sections, the first entry is just the
        # interface name.
        iface_name = section[0]
        header_line = section[1]
        headers = [
            col.strip()
            for col in re.split("\t[\t ]*", header_line)
        ]

        entries = []
        for line in section[2:]:
            entry = {}
            columns = [col.strip() for col in line.split("\t")]
            for key, col in zip(headers, columns):
                entry[key] = col.strip()
            entries.append(entry)

        result[iface_name] = sorted(entries)

    return result


def get_port_security_mac_dump():
    # Enable MAC dump
    write_file(PORT_SECURITY_DUMP_FILE, "1\n")

    with open(PORT_SECURITY_MAC_DUMP_FILE) as f:
        dump = f.read()
    return parse_port_security_mac_dump(dump)


def get_static_macs():
    mac_dump = get_port_security_mac_dump()
    # Trim the parsed results to just a list of static macs for each interface
    static_macs = {
        iface: [
            entry[DUMP_KEY_MAC_ADDRESS]
            for entry in entries
            if entry[DUMP_KEY_TYPE] == "Static"
        ]
        for iface, entries in mac_dump.items()
    }
    # Filter out empty arrays
    return {
        iface: macs
        for iface, macs in static_macs.items()
        if macs != []
    }


######################
# Parsing switchd.conf
######################
Line = namedtuple(
    "Line",
    [
        "comment_split",
        "equal_split",
        "key_split",
        "value_split",
    ]
)


Token = namedtuple(
    "Token",
    [
        "token",
        "before",
        "after",
    ],
)


DataValue = namedtuple(
    "DataValue",
    [
        "value",
        "index",
    ]
)


ParsedSwitchdConf = namedtuple(
    "ParsedSwitchdConf",
    [
        "lines",
        "data",
    ],
)


def _parse_token(raw, token):
    # Don't split with empty string separator
    if token != "":
        split = raw.split(token, 1)
    else:
        split = [raw, '']
    return Token(
        before=split[0],
        after=split[1] if len(split) > 1 else None,
        token=token if token in raw else None,
    )


def edit_switchd_conf(contents, overrides):
    """
    Parse a switchd conf file's contents, update some values and return the
    updated contents.

    Args:
        contents: The string contents of the switchd conf file.
        overrides: A dict mapping the key string to the new value to update the
            file with

    Returns:
        The updated string contents.
    """
    # NOTE: mutates parsed.lines in place
    parsed = parse_switchd_conf(contents)
    lines = parsed.lines
    data = parsed.data
    for key, val in overrides.items():
        if key in data:
            idx = data[key].index
            old_line = lines[idx]
            old_value_split = old_line.value_split
            new_value_split = Token(
                token=val,
                before=old_value_split.before,
                after=old_value_split.after,
            )
            new_line = Line(
                comment_split=old_line.comment_split,
                equal_split=old_line.equal_split,
                key_split=old_line.key_split,
                value_split=new_value_split,
            )
            lines[idx] = new_line
        else:
            new_line = Line(
                comment_split=Token(
                    before=None,
                    after=None,
                    token=None,
                ),
                equal_split=Token(
                    before=None,
                    after=None,
                    token=None,
                ),
                key_split=Token(
                    before="",
                    token=key,
                    after=" ",
                ),
                value_split=Token(
                    before=" ",
                    token=val,
                    after="",
                ),
            )
            lines.append(new_line)
    return rebuild_switchd_conf(parsed)


def parse_switchd_conf(contents):
    """
    Parse a switchd conf file.

    Args:
        contents: The string contents of the file.

    Returns:
        A ParsedSwitchdConf namedtuple.
    """
    raw_lines = contents.splitlines()
    lines = []
    data = {}
    for idx, line in enumerate(raw_lines):
        comment_token = "#"
        comment_split = _parse_token(line, comment_token)

        equal_token = "="
        equal_split = _parse_token(comment_split.before, equal_token)

        key_split = None
        value_split = None
        if equal_split.token is not None:
            key_token = equal_split.before.strip()
            value_token = equal_split.after.strip()
            key_split = _parse_token(equal_split.before, key_token)
            value_split = _parse_token(equal_split.after, value_token)

            data[key_token] = DataValue(value_token, idx)

        lines.append(Line(
            comment_split=comment_split,
            equal_split=equal_split,
            key_split=key_split,
            value_split=value_split,
        ))

    return ParsedSwitchdConf(
        lines=lines,
        data=data,
    )


def _reconstruct_line(line):
    """
    Reconstruct a line from the Line namedtuple into a string line for the
    config file.

    Args:
        line: A Line namedtuple

    Returns:
        A string line.
    """
    comment_split = line.comment_split
    key_split = line.key_split
    value_split = line.value_split

    constructed_line = ""
    if key_split is not None:
        constructed_line += (
            key_split.before
            + key_split.token
            + key_split.after
        )

    if value_split is not None:
        constructed_line += (
            "="
            + value_split.before
            + value_split.token
            + value_split.after
        )

    if comment_split.token is not None and constructed_line != "":
        constructed_line += (
            comment_split.token
            + comment_split.after
        )
    elif comment_split.token is not None:
        constructed_line += (
            comment_split.before
            + comment_split.token
            + comment_split.after
        )
    return constructed_line


def rebuild_switchd_conf(parsed):
    """
    Rebuild a switchd conf file from the parsed data stucture.

    Args:
        parsed: The parsed data structure as returned by `parse_switchd_conf`.

    Returns:
        The string contents of the rebuilt file.
    """
    rebuilt_lines = []
    for line in parsed.lines:
        rebuilt_lines.append(_reconstruct_line(line))
    return "\n".join(rebuilt_lines) + "\n"


########################
# Applying switchd.conf
########################
def apply_switchd_conf(contents):
    supported_ports = list_supported_ports()
    parsed = parse_switchd_conf(contents)
    for key, dataValue in parsed.data.items():
        value = dataValue.value
        key_path = key.split(".")
        # Get the interface name
        has_interface = len(key_path) > 2 and key_path[0] == "interface"
        # Skip over interfaces that aren't there
        if has_interface and key_path[1] not in supported_ports:
            continue

        target = os.path.join(SWITCHD_CONFIG_DIR, *key_path)
        parent = os.path.dirname(target)
        try:
            os.makedirs(parent)
        except Exception:
            # Not a problem if it already exists
            pass
        write_file(target, value + "\n")
