"""
Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt

----------------------------------------------------------------------------------------

Usage:

    net (add|del) reflexive-acl-enable

Options:
    reflexive-acl-enable    : enables/disables Reflexive ACL on the switch
"""
import logging
import os
import json
import collections
import itertools

from tabulate import tabulate

from nclu import (
    ifname_expand_glob,
    ifnames_to_glob,
    glob_to_numbers,
    numbers_to_glob,
    WORKING_DIRECTORY,
)

from nclu.plugins import (
    use_exceptions,
    get_config_files_helper,
    get_pending_diffs,
    make_scratchpads,
    persist_configurations,
    remove_scratch_files_and_backups,
    run_commands,
)

from nclu.utils import(
    get_platform_info,
)

from configobj import ConfigObj
from network_docopt import sort_for_humans

log = logging.getLogger("netd")

SWITCHD_CONFIG = "/etc/cumulus/switchd.conf"
SWITCHD_CONFIG_SCRATCHPAD = os.path.join(
    WORKING_DIRECTORY, "reflexive-acl", "switchd.conf"
)

SWITCHD_CONF_RESTART_UTIL = "/usr/bin/systemctl"
SWITCHD_CONF_RESTART_COMMAND = ["/usr/bin/systemctl", "restart", "switchd.service"]


####################################
# EXPECTED NCLU PLUGIN FUNCTIONS
####################################
# Override keyword definitions created by other plugins/wrappers
per_command_helptext = {}



def commit_pending(verbose, net_del_all):
    """
    Copy scratchpad files over configuration files, call necessary restart commands, etc.

    Return a 2-tuple of (bool, dict).  The boolean indicates if the commit succeeded.
    The dictionary is as described in nclu/plugins/__init__.py:execute_modular.

    net_del_all - A boolean; if True, the user is committing "net del all"
    """

    pending_switchd_conf = get_pending_diffs(
        [(SWITCHD_CONFIG, SWITCHD_CONFIG_SCRATCHPAD)],
        False,
    )

    # copy the scratchpad over the config file to prep for running the
    # commands to apply the configuration.
    persist_configurations([(SWITCHD_CONFIG, SWITCHD_CONFIG_SCRATCHPAD)])

    commands = []

    if net_del_all:
        # Reload the port-mirror config on `del all`
        commands = get_restart_commands()
    elif pending_switchd_conf != "":
        # For everything else, reload port-mirror config if pending changes
        commands = get_restart_commands()

    success = True
    if commands:
        success, _, messages = run_commands(commands, verbose)

    if success:
        abort_pending()
        return (True, {})
    else:
        # If the command fails, restore the config file to the backup to keep the changes
        # in 'pending'.
        if os.path.isfile(SWITCHD_CONFIG_SCRATCHPAD + ".bak"):
            persist_configurations([(SWITCHD_CONFIG, SWITCHD_CONFIG_SCRATCHPAD + ".bak")], make_backup=False)
        messages["error"] = ["Unable to apply reflexive ACL configuration.  See switchd.log for details"]
        return (False, messages)


def show_config_commands(_ifupdown2_wrapper):
    """
    net show configuration commands
    """

    command_tree = {}

    # Walk the tree and create the list of commands.
    commands = []


    try:
        iniobj = create_ini_object(SWITCHD_CONFIG)
        if iniobj["rflx.reflexive_acl_enable"] == "TRUE":
            command = "net add reflexive-acl-enable"
            commands.append(command)
    except Exception as e:
        pass


    return (commands, [])


def show_config_summary(summary):
    """
    net show configuration reflexive-acl

    Populate the "summary" dictionary with information about the port_mirror configuration.
    """
    # nope
    pass


@use_exceptions(Exception)
def execute_command(net_daemon):
    """
    Execute an NCLU reflexive-acl command. Assumes that the NCLU core is only sending it
    valid commands.

    (Called by NCLU framework code.)

    Side effects:
        Writes files to NCLU scratchpad.

    Returns:
        A 2-tuple of (bool, dict), where bool represents whether or not the
        command succeeded, and dict contains output/error messages to display
        to user.
    """
    args = net_daemon.nclu_parser.args
    if not os.path.isfile(SWITCHD_CONFIG_SCRATCHPAD):
        make_scratchpads([
            (
                SWITCHD_CONFIG,
                SWITCHD_CONFIG_SCRATCHPAD,
                '',
            ),
        ])

    iniobj = create_ini_object(SWITCHD_CONFIG_SCRATCHPAD)

    if args.get("add"):
        rflx_enable = "rflx.reflexive_acl_enable"
        rflx_enable_info = create_ini_object(SWITCHD_CONFIG).get(rflx_enable)

        if rflx_enable_info is not None:
            if rflx_enable_info == 'TRUE':
                log.info("No change in config required. Reflexive ACL is already enabled")
                os.remove(SWITCHD_CONFIG_SCRATCHPAD)
            elif rflx_enable_info == 'FALSE':
                log.info('Current state of rflx enable is False, enabling it')
                write_section(iniobj)
        else:
            write_section(iniobj)
        
        
    elif args.get("del"):
        rflx_enable = "rflx.reflexive_acl_enable"
        rflx_enable_info = create_ini_object(SWITCHD_CONFIG).get(rflx_enable)

        if rflx_enable_info is not None:
            if rflx_enable_info == 'FALSE':
                log.info("No change in config required. Reflexive ACL is already disabled")
                os.remove(SWITCHD_CONFIG_SCRATCHPAD)
            elif rflx_enable_info == 'TRUE':
                log.info('Current state of rflx enable is TRUE, disabling it')
                write_section(iniobj, enable="FALSE")
        else:
            write_section(iniobj, enable="FALSE")
        
    else:
        log.info("Could not determine which command was executed")


def del_all():
    """
    net del all

    Revert all configuration files associated with this plug-in to their defaults.
    """
    #with open(SWITCHD_CONFIG_SCRATCHPAD, "w") as f:
    #    f.write(PORT_MIRROR_CONFIG_DEFAULT)
    if not os.path.isfile(SWITCHD_CONFIG_SCRATCHPAD):
        make_scratchpads([
            (
                SWITCHD_CONFIG,
                SWITCHD_CONFIG_SCRATCHPAD,
                '',
            ),
        ])

        iniobj = create_ini_object(SWITCHD_CONFIG_SCRATCHPAD)

        rflx_enable = "rflx.reflexive_acl_enable = "

        try:
            rflx_enable_info = create_ini_object(SWITCHD_CONFIG).get(rflx_enable)
            if rflx_enable_info is not None:
                rflx_enable_state = rflx_enable_info.get('rflx.reflexive_acl_enable = FALSE')
                if rflx_enable_state == 'FALSE':
                    log.info("No change in config required. Reflexive ACL is already disabled")
                    os.remove(SWITCHD_CONFIG_SCRATCHPAD)
                elif rflx_enable_state == 'TRUE':
                    log.info('Current state of rflx enable is TRUE, disabling it')
            else:
                write_section(iniobj, enable="FALSE")
        

        except Exception as e:
            pass



def get_config_files():
    """
    Return a string with the content of all existing, non-empty configuration files managed with this plug-in.
    """
    return get_config_files_helper(get_managed_files())


def get_pending(use_colors):
    """ Return a string of the pending changes in all managed files. """
    return get_pending_diffs(
        [
            (SWITCHD_CONFIG, SWITCHD_CONFIG_SCRATCHPAD),
        ],
        use_colors
    )


def abort_pending():
    remove_scratch_files_and_backups([
        SWITCHD_CONFIG_SCRATCHPAD,
    ])


def get_managed_files():
    """ Return a list of managed files. """

    return [SWITCHD_CONFIG]


def get_restart_commands(filename=None):
    """
    Return a list of commands that should be called for the given file after a rollback.
    Some plug-ins use this in commit_pending as well.
    """
    switchd_conf_commands = []

    platform_info = get_platform_info()

    # Get the platform info.  If on a Vx ('software' platform) we can't support any
    # switchdctl commands so return empty command list.
    hw_sw = platform_info.get('platform')
    if hw_sw is None or hw_sw == 'software':
        return switchd_conf_commands

    # otherwise on a hardware platform return the command to reload port-mirror config
    # if switchdctl is present.
    if os.path.isfile(SWITCHD_CONF_RESTART_UTIL):
        switchd_conf_commands.append(SWITCHD_CONF_RESTART_COMMAND)

    return switchd_conf_commands


###########################
# NON-PLUGIN FUNCTIONS
###########################
def create_ini_object(from_file):
    try:
        return(ConfigObj(
            infile=from_file,
            write_empty_values=True,
            list_values=False))
    except Exception as e:
        log.exception(e)
        raise Exception("Unable to parse {0}".format(from_file))


def write_section(iniobj, enable='TRUE'):

    rflx_enable_str = 'rflx.reflexive_acl_enable'

    iniobj[rflx_enable_str] = enable
    iniobj.write()


