"""
Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt

----------------------------------------------------------------------------------------

Usage:

    net add reflexive-acl [unreplied-ingress-policer-rate <1000-5000>] [trap-group-policer-rate <100-1000>] [age-poll-interval <1-15>]
    net del reflexive-acl [unreplied-ingress-policer-rate | trap-group-policer-rate | age-poll-interval]

    net show reflexive-acl status [json]

Options:
    reflexive-acl                   : keyword for Reflexive acl
    unreplied-ingress-policer-rate  : Rate in Packets per second(PPS) for unreplied ingress flow
    trap-group-policer-rate         : Rate in Packets per second(PPS) for trapped packets
    age-poll-interval               : Timeout interval in mins for inactvity (applicable for UDP/ICMP Flows)
"""
import logging
import os
import json
import collections
import itertools

from tabulate import tabulate
from json import dumps

from nclu import (
    ifname_expand_glob,
    ifnames_to_glob,
    glob_to_numbers,
    numbers_to_glob,
    WORKING_DIRECTORY,
)

from nclu.plugins import (
    use_exceptions,
    get_config_files_helper,
    get_pending_diffs,
    make_scratchpads,
    persist_configurations,
    remove_scratch_files_and_backups,
    run_commands,
)

from nclu.utils import(
    get_platform_info,
)

from configobj import ConfigObj
from network_docopt import sort_for_humans

log = logging.getLogger("netd")

RFLX_ACL_UTIL_CONFIG = "/etc/cumulus/switchd.d/reflex_acl.conf"
RFLX_ACL_UTIL_CONFIG_SCRATCHPAD = os.path.join(
    WORKING_DIRECTORY, "reflexive-acl", "reflex_acl.conf"
)

RFLX_ACL_UTIL_RESTART_UTIL = "/usr/lib/cumulus/switchdctl"
RFLX_ACL_UTIL_RESTART_COMMAND = [RFLX_ACL_UTIL_RESTART_UTIL, "--load", RFLX_ACL_UTIL_CONFIG]

RFLX_ACL_DEF_UNREP_ING_POL_RATE = 1000
RFLX_ACL_DEF_TRPGRP_POL_RATE    = 500
RFLX_ACL_DEF_AGE_POLL_INTVL     = 2


####################################
# EXPECTED NCLU PLUGIN FUNCTIONS
####################################
# Override keyword definitions created by other plugins/wrappers
per_command_helptext = {}



def commit_pending(verbose, net_del_all):
    """
    Copy scratchpad files over configuration files, call necessary restart commands, etc.

    Return a 2-tuple of (bool, dict).  The boolean indicates if the commit succeeded.
    The dictionary is as described in nclu/plugins/__init__.py:execute_modular.

    net_del_all - A boolean; if True, the user is committing "net del all"
    """

    pending_switchd_conf = get_pending_diffs(
        [(RFLX_ACL_UTIL_CONFIG, RFLX_ACL_UTIL_CONFIG_SCRATCHPAD)],
        False,
    )

    # copy the scratchpad over the config file to prep for running the
    # commands to apply the configuration.
    persist_configurations([(RFLX_ACL_UTIL_CONFIG, RFLX_ACL_UTIL_CONFIG_SCRATCHPAD)])

    commands = []

    if net_del_all:
        # Reload the port-mirror config on `del all`
        commands = get_restart_commands()
    elif pending_switchd_conf != "":
        # For everything else, reload port-mirror config if pending changes
        commands = get_restart_commands()

    success = True
    if commands:
        success, _, messages = run_commands(commands, verbose)

    if success:
        abort_pending()
        return (True, {})
    else:
        # If the command fails, restore the config file to the backup to keep the changes
        # in 'pending'.
        if os.path.isfile(RFLX_ACL_UTIL_CONFIG_SCRATCHPAD + ".bak"):
            persist_configurations([(RFLX_ACL_UTIL_CONFIG, RFLX_ACL_UTIL_CONFIG_SCRATCHPAD + ".bak")], make_backup=False)
        messages["error"] = ["Unable to apply reflexive ACL Util configuration.  See switchd.log for details"]
        return (False, messages)


def show_config_commands(_ifupdown2_wrapper):
    """
    net show configuration commands
    """

    command_tree = {}

    # Walk the tree and create the list of commands.
    commands = []

    iniobj = create_ini_object(RFLX_ACL_UTIL_CONFIG)

    try:
        unrep_ing_pol_rate = iniobj["rflx.unreplied_ing_pol_rate"]
        if unrep_ing_pol_rate != str(RFLX_ACL_DEF_UNREP_ING_POL_RATE):
            command = "net add reflexive-acl unreplied-ingress-policer-rate {0}".format(unrep_ing_pol_rate)
            commands.append(command)
    except Exception as e:
        pass

    try:
        trpgrp_pol_rate = iniobj["rflx.trapgroup_pol_rate"]
        if trpgrp_pol_rate != str(RFLX_ACL_DEF_TRPGRP_POL_RATE):
            command = "net add reflexive-acl trap-group-policer-rate {0}".format(trpgrp_pol_rate)
            commands.append(command)
    except Exception as e:
        pass

    try:
        age_poll_intvl = iniobj["rflx.age_poll_interval"]
        if age_poll_intvl != str(RFLX_ACL_DEF_AGE_POLL_INTVL):
            command = "net add reflexive-acl age-poll-interval {0}".format(age_poll_intvl)
            commands.append(command)
    except Exception as e:
        pass


    return (commands, [])


def show_config_summary(summary):
    """
    net show configuration reflexive-acl

    Populate the "summary" dictionary with information about the port_mirror configuration.
    """
    # nope
    pass


@use_exceptions(Exception)
def execute_command(net_daemon):
    """
    Execute an NCLU reflexive-acl command. Assumes that the NCLU core is only sending it
    valid commands.

    (Called by NCLU framework code.)

    Side effects:
        Writes files to NCLU scratchpad.

    Returns:
        A 2-tuple of (bool, dict), where bool represents whether or not the
        command succeeded, and dict contains output/error messages to display
        to user.
    """
    args = net_daemon.nclu_parser.args
    if not os.path.isfile(RFLX_ACL_UTIL_CONFIG_SCRATCHPAD):
        make_scratchpads([
            (
                RFLX_ACL_UTIL_CONFIG,
                RFLX_ACL_UTIL_CONFIG_SCRATCHPAD,
                '',
            ),
        ])

    iniobj = create_ini_object(RFLX_ACL_UTIL_CONFIG_SCRATCHPAD)

    do_unrep_ing_pol_cfg = False
    unrep_ing_pol_cfg = 0
    do_trpgrp_pol_cfg = False
    trpgrp_pol_cfg = 0
    do_age_poll_cfg = False
    age_poll_cfg = 0

    if args.get("add") or args.get("del"):
        if args.get("unreplied-ingress-policer-rate"):
            unrep_ing_pol_str = "rflx.unreplied_ing_pol_rate"
            unrep_ing_pol_info = create_ini_object(RFLX_ACL_UTIL_CONFIG).get(unrep_ing_pol_str)
            if args.get("add"):
                in_unrep_ing_pol_rate = args.get("<1000-5000>")
                in_unrep_ing_pol_val = int(in_unrep_ing_pol_rate)
            else:
                in_unrep_ing_pol_val = RFLX_ACL_DEF_UNREP_ING_POL_RATE
            if unrep_ing_pol_info is not None:
                unrep_ing_pol_val = int(unrep_ing_pol_info)
                if unrep_ing_pol_val != in_unrep_ing_pol_val:
                    do_unrep_ing_pol_cfg = True
                    unrep_ing_pol_cfg = in_unrep_ing_pol_val
                else:
                    do_unrep_ing_pol_cfg = False
            else:
                do_unrep_ing_pol_cfg = True
                unrep_ing_pol_cfg = in_unrep_ing_pol_val

        if args.get("trap-group-policer-rate"):
            trpgrp_pol_str = "rflx.trapgroup_pol_rate"
            trpgrp_pol_info = create_ini_object(RFLX_ACL_UTIL_CONFIG).get(trpgrp_pol_str)
            if args.get("add"):
                in_trpgrp_pol_rate = args.get("<100-1000>")
                in_trpgrp_pol_val = int(in_trpgrp_pol_rate)
            else:
                in_trpgrp_pol_val = RFLX_ACL_DEF_TRPGRP_POL_RATE
            if trpgrp_pol_info is not None:
                trpgrp_pol_val = int(trpgrp_pol_info)
                if trpgrp_pol_val != in_trpgrp_pol_val:
                    do_trpgrp_pol_cfg = True
                    trpgrp_pol_cfg = in_trpgrp_pol_val
                else:
                    do_trpgrp_pol_cfg = False
            else:
                do_trpgrp_pol_cfg = True
                trpgrp_pol_cfg = in_trpgrp_pol_val

        if args.get("age-poll-interval"):
            age_poll_str = "rflx.age_poll_interval"
            age_poll_info = create_ini_object(RFLX_ACL_UTIL_CONFIG).get(age_poll_str)
            if args.get("add"):
                in_age_poll = args.get("<1-15>")
                in_age_poll_val = int(in_age_poll)
            else:
                in_age_poll_val = RFLX_ACL_DEF_AGE_POLL_INTVL
            if age_poll_info is not None:
                age_poll_val = int(age_poll_info)
                if age_poll_val != in_age_poll_val:
                    do_age_poll_cfg = True
                    age_poll_cfg = in_age_poll_val
                else:
                    do_age_poll_cfg = False
            else:
                do_age_poll_cfg = True
                age_poll_cfg = in_age_poll_val

    
        write_section(iniobj, do_unrep_ing_pol_cfg=do_unrep_ing_pol_cfg, unrep_ing_pol_rate=unrep_ing_pol_cfg, do_trapgroup_pol_cfg=do_trpgrp_pol_cfg, trapgroup_pol_rate=trpgrp_pol_cfg, do_age_poll_cfg=do_age_poll_cfg, age_poll_intvl=age_poll_cfg)
        
    elif args.get('show'):
        if args.get('status'):

            is_json = False
            if args.get('json'):
                is_json = True

            switchd_base_path = "/cumulus/switchd/"
            switchd_run_path = switchd_base_path + "run/"
            switchd_cfg_path = switchd_base_path + "config/"
            rflx_run_path = switchd_run_path + "rflx_info/"
            rflx_cfg_path = switchd_cfg_path + "rflx/"


            rflx_info_list = [
                {
                    "path":rflx_cfg_path+"reflexive_acl_enable",
                    "description" : "Reflexive ACL Enable",
                    "j_key" : "rflx_acl_enable",
                },
                {
                    "path":rflx_cfg_path+"table_size",
                    "description" : "Max number of rules",
                    "j_key" : "rflx_acl_tbl_sz",
                },
                {
                    "path":rflx_cfg_path+"unreplied_ing_pol_rate",
                    "description" : "Unreplied Ingress Flow Policer rate(PPS)",
                    "j_key" : "rflx_unrep_in_pol_rate",
                },
                {
                    "path":rflx_cfg_path+"trapgroup_pol_rate",
                    "description" : "Trap Policer rate(PPS)",
                    "j_key" : "rflx_trap_pol_rate",
                },
                {
                    "path":rflx_cfg_path+"age_poll_interval",
                    "description" : "Aging timer for inactivity(minutes)",
                    "j_key" : "rflx_trap_pol_rate",
                },
                {
                    "path":rflx_run_path+"reflexive/irflx_count",
                    "description" : "Number of Ingress Entries",
                    "j_key" : "rflx_ing_num_entries",
                },
                {
                    "path":rflx_run_path+"reflexive/irflx_max",
                    "description" : "Max number of Ingress Entries",
                    "j_key" : "rflx_max_ing_num_entries",
                },
                {
                    "path":rflx_run_path+"reflexive/erflx_count",
                    "description" : "Number of Egress Entries",
                    "j_key" : "rflx_eg_num_entries",
                },
                {
                    "path":rflx_run_path+"reflexive/erflx_max",
                    "description" : "Max number of Egress Entries",
                    "j_key" : "rflx_max_eg_num_entries",
                },
                {
                    "path":rflx_run_path+"config_rule/rflx_count",
                    "description" : "Number of interfaces configured",
                    "j_key" : "rflx_num_intf_cfg",
                },
                {
                    "path":rflx_run_path+"config_rule/rflx_max",
                    "description" : "Max number of interfaces supported",
                    "j_key" : "rflx_max_num_intf_cfg_supp",
                },
            ]

            rows = []
            json_result = {}
            for rflx_itr in rflx_info_list:
                rflx_file = rflx_itr["path"]

                if os.path.isfile(rflx_file):
                    with open(rflx_file) as f:
                        contents = f.read()
                        contents = contents.strip("\n")

                        if True == is_json:
                            json_result[rflx_itr["j_key"]] = contents
                        else:
                            rows.append([rflx_itr["description"], contents])

            if True == is_json:
                result = dumps(json_result, indent=2)
            else:
                result = tabulate(rows,tablefmt="simple")

            return result
        else:
            log.info('Unsupported show command')
    else:
        log.info("Could not determine which command was executed")


RFLX_ACL_UTIL_CONFIG_DEFAULT = "rflx.unreplied_ing_pol_rate = {0}\n".format(RFLX_ACL_DEF_UNREP_ING_POL_RATE)
RFLX_ACL_UTIL_CONFIG_DEFAULT += "rflx.trapgroup_pol_rate = {0}\n".format(RFLX_ACL_DEF_TRPGRP_POL_RATE)
RFLX_ACL_UTIL_CONFIG_DEFAULT += "rflx.age_poll_interval = {0}\n".format(RFLX_ACL_DEF_AGE_POLL_INTVL)

def del_all():
    """
    net del all

    Revert all configuration files associated with this plug-in to their defaults.
    """
    with open(RFLX_ACL_UTIL_CONFIG_SCRATCHPAD, "w") as f:
        f.write(RFLX_ACL_UTIL_CONFIG_DEFAULT)


def get_config_files():
    """
    Return a string with the content of all existing, non-empty configuration files managed with this plug-in.
    """
    return get_config_files_helper(get_managed_files())


def get_pending(use_colors):
    """ Return a string of the pending changes in all managed files. """
    return get_pending_diffs(
        [
            (RFLX_ACL_UTIL_CONFIG, RFLX_ACL_UTIL_CONFIG_SCRATCHPAD),
        ],
        use_colors
    )


def abort_pending():
    remove_scratch_files_and_backups([
        RFLX_ACL_UTIL_CONFIG_SCRATCHPAD,
    ])


def get_managed_files():
    """ Return a list of managed files. """

    return [RFLX_ACL_UTIL_CONFIG]


def get_restart_commands(filename=None):
    """
    Return a list of commands that should be called for the given file after a rollback.
    Some plug-ins use this in commit_pending as well.
    """
    switchd_conf_commands = []

    platform_info = get_platform_info()

    # Get the platform info.  If on a Vx ('software' platform) we can't support any
    # switchdctl commands so return empty command list.
    hw_sw = platform_info.get('platform')
    if hw_sw is None or hw_sw == 'software':
        return switchd_conf_commands

    # otherwise on a hardware platform return the command to reload port-mirror config
    # if switchdctl is present.
    if os.path.isfile(RFLX_ACL_UTIL_RESTART_UTIL):
        switchd_conf_commands.append(RFLX_ACL_UTIL_RESTART_COMMAND)

    return switchd_conf_commands


###########################
# NON-PLUGIN FUNCTIONS
###########################
def create_ini_object(from_file):
    try:
        return(ConfigObj(
            infile=from_file,
            write_empty_values=True,
            list_values=False))
    except Exception as e:
        log.exception(e)
        raise Exception("Unable to parse {0}".format(from_file))


def write_section(iniobj, do_unrep_ing_pol_cfg=False, unrep_ing_pol_rate=1000, do_trapgroup_pol_cfg=False, trapgroup_pol_rate=100, do_age_poll_cfg=False, age_poll_intvl=1):


    if do_unrep_ing_pol_cfg == True:
        iniobj['rflx.unreplied_ing_pol_rate'] = unrep_ing_pol_rate

    if do_trapgroup_pol_cfg == True:
        iniobj['rflx.trapgroup_pol_rate'] = trapgroup_pol_rate

    if do_age_poll_cfg == True:
        iniobj['rflx.age_poll_interval'] = age_poll_intvl

    if do_unrep_ing_pol_cfg == True or do_trapgroup_pol_cfg == True or do_age_poll_cfg == True:
        iniobj.write()


