"""
Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
Copyright (C) 2020-2021 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt

----------------------------------------------------------------------------------------
"""

import logging
import os
from nclu.NetDaemon import NetDaemon
from nclu import (
    WORKING_DIRECTORY,
    platform_vendor,
)
from nclu.utils import check_output
from nclu.plugins import (
    use_exceptions,
)
from subprocess import CalledProcessError


log = logging.getLogger("netd")

MLNX_DOCSTRING = """
Usage:

    net system maintenance mode (enable|disable)
    net system maintenance upgrade (all|dry-run)
    net system maintenance show status
    net system maintenance restart (cold|fast|warm)
    net system maintenance ports (down|up)

Options:
    system       : System
    maintenance  : System Maintenance
    mode         : System Maintenance mode
    enable       : Enable system maintenance mode
    disable      : Disable system maintenance mode
    upgrade      : System Maintenance Upgrade
    all          : System Maintenance full upgrade
    dry-run      : System Maintenance Upgrade dry-run
    show         : System Maintenance show
    status       : Show system Maintenance status
    restart      : Issue system restart
    cold         : Issue a full system restart
    fast         : Issue a fast system restart
    warm         : Issue a warm system restart
    ports        : Admin (down|up) physical ports for System Maintenance
    up           : Admin up physical ports for System Maintenance
    down         : Admin down physical ports for System Maintenance
"""

def build_doc_string():
    global __doc__

    # System Management commands are supported on Mellanox platforms only.
    if platform_vendor() in ("mlnx", "cumulus"):
        __doc__ += MLNX_DOCSTRING

    return __doc__



def commit_pending(verbose, net_del_all):
    """
    Copy scratchpad files over configuration files, call necessary restart commands, etc.

    Return a 2-tuple of (bool, dict).  The boolean indicates if the commit succeeded.
    The dictionary is as described in nclu/plugins/__init__.py:execute_modular.

    net_del_all - A boolean; if True, the user is committing "net del all"
    """
    return (True, {})


def del_all():
    """
    net del allow

    Revert all configuration files associated with this plug-in to their defaults.
    """
    pass


def show_config_summary(summary):
    """
    net show configuration port-mirror

    Populate the "summary" dictionary with information about the port_mirror configuration.
    """
    # nope
    pass


def show_config_commands(_ifupdown2_wrapper):
    return ([], [])


def run_sysmgmt_command(sysmgmt_cmd):
    try:
        output = check_output(sysmgmt_cmd)

    except CalledProcessError as e:
        log.exception(e)
        raise Exception("\"{0}\" failed with \"{1}\".".format(''.join(sysmgmt_cmd), e.output.strip()))

    except OSError as e:
        log.exception(e)
        raise Exception("\"{0}\" failed with errno= {1}: {2}.".format(''.join(sysmgmt_cmd), e.args[0], e.args[1]))

    except Exception as e:
        log.exception(e)
        raise Exception("\"{0}\" failed.".format(''.join(sysmgmt_cmd)))

    return output


@use_exceptions(Exception)
def execute_command(net_daemon):
    """
    Execute an NCLU system management commands and return a standard 2-tuple

    (Called by NCLU framework code.)

    Side effects:
        Writes files to NCLU scratchpad.

    Returns:
        A 2-tuple of (bool, dict), where bool represents whether or not the
        command succeeded, and dict contains output/error messages to display
        to user.
    """

    assert isinstance(net_daemon, NetDaemon)

    # Convenience variables
    tokens = net_daemon.nclu_parser.args
    assert "system" in tokens and "maintenance" in tokens, "Why are you here?"

    cmd = ['/usr/sbin/csmgrctl']

    if "show" in tokens:
        cmd.append('-s')

    elif "mode" in tokens:
        cmd.append('-m1') if "enable" in tokens else cmd.append('-m0')

    elif "upgrade" in tokens:
        # The underlying command executes an apt-get that must run in the mgmt
        # vrf to reach the repos.  Force this command to run in the mgmt vrf
        cmd = ['ip', 'vrf', 'exec', 'mgmt', cmd[0]]
        cmd.append('-u') if 'all' in tokens else cmd.append('-d')

    elif "restart" in tokens:
        if 'cold' in tokens:
            cmd.append('-cf')
        elif 'warm' in tokens:
            cmd.append('-wf')
        else:
            cmd.append('-ff')

    elif "ports" in tokens:
        cmd.append('-p0') if 'down' in tokens else cmd.append('-p1')

    return run_sysmgmt_command(cmd)


def get_config_files():
    """
    Return a string with the content of all existing, non-empty configuration files managed
    with this plug-in
    """
    return ''


def get_pending(use_colors):
    """
    Return a string of the pending changes in all managed files.
    """
    return ''


def abort_pending():
    pass


def get_managed_files():
    """
    Return a list of managed files.
    """
    return []


def get_restart_commands(filename=None):
    """
    Return a list of commands that should be called for the given file
    after a rollback.
    """
    return []
