# --------------------------------------> IMPORTANT <-------------------------------------------
# PRETTY PLEASE, WITH SUGAR ON TOP, COMMENT YOUR CODE.  PEOPLE SHOULD NOT HAVE TO READ CODE LINE
# BY LINE TO KNOW WHAT IT DOES.  ANY FUNCTION OF MORE THAN A COUPLE OF LINES SHOULD HAVE A
# DOCSTRING EXPLAINING THE INPUTS AND RETURN VALUE.  THIS SHOULD INCLUDE EXPLICIT MENTION OF THE
# EXPECTED TYPE(S).
#
# FULLY SPELL OUT WORDS.  USE PROPER ENGLISH GRAMMAR, PUNCTUATION, AND CAPITALIZATION.  USE
# DESCRIPTIVE, UNABRIDGED VARIABLE NAMES.
#
# DO NOT USE C-STYLE STRING FORMATTING, WHICH IS DEPRECATED.  DO NOT TREAT None LIKE A BOOLEAN.
# COMPLY WITH THE SYTLE GUIDE: https://www.python.org/dev/peps/pep-0008/.
#
# DO NOT USE CLASSES; THERE IS NO STATE.  FUNCTIONS SHOULD TAKE INPUTS AND PRODUCE OUTPUTS WITH
# NO SIDE EFFECTS OTHER THAN LOGGING.  DO NOT PRINT.
#
# CODE THAT DOES NOT FOLLOW THIS GUIDANCE SHALL BE CONSIDERED UNMAINTAINABLE, AT BEST, AND
# THEREFORE NOT APPROPRIATE FOR RELEASE.
# -----------------------------------------------------------------------------------------------
"""
Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
Copyright (C) 2018 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt

----------------------------------------------------------------------------------------

Usage:
    # New trap names needs to be appended to the list here
    net add trap (l3-mtu-err|<text>) action (on|off)

Options:
    trap          : configures the trap
    l3-mtu-err    : packet exceeded router interface mtu size
    action        : determines the action for the configured trap
    on            : sets the trap action to trap_to_cpu
    off           : sets the trap action to discard or drop
"""

import logging
import os
import json
import collections
import itertools

from tabulate import tabulate

from nclu import (
    ifname_expand_glob,
    ifnames_to_glob,
    glob_to_numbers,
    numbers_to_glob,
    WORKING_DIRECTORY,
)

from nclu.plugins import (
    use_exceptions,
    get_config_files_helper,
    get_pending_diffs,
    make_scratchpads,
    persist_configurations,
    remove_scratch_files_and_backups,
    run_commands,
)

from nclu.utils import(
    get_platform_info,
)

from configobj import ConfigObj
from network_docopt import sort_for_humans

log = logging.getLogger("netd")

TRAP_CONFIG_FILE = "/etc/cumulus/switchd.d/trap.conf"
TRAP_CONFIG_SCRATCHPAD = os.path.join(
        WORKING_DIRECTORY, "trap", "trap.conf"
)

# trap config gets applied through this switchdctl command
TRAP_CONFIG_COMMAND = "/usr/lib/cumulus/switchdctl"

# this is the full trap config reload command
TRAP_CONFIG_RELOAD = [TRAP_CONFIG_COMMAND, "--load", TRAP_CONFIG_FILE,
                      "-prefix", "trap"]

TRAP_CONFIG_DEFAULT = """
# Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt
#
# trap.trap_name.enable = (true|false)
"""

####################################
# EXPECTED NCLU PLUGIN FUNCTIONS
####################################
# Override keyword definitions created by other plugins/wrappers
per_command_helptext = {}

###########################
# NON-PLUGIN FUNCTIONS
###########################
def create_ini_object(from_file):
    try:
        return(ConfigObj(
            infile=from_file,
            write_empty_values=True,
            list_values=False))
    except Exception as e: 
        log.exception(e)
        raise Exception("Unable to parse {0}".format(from_file))

def get_restart_commands(filename=None):
    """
    Return a list of commands that should be called for the given file after a rollback.
    Some plug-ins use this in commit_pending as well.
    """
    trap_commands = [] 

    platform_info = get_platform_info()

    # Get the platform info.  If on a Vx ('software' platform) we can't support any
    # switchdctl commands so return empty command list.
    hw_sw = platform_info.get('platform')
    if hw_sw is None or hw_sw == 'software':
        return trap_commands

    # otherwise on a hardware platform return the command to reload trap config
    # if switchdctl is present.
    if os.path.isfile(TRAP_CONFIG_COMMAND):
        trap_commands.append(TRAP_CONFIG_RELOAD)

    return trap_commands


def commit_pending(verbose, net_del_all):
    """
    Copy scratchpad files over configuration files, call necessary restart commands, etc.

    Return a 2-tuple of (bool, dict).  The boolean indicates if the commit succeeded.
    The dictionary is as described in nclu/plugins/__init__.py:execute_modular.

    net_del_all - A boolean; if True, the user is committing "net del all"
    """
    pending_trap_config = get_pending_diffs(
                            [(TRAP_CONFIG_FILE, TRAP_CONFIG_SCRATCHPAD)],
                            False,
                            )
    persist_configurations([(TRAP_CONFIG_FILE, TRAP_CONFIG_SCRATCHPAD)])
    commands = []

    if net_del_all:
        # Reload the trap config on `del all`
        commands = get_restart_commands()
    elif pending_trap_config != "":
        # Reload on any pending changes #
        commands = get_restart_commands()

    success = True
    if commands:
        success, _, messages = run_commands(commands, verbose)

    if success:
        abort_pending()
        return (True, {})
    else:
        # If the command fails, restore the config file to the backup to keep the changes
        # in pending
        if os.path.isfile(TRAP_CONFIG_SCRATCHPAD + ".bak"):
            persist_configurations([(TRAP_CONFIG_FILE, TRAP_CONFIG_SCRATCHPAD + ".bak")], make_backup=False)
        messages["error"] = ["Unable to apply trap configuration.  See switchd.log for details"]
        return (False, messages)
    

def show_config_summary(summary):
    """
    net show configuration trap
    """
    # nope
    pass

def show_config_commands(ifupdown2_wrapper):
    """
    net show configuration commands
    """
    command_tree = {}
    
    # Walk the tree and create the list of commands.
    commands = []
    iniobj = create_ini_object(TRAP_CONFIG_FILE)

    # walk through the ini file and construct a 'net add' command for each section
    for section in iniobj.sections:
        args = dict(iniobj[section].items())
        # form every command from the session sections in the ini file.
        try:
            trap_name = args['Name']
            val = args['trap.{0}.enable'.format(trap_name)]
            if val == 'True':
                trap_action = 'on'
            else:
                trap_action = 'off'
            command = "net add trap {name} action {action}".format(
                              name=trap_name , action = trap_action)
        except Exception as e:
            # if something goes wrong finding all the keys, etc. for this command, skip it.  This
            # could happen if someone mucks with the config file for example and deletes something
            # important.
            log.exception(e)
            log.warning("Missing keys in {0}".format(TRAP_CONFIG_FILE))
            continue

     
        commands.append(command) 
    return (commands, [])

# When the command is executed the following happens in this function:
#   - create the ini object with the scratchpad file
#   - parse the tokens and match the trap names
#   - populate string with the appropriate config
#   - write the config to the scratchpad file - /run/nclu/trap/trap.conf
def execute_command(net_daemon):
    """
    Given an instance of NetDaemon that has already run an input command through NetworkDocopt,
    perform the appropriate actions, and return the standard 2-tuple.
    """
    
    # Convenience variable
    tokens = net_daemon.nclu_parser.args
    argv_expanded = net_daemon.nclu_parser.argv_expanded
    assert "configuration" not in tokens, "\"net show \" does not use this code path."

    mode = next((
             mode_token
             for mode_token in ["add", "del", "show"]
             if mode_token in tokens
    ))
   
    # if the config file isn't present open one and write default header 
    if not os.path.isfile(TRAP_CONFIG_FILE):
        with open(TRAP_CONFIG_FILE, "w") as f:
            f.write(TRAP_CONFIG_DEFAULT)

    # Create the relevant scratchpads (if necessary) and track which ones were
    # created for this command. If the command creates scratchfiles, but
    # doesn't actually change anything, we'll clean up the scratchpad at the
    # end

    did_create_trap_config_scratchpad = False
    if not os.path.isfile(TRAP_CONFIG_SCRATCHPAD):
       did_create_trap_config_scratchpad = True
       make_scratchpads([
           (
                TRAP_CONFIG_FILE,
                TRAP_CONFIG_SCRATCHPAD,
                TRAP_CONFIG_DEFAULT,
           ),
       ])
    # Retrieve the trap name. If new traps are added an or condition to check for the trap_name should
    # be implemented here.
    trap_name = tokens.get("l3-mtu-err")
    config = tokens.get("on") or tokens.get("off")
    if config == 'on':
        enable_val = 'True'
    else:
        enable_val = 'False'

    # Create an object of the scratchpad to update the trap config
    iniobj = create_ini_object(TRAP_CONFIG_SCRATCHPAD)
    info = 'Trap = {0}'.format(trap_name)
    trap_config = {
              'trap.{0}.enable'.format(trap_name) : enable_val,
              'Name' : trap_name,
    }
    iniobj[info] = trap_config

    iniobj.write()
    return (True, {})

def del_all():
    """
    net del all

    Revert all configuration files associated with this plug-in to their defaults.
    """
    with open(TRAP_CONFIG_SCRATCHPAD, "w") as f:
        f.write(TRAP_CONFIG_DEFAULT)

def get_config_files():
    """
    Return a string with the content of all existing, non-empty configuration files managed with this plug-in.
    """
    return ''

def get_pending(use_colors):
    """ Return a string of the pending changes in all managed files. """
    return get_pending_diffs(
            [
                (TRAP_CONFIG_FILE, TRAP_CONFIG_SCRATCHPAD),
            ],
            use_colors
    )

def abort_pending():
    remove_scratch_files_and_backups([
         TRAP_CONFIG_SCRATCHPAD,
    ])


def get_managed_files():
    """ Return a list of managed files. """

    return []
