"""
Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
Copyright (C) 2020 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt

----------------------------------------------------------------------------------------
"""
import logging
import os
import re

from nclu import (
    WORKING_DIRECTORY,
    file_set_key_value,
    file_get_key_value,
    service_is_active,
    service_is_enabled,
)

from nclu.utils import(
    check_output,
    get_platform_info,
)

from nclu.plugins import (
    use_exceptions,
    get_pending_diffs,
    make_scratchpads,
    persist_configurations,
    remove_scratch_files_and_backups,
    run_commands,
)
from subprocess import CalledProcessError

log = logging.getLogger("netd")

WFI_KERNEL_ROUTE_OFFLOAD_FLAGS_KEY = "kernel_route_offload_flags"
WFI_KERNEL_ROUTE_OFFLOAD_FLAG_SET = 2
WFI_KERNEL_ROUTE_OFFLOAD_FLAG_CLEAR = 0
WFI_KERNEL_CONFIG = "/etc/cumulus/switchd.d/kernel_route_offload_flags.conf"
WFI_KERNEL_SCRATCHPAD = os.path.join(
    WORKING_DIRECTORY, "wfi", "kernel_route_offload_flags.conf")
WFI_KERNEL_CONFIG_DEFAULT = """
# Set routing-forwarding-sync mode for routes.
#  0: No notification on HW install success or failure (default mode)
#  1: Notify HW install failure
#  2: Notify HW install success/failure
{0} = 0
""".format(WFI_KERNEL_ROUTE_OFFLOAD_FLAGS_KEY)

# This is the persistent setting in the file system.
#
# This is a bit odd for set and clear (net add and net del) to set this flag
# to the same value... but the defaults changed and this is correct.  If a
# user modified the config manually we want NCLU to bring it back to the
# default setting of 0 for any net add or net del command.
MODPROBE_ROUTE_OFFLOAD_FLAGS_KEY = "options sx_netdev route_offload_flags"
MODPROBE_ROUTE_OFFLOAD_FLAG_SET = 0
MODPROBE_ROUTE_OFFLOAD_FLAG_CLEAR = 0
MODPROBE_CONFIG = "/etc/modprobe.d/sx_netdev.route_offload_flags.conf"
MODPROBE_SCRATCHPAD = os.path.join(
    WORKING_DIRECTORY, "modprobe", "sx_netdev.route_offload_flags.conf")
MODPROBE_CONFIG_DEFAULT =  "{0}={1}\n".format(
    MODPROBE_ROUTE_OFFLOAD_FLAGS_KEY,
    MODPROBE_ROUTE_OFFLOAD_FLAG_CLEAR
)

# and this is the non-persistent kernel setting.  Same caveat applies to this
# flag as noted above about set and clear being the same value.
SX_NETDEV_CONFIG = "/sys/module/sx_netdev/parameters/route_offload_flags"
SX_NETDEV_CONFIG_SET = "0"
SX_NETDEV_CONFIG_CLEAR = "0"

SWITCHD_SERVICE = "switchd.service"
FRR_SERVICE = "frr.service"


def write_sx_netdev_parm(set_clear):

    # don't create this file if it does not exist.
    # EARLY RETURN
    if not os.path.exists(SX_NETDEV_CONFIG):
        return

    with open(SX_NETDEV_CONFIG, "w") as f:
        contents = (SX_NETDEV_CONFIG_SET if set_clear == "set" else SX_NETDEV_CONFIG_CLEAR) + "\n"
        f.write(contents)


def commit_pending(verbose, net_del_all):
    """
    Copy scratchpad files over configuration files, call necessary restart commands, etc.

    Return a 2-tuple of (bool, dict).  The boolean indicates if the commit succeeded.
    The dictionary is as described in nclu/plugins/__init__.py:execute_modular.

    net_del_all - A boolean; if True, the user is committing "net del all"
    """
    pending_diffs = get_pending_diffs(
        [
            (WFI_KERNEL_CONFIG, WFI_KERNEL_SCRATCHPAD),
            (MODPROBE_CONFIG, MODPROBE_SCRATCHPAD),
        ],
        False
    )

    persist_configurations(
        [
            (WFI_KERNEL_CONFIG, WFI_KERNEL_SCRATCHPAD),
            (MODPROBE_CONFIG, MODPROBE_SCRATCHPAD),
        ]
    )

    commands = []
    if net_del_all or pending_diffs != "":
        commands = get_restart_commands()

    success = True
    if commands:
        success, _, messages = run_commands(commands, verbose)

    if success:
        # TODO:  determine set versus clear at commit time ? (was this and 'add' or 'del' command?

        kro_flags = file_get_key_value(WFI_KERNEL_SCRATCHPAD, WFI_KERNEL_ROUTE_OFFLOAD_FLAGS_KEY)

        value = kro_flags.get("kernel_route_offload_flags")
        if value:
            write_sx_netdev_parm("set" if int(value) == WFI_KERNEL_ROUTE_OFFLOAD_FLAG_SET else "clear")

        abort_pending()
        return (True, {})

    else:
        # If the commands fail, restore the config files to the backup to keep the
        # changes in 'pending'
        if os.path.isfile(WFI_KERNEL_SCRATCHPAD + ".bak"):
            persist_configurations([(WFI_KERNEL_CONFIG, WFI_KERNEL_SCRATCHPAD + ".bak")], make_backup=False)

        if os.path.isfile(MODPROBE_SCRATCHPAD + ".bak"):
            persist_configurations([(MODPROBE_CONFIG, MODPROBE_SCRATCHPAD + ".bak")], make_backup=False)

        messages["error"] = ["Unable to apply wait-for-install configuration.  See switchd.log for details"]

        return (False, messages)


def del_all():
    """
    net del allow

    Revert all configuration files associated with this plug-in to their defaults.
    """
    if os.path.isfile(WFI_KERNEL_SCRATCHPAD):
        with open(WFI_KERNEL_SCRATCHPAD, "w") as f:
            f.write(WFI_KERNEL_CONFIG_DEFAULT)

    if os.path.isfile(MODPROBE_SCRATCHPAD):
        with open(MODPROBE_SCRATCHPAD, "w") as f:
            f.write(MODPROBE_CONFIG_DEFAULT)


def show_config_summary(summary):
    """
    net show configuration port-mirror

    Populate the "summary" dictionary with information about the port_mirror configuration.
    """
    # nope
    pass


def show_config_commands(_ifupdown2_wrapper):
    return ([], [])


@use_exceptions(Exception)
def execute_command(net_daemon):
    """
    Execute an NCLU system management commands and return a standard 2-tuple

    (Called by NCLU framework code.)

    Side effects:
        Writes files to NCLU scratchpad.

    Returns:
        A 2-tuple of (bool, dict), where bool represents whether or not the
        command succeeded, and dict contains output/error messages to display
        to user.
    """

    assert isinstance(net_daemon, NetDaemon)

    # we should not get here.  there are no direct APIs into this module so NCLU should
    # never dispatch a command here.  Everything comes in through the backdoor add_del API
    # starting with a command that is dispatched to frr.
    raise NotImplementedError
    return (False, {"error": ["NCLU doesn't know how to process this command."]})


def get_config_files():
    """
    Return a string with the content of all existing, non-empty configuration files managed
    with this plug-in
    """
    return ''


def get_pending(use_colors):
    """
    Return a string of the pending changes in all managed files.
    """
    result = get_pending_diffs(
        [
            (WFI_KERNEL_CONFIG, WFI_KERNEL_SCRATCHPAD),
            (MODPROBE_CONFIG, MODPROBE_SCRATCHPAD),
        ],
        use_colors
    )

    if result:
        if use_colors:
            color_marker = "\n\033[91m\n"
        else:
            color_marker = "\n"
        result += """{0}
********************************************************************
********************************************************************
NOTE: "net commit" will run 'systemctl restart switchd' to apply the
configuration changes. This will be service impacting.
********************************************************************
********************************************************************
{1}
""".format(color_marker, color_marker)

    return result


def abort_pending():
    remove_scratch_files_and_backups([
        WFI_KERNEL_SCRATCHPAD,
        MODPROBE_SCRATCHPAD,
    ])


def get_managed_files():
    """
    Return a list of managed files.
    """
    return [
        WFI_KERNEL_CONFIG,
        MODPROBE_CONFIG,
    ]


def get_restart_commands(filename=None):
    """
    Return a list of commands that should be called for the given file
    after a rollback.
    """
    restart_commands = [
        ("/bin/systemctl", "enable", FRR_SERVICE),
        ("/bin/systemctl", "reset-failed", FRR_SERVICE),
        ("/bin/systemctl", "reload", FRR_SERVICE),
    ]

    # on Vx, (a 'software' platform) switchd does not exist so don't return
    # commands unless we are on a hardware platform.
    hw_sw = get_platform_info().get('platform')
    if hw_sw is None or hw_sw == 'software':
        # EARLY RETURN
        return restart_commands

    # don't enable switchd service if it is not enabled.
    if service_is_enabled(SWITCHD_SERVICE) or serivce_is_active(SWITCHD_SERVICE):
        restart_commands.append(("/bin/systemctl", "reset-failed", SWITCHD_SERVICE))
        restart_commands.append(("/bin/systemctl", "restart", SWITCHD_SERVICE))

    # otherwise, tet the reload commands and return those.
    return restart_commands


def add_del(tokens):

    mode = next((
        mode_token
        for mode_token in ["add", "del"]
        if mode_token in tokens
    ))

    did_create_wfi_kernel_scratchpad = False
    if not os.path.isfile(WFI_KERNEL_SCRATCHPAD):
        did_create_wfi_kernel_scratchpad = True
        make_scratchpads([
            (
                WFI_KERNEL_CONFIG,
                WFI_KERNEL_SCRATCHPAD,
                WFI_KERNEL_CONFIG_DEFAULT,
            ),
            (
                MODPROBE_CONFIG,
                MODPROBE_SCRATCHPAD,
                MODPROBE_CONFIG_DEFAULT,
            ),
        ])


    if "add" == mode:
        file_set_key_value(
            WFI_KERNEL_SCRATCHPAD,
            WFI_KERNEL_ROUTE_OFFLOAD_FLAGS_KEY,
            WFI_KERNEL_ROUTE_OFFLOAD_FLAG_SET
        )

        file_set_key_value(
            MODPROBE_SCRATCHPAD,
            MODPROBE_ROUTE_OFFLOAD_FLAGS_KEY,
            MODPROBE_ROUTE_OFFLOAD_FLAG_SET,
        )

    elif "del" == mode:
        file_set_key_value(
            WFI_KERNEL_SCRATCHPAD,
            WFI_KERNEL_ROUTE_OFFLOAD_FLAGS_KEY,
            WFI_KERNEL_ROUTE_OFFLOAD_FLAG_CLEAR
        )

        file_set_key_value(
            MODPROBE_SCRATCHPAD,
            MODPROBE_ROUTE_OFFLOAD_FLAGS_KEY,
            MODPROBE_ROUTE_OFFLOAD_FLAG_CLEAR,
        )

    return


def write_sx_netdev_parm(set_clear):
    # don't create this file if it does not exist.
    # EARLY RETURN
    if not os.path.exists(SX_NETDEV_CONFIG):
        return

    contents = (SX_NETDEV_CONFIG_SET if set_clear == "set" else SX_NETDEV_CONFIG_CLEAR) + "\n"
    with open(SX_NETDEV_CONFIG, "w") as f:
        f.write(contents.strip())
