"""
Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
Copyright (C) 2018, 2019 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt
"""

from collections import OrderedDict
from nclu import tabulate_remove_unicode
from nclu.plugins import DYNAMIC_RESTART, format_messages, run_commands
from nclu.plugins.frr.core import SERVICE_NAME
from pprint import pformat
from subprocess import CalledProcessError, Popen, STDOUT
from nclu.utils import check_output
from tabulate import tabulate
from threading import Thread, Event
from time import sleep
import logging
import string
import re
import os
import nclu.execute_command

log = logging.getLogger("netd")

CUMULUS_LIB = "/usr/lib/cumulus/"
BACKUP_SHORT = "config-backup"
BACKUP_TOOL = CUMULUS_LIB + BACKUP_SHORT
RESTORE_SHORT = "config-restore"
RESTORE_TOOL = CUMULUS_LIB + RESTORE_SHORT

class FutureRollback(Thread):
    """
    When the user does "net commit confirm", an instance of FutureRollback will be created.
    FutureRollback will revert the commit unless the user presses <ENTER> within "seconds".
    """

    def __init__(self, uid, gid, rollback_target, seconds, future_rollbacks):
        Thread.__init__(self)
        self.uid = uid
        self.gid = gid
        self.seconds = seconds
        self.rollback_target = rollback_target
        self.future_rollbacks = future_rollbacks
        self.shutdown_event = Event()

    def run(self):

        # Sleep for self.seconds, but wake up every 1s to see if we've been asked to terminate.
        for x in xrange(self.seconds):
            if self.shutdown_event.is_set():
                return
            sleep(1)

        log.info("The user did not press <ENTER> after \"net commit confirm\".  Rolled back to {0}.".format(self.rollback_target))
        Popen(["/usr/bin/net", "rollback", self.rollback_target])

        # Delete this instance from future_rollbacks.
        rollback_instance = self.future_rollbacks.get((self.uid, self.gid))

        if rollback_instance is not None:
            del self.future_rollbacks[(self.uid, self.gid)]


def rollback_entries():
    """
    Return a 2-tuple.  The first element is a dictionary from BACKUP_SHORT with -l
    output (fields are TAB separated).  For example:
#	Type	Description	Name
1	Initial	First system boot	config_backup-2019-03-13-21.24.01_superm-redxp-06
2	Dave-1	more play	config_backup-2019-03-13-21.26.39_superm-redxp-06
3	Pre-restore		config_backup-2019-03-13-21.57.11_superm-redxp-06
4	pre	nclu "net commit" (user root)	config_backup-2019-03-16-19.01.58_superm-redxp-06
5	Pre-restore		config_backup-2019-03-17-19.21.36_dave
6	Pre-restore		config_backup-2019-03-17-19.21.52_dave

    If config-restore is not installed, return ({}, None).
    """

    if not os.path.isfile(RESTORE_TOOL):
        return ({}, None)

    last_rollback_id = None
    entries = OrderedDict()

    try:
        output = check_output([RESTORE_TOOL, "-l"], stdin=None)
    except CalledProcessError:
        output = ''

    # Ignore the table header while iterating throught the lines of the output.
    for line in output.splitlines()[1:]:

        try:
            (rollback_id, rollback_type, desc, name) = map(str.strip, line.split('\t'))
            rollback_id = int(rollback_id)
        except ValueError:
            log.debug("This line from the output of " + RESTORE_SHORT + " raised ValueError:\n{0}".format(line))
            continue

        # convert backup name to human date to be similar to old output
        date = string.replace(name, 'config_backup-', '')
        date = re.sub('_.*', '', date)
        date = re.sub(r'-([0-5][0-9])[.]', r' \1.', date)
        date = re.sub('\.', ':', date)

        entries[rollback_id] = {
            "type": rollback_type,
            "id": rollback_id,
            "name": name,
            "date": date,
            "desc": desc,
        }

        last_rollback_id = rollback_id

    return (entries, last_rollback_id)


def rollback_last_nclu_pre_commit():
    """
    Return the ID of the most recent pre-commit.
    """

    entries, _ = rollback_entries()
    last_nclu_pre_commit = None

    for backup in entries.itervalues():
        if backup["type"] == "pre" and backup["desc"].startswith("nclu "):
            last_nclu_pre_commit = backup["id"]

    return last_nclu_pre_commit


def rollback_list_show(start, wrappers, plugins, color_diffs):
    """
    If the user did not specify a rollback, return the listing
    Otherwise, return the delta from start->stop for all of the files managed by NCLU.
    """

    assert start is None or isinstance(start, int)

    if not os.path.isfile(RESTORE_TOOL):
        return (False, "Cannot show backups because the " + RESTORE_SHORT + " utility is not installed.")

    entry, _ = rollback_entries()

    if start is None:
        # Return a list of all rollbacks.
        header = ('#', "Date     ", "Description")
        data = [(backup["id"], backup["date"], backup["desc"]) for backup in entry.itervalues() \
            if backup["desc"].startswith("nclu") and backup["type"] == "pre"]
        return (True, tabulate(tabulate_remove_unicode(data), header))

    if start not in entry:
        return (False, "Backup #{0} is not a NCLU related rollback.".format(start))

    # rollback_create_post() puts the matching pre-backup number as part
    # of the tag, so post-4 means that backup 4 is the matching pre to the
    # nclu post.
    # This code relies on that, and checks.  Other users of the backup tool may
    # not follow this, hence the checking.
    rollback_start = entry[start]

    if rollback_start["type"][0:4] == "post":
        if rollback_start["type"][4:5] == "-":
            preid = rollback_start["type"][5:]
            if preid:
                start = entry[int(preid)]["id"]
    elif rollback_start["type"][0:3] == "pre":
        preid = int(rollback_start["id"])
        # Find matching post
        for backup in entry.itervalues():
            if backup["type"][0:4] == "post" and backup["type"][4:5] == "-":
                id = backup["type"][5:]
                if id and int(id) == preid:
                    start = backup["id"]
                    break

    if start is None or start not in entry:
        return (False, "Backup #{0} is not a NCLU related rollback.".format(start))

    managed_files = get_all_managed_files(wrappers, plugins)  # A set

    reply = ''

    for filename in managed_files:
        try:
            # This is a really inefficient way to do it, but...
            output = check_output([RESTORE_TOOL, "-qqD",
                                  "-n{0}".format(start), "-F", filename],
                                  stderr=STDOUT, stdin=None)

            if color_diffs:
                for line in output.splitlines():
                    # Display the deletes in red and the adds in green.

                    if line.startswith("---") or line.startswith("+++"):
                        reply += "\033[94m{0}\033[0m\n".format(line)
                    elif line.startswith('-'):
                        reply += "\033[91m{0}\033[0m\n".format(line)
                    elif line.startswith('+'):
                        reply += "\033[92m{0}\033[0m\n".format(line)
                    else:
                        reply += line + '\n'
            else:
                reply = output

        except CalledProcessError:
            # There was no change for the filename in this rollback
            # This is okay.
            pass

    return (True, reply)


def rollback_create_pre(description, permanent):
    """
    Using config-backup, create a "pre" backup, and return the backup number as a string.
    If the backup number cannot be retreived, return None.
    """

    assert isinstance(description, str)
    assert isinstance(permanent, bool)

    # Do not allow TAB in the description because we will split on that character.
    description = description.replace('\t', ' ')

    command = [BACKUP_TOOL,
               "-D", description,
               "-t", "pre"]
    if permanent:
        command.extend(["-p"])

    # Fail reasonably if tool not available
    try:
        pre_rollback_number = check_output(command, stderr=STDOUT).strip()

        # Tool might print some messages.
        # The backup number will be the last line of the output.
        pre_rollback_number = re.sub(' .*', '', pre_rollback_number.splitlines()[-1])

        # Tool might fail, print an error message, and still exit 0.
        # Only honor the output if it is a number.
        if not pre_rollback_number.isdigit():
            log.warning("\"{0}\" returned\n\n{1}\n\n".format(pformat(command),
                                                             pre_rollback_number))
            pre_rollback_number = None

    except (CalledProcessError, OSError) as e:
        pre_rollback_number = None
        log.exception(e)

    assert isinstance(pre_rollback_number, str) or pre_rollback_number is None
    return pre_rollback_number


def rollback_create_post(description, pre_rollback_number, permanent):
    """
    Create a "post" rollback, and return the rollback number as a string.
    If the rollback number cannot be retreived, return None.
    """

    assert isinstance(description, str)
    assert isinstance(permanent, bool)

    # Do not allow '|' in the description because we will split on that character.
    description = description.replace('|', ' ')

    # encode the matching pre-number into the post- tag.  Used in rollback_show()
    # and in
    cmd = [BACKUP_TOOL,
           "-D", description,
           "-t", "post-{0}".format(pre_rollback_number)]
    if permanent:
        cmd.extend(["-p"])

    try:
        post_rollback_number = check_output(cmd, stderr=STDOUT).strip()
    except (CalledProcessError, OSError) as e:
        post_rollback_number = None
        log.exception(e)


def get_rollback_id_from_description(target_desc):
    """
    Return a 2-tuples of (True, int) or (False, str).  If True, the second element is a rollback ID.
    If False, it is an error message.
    """

    assert isinstance(target_desc, str)

    if not target_desc.startswith("nclu"):
        target_desc = "nclu " + target_desc

    entry, _ = rollback_entries()

    rollback_ids = [rollback_id for (rollback_id, value) in entry.iteritems() \
            if value.get("type") == "pre" and value.get("desc") == target_desc]

    if len(rollback_ids) == 1:
        return (True, rollback_ids[0])

    message = "There are {0} rollbacks with that description.  Use \"net show rollback <number>\" or \"net rollback <number>\" instead.".format(len(rollback_ids))
    return (False, message)


def rollback(rollback_target, wrappers, plugins, user):
    """
    For each file that NCLU manages, revert that file to rollback_target
    by running the restore tool with the given backup number
    Retrun (bool, str).  The boolean indicates if the rollback succeeded.  The string contains any
    messages for the user.
    """

    if not os.path.isfile(RESTORE_TOOL):
        return (False, "Rollback is not available because the " + RESTORE_SHORT + " utility is not installed.")

    entry, unused_rollback_id = rollback_entries()

    if rollback_target not in entry:
        return (False, "Backup #{0} is not a NCLU related rollback.".format(rollback_target))

    nclu.execute_command.abort_pending(wrappers, plugins)
    desc = "nclu \"{0} rollback\" (user {1})".format(rollback_target, user)
    pre_rollback_number = rollback_create_pre(desc, False)
    commands = []
    reply = ''

    # TODO - we need a better way to control the order in which these are applied
    #
    # We apply wrappers before plugins because ifupdown2 uses the wrapper model
    # and we need to apply the changes to /etc/network/interfaces first.

    # The way this was coded was somewhat confusing.  What it does is undo the
    # changes between the given backup, and now.  That is, it rolls back to the
    # named backup, but originally only for one file at a time.
    # Now we restore all at once.

    output = ""
    rollback_files_command = [RESTORE_TOOL, "-L", "-n{0}".format(rollback_target)]
    rollback_command = [RESTORE_TOOL, "-B", "-n{0}".format(rollback_target)]
    try:
        output = check_output(rollback_files_command, stdin=None)
    except CalledProcessError:
        reply += "WARNING: problems during configuration restore"
        # keep going, since we don't know what got restored
        # and want to handle everything that did

    diffdict = {}
    flist = output.splitlines()

    # Don't want lines starting with config-restore
    out_files = []
    for line in flist:
        if  line.startswith('config-restore'):
            continue
        filen = line.lstrip()
        out_files.append(filen)
        # Get the diff for each file in the list, and store it for later
        # We have to do this before the files are restored, otherwise there
        # will be no differences
        diff_command = [RESTORE_TOOL, "-qqD", "-n{0}".format(rollback_target), "-F",
                filen]
        diffdict[filen] = ""
        try:
            diffdict[filen] = check_output(diff_command, stderr=STDOUT,
                                      stdin=None).splitlines()
        except CalledProcessError:
            reply += "WARNING: problems during configuration diff for {0}".format(filen)

    if out_files == []:
        log.info("No files were rolled back, return early")
        reply += "No files were rolled back"
        return (True, reply)

    try:
        output = check_output(rollback_command, stdin=None)
    except CalledProcessError:
        reply += "WARNING: problems during configuration restore"
        # keep going, since we don't know what got restored
        # and want to handle everything that did

    # Get a list of associated restart commands for rolled back files from
    # wrappers.
    assert len(wrappers) > 0, "If the wrappers paradigm has been removed, edit this function."
    for _, wrapper in wrappers:
        for filename in wrapper.get_managed_files():
            for line in out_files:
                if line != filename:
                    continue
                commands.extend(wrapper.get_service_cmd_for_file(filename))

    # Get a list of associated restart commands for rolled back files from
    # plugins
    for plugin in plugins:
        m = __import__("nclu.plugins." + plugin, fromlist=["get_managed_files", "get_restart_commands"])

        for filename in m.get_managed_files():
            for line in out_files:
                if line != filename:
                    continue
                # The call succeeded.  Get any associated restart commands.
                if plugin in DYNAMIC_RESTART:
                    # The plug-in has non-static restart commands depending on
                    # the specific changes.
                    if not diffdict[line]:
                        continue
                    try:
                        if plugin == "dns":
                            filename_commands, _, _ = \
                                m.get_restart_commands(filename,
                                    diffdict[line])
                        else:
                            filename_commands = m.get_restart_commands(filename,
                                    diffdict[line])
                        commands.extend(filename_commands)
                    except CalledProcessError:
                        reply += "WARNING: Couldn't get the restart commands for {0}.\n".format(plugin.upper())
                else:
                    commands.extend(m.get_restart_commands(filename))

    rollback_create_post(desc, pre_rollback_number, False)

    # Don't call "systemctl ... switchd.service" commands multiple times.  They are expensive.

    assert ["/bin/systemctl", "enable", "switchd"] not in commands, "Use the full service name, switchd.service."
    while commands.count(["/bin/systemctl", "enable", "switchd.service"]) > 1:
        commands.remove(["/bin/systemctl", "enable", "switchd.service"])

    assert ["/bin/systemctl", "reset-failed", "switchd"] not in commands, "Use the full service name, switchd.service."
    while commands.count(["/bin/systemctl", "reset-failed", "switchd.service"]) > 1:
        commands.remove(["/bin/systemctl", "reset-failed", "switchd.service"])

    assert ["/bin/systemctl", "restart", "switchd"] not in commands, "Use the full service name, switchd.service."
    while commands.count(["/bin/systemctl", "restart", "switchd.service"]) > 1:
        commands.remove(["/bin/systemctl", "restart", "switchd.service"])

    if ("/bin/systemctl", "restart", SERVICE_NAME) in commands or ("/bin/systemctl", "stop", SERVICE_NAME) in commands:
        # Don't try to reload FRR.  It is already being restarted or stopped.
        log.debug("Removing FRR \"reload\" commands from the rollback commands.")

        while ("/bin/systemctl", "reload", SERVICE_NAME) in commands:
            commands.remove(("/bin/systemctl", "reload", SERVICE_NAME))

    log.debug("Removing duplicate FRR \"reset-failed\" commands, if any.")

    while commands.count(("/bin/systemctl", "reset-failed", SERVICE_NAME)) > 1:
        commands.remove(("/bin/systemctl", "reset-failed", SERVICE_NAME))

    if ("/bin/systemctl", "stop", SERVICE_NAME) in commands:
        # Don't try to enable FRR.
        log.debug("Removing FRR \"enable\" commands from the rollback commands.")

        while ("/bin/systemctl", "enable", SERVICE_NAME) in commands:
            commands.remove(("/bin/systemctl", "enable", SERVICE_NAME))

    # Perform "ifreload -a", "systemctl reload frr.service", etc.
    outcome, _, messages = run_commands(commands, False)
    reply += format_messages(messages, False)

    return (outcome, reply)


def rollback_show(rollback_target, wrappers, plugins, color_diffs):

    if not os.path.isfile(RESTORE_TOOL):
        return (False,
                "Rollback is not available because the " + RESTORE_SHORT + " utility is not installed.")

    entry, unused_rollback_id = rollback_entries()

    if rollback_target not in entry:
        return (False, "Backup #{0} is not a NCLU related rollback.".format(rollback_target))

    # We need to build a list of files that we manage that have changed between
    # the current state of the filesystem (rollback 0) and rollback_target.
    # config-restore -D -n rollback_target' will give us a list of all files that
    # changed, look through that output for the files that we manage.
    #
    # We need said list because we have to tell 'config-restore -D' which files
    # to diff.
    # If we ask it to diff a file that hasn't changed it will return non-zero.
    nclu_managed_files = get_all_managed_files(wrappers, plugins)
    nclu_managed_files_in_rollback = []

    command = [RESTORE_TOOL, "-Lqq", "-n{0}".format(rollback_target)]
    try:
        output = check_output(command)
    except CalledProcessError as e:
        return (False, "\"{0}\" failed due to \"{1}\".".format(' '.join(command), e.output.strip()))

    for filename in output.splitlines():
        if filename in nclu_managed_files:
            nclu_managed_files_in_rollback.append(filename)

    command = [RESTORE_TOOL, "-qqD", "-n{0}".format(rollback_target)]
    command.extend(nclu_managed_files_in_rollback)
    reply = ''

    try:
        output = check_output(command, stderr=STDOUT, stdin=None)

        if color_diffs:
            # Output the deletes in red and the adds in green.
            for line in output.splitlines():
                if line.startswith("---") or line.startswith("+++"):
                    reply += "\033[94m{0}\033[0m\n".format(line)
                elif line.startswith('-'):
                    reply += "\033[91m{0}\033[0m\n".format(line)
                elif line.startswith('+'):
                    reply += "\033[92m{0}\033[0m\n".format(line)
                else:
                    reply += line + '\n'
        else:
            reply = output

    except CalledProcessError:
        return (False, "\"{0}\" failed due to \"{1}\".".format(' '.join(command), e.output.strip()))

    return (True, reply)


def get_all_managed_files(wrappers, plugins):
    """
    Return a set of all files managed by the given wrappers and plugins.
    """

    assert len(wrappers) > 0, "If the wrappers paradigm has been removed, edit this function."

    managed_files = set()

    for plugin in plugins:
        m = __import__("nclu.plugins." + plugin, fromlist=["get_managed_files"])
        managed_files.update(m.get_managed_files())

    for _, wrapper in wrappers:
        managed_files.update(wrapper.get_managed_files())

    return managed_files
