# Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
# Copyright (C) 2020 Cumulus Networks, inc.
#
# All rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import subprocess
import json
import logging
import re


logger = logging.getLogger('netd')


MIRROR_PREFIX = "mirror"


def match_swp(ifname):
    """
    Try to match an swp in the given string 'ifname'.

    Matches any of the following examples:
        swp0
        swp0s0
        sw0
        sw0p0
        sw0p0s0

    Returns a match object or None.

    Given the example 'sw0p1s2', the following groups can be retrieved
    from the match object, m:
        m.group(0): 'sw0p1s2'
        m.group('base_name'): 'sw'
        m.group('suffix'): '0p1s2'
        m.group('module'): '0'
        m.group('port'): '1'
        m.group('breakout'): '2'
    """

    pattern = r'^(?P<base_name>sw)(?P<suffix>(?P<module>\d*)p(?P<port>\d+)(s(?P<breakout>\d+))?)$'
    return re.match(pattern, ifname)



def is_mirror_interface(ifname):
    """
    Determine if an interface is a mirror interface.

    Args:
        ifname: A string interface name.

    Returns:
        True if the `ifname` is a mirror interface, otherwise False.
    """
    return ifname.startswith(MIRROR_PREFIX)


def check_output(command, *args, **kwargs):
    '''
    Basically a facade over the check_output
    command that we can do logging with.
    '''
    messages = kwargs.pop('messages', None)
    verbose = kwargs.pop('verbose', False)
    fmt_command = command
    if isinstance(fmt_command, (list, tuple)):
        fmt_command = ' '.join(fmt_command)
    try:
        result = subprocess.check_output(command, *args, **kwargs)
        # success means return code 0
        logger.debug('Ran command: {}, Return code: {}'.format(
            fmt_command, 0))

        if verbose and messages is not None:
            if "output" not in messages:
                messages["output"] = ''
            if result:
                messages["output"] += "{0}\n---------\n{1}\n\n".format(
                    fmt_command, result)
            else:
                messages["output"] += "Ran {0}.\n\n".format(fmt_command)

        return result
    except subprocess.CalledProcessError as exc:
        logger.debug('Ran command: {}, Return code: {}, Output: {}'.format(
            fmt_command, exc.returncode, exc.output))
        if verbose and messages:
            if "output" not in messages:
                messages["output"] = ''
            if exc.output:
                messages["output"] += "{0}\n---------\n{1}\n\n".format(
                    fmt_command, exc.output)
            else:
                messages["output"] += "Ran {0}. Exited {1}\n\n".format(
                    fmt_command, exc.returncode)

        # just bubble, we're here for the logging
        raise


def safe_unidecode(val):
    '''
    Convert a string in a safe way to unicode
    >>> safe_unidecode(b'foobar')
    'foobar'
    :param val: str to convert
    '''
    if isinstance(val, unicode):
        # already decoded
        return val

    for codec in ('utf-8', 'windows-1252', 'latin-1'):
        try:
            return val.decode(codec)
        except UnicodeDecodeError:
            pass
    return val.decode('utf-8', errors='replace')


class Config:
    def __init__(self, config_file):
        self.config_file = config_file
        self.show_linux_command = None
        self.color_diffs = None
        self.blacklist = None
        self.components = None


    def parse(self):
        try:
            # Read the configuration file.
            with open(self.config_file) as f:
                self.config_text = f.read()
        except OSError as e:
            exit("Couldn't read the configuration file, {0}.  ({1})".format(self.config_file, e))

        # These boolean attributes pertain to how the output returned to the user should look.
        self.show_linux_command = re.search(r"""^\s*show_linux_command\s*=\s*True""", self.config_text, re.IGNORECASE | re.MULTILINE) is not None
        self.color_diffs = re.search(r"""^\s*color_diffs\s*=\s*True""", self.config_text, re.IGNORECASE | re.MULTILINE) is not None

        # Populate the permissions attributes.  These are four frozensets that specify who can do what.  If the permission is
        # not specified in the configuration file, or if it is set to blank, the resulting attribute will be an empty frozenset.
        for permissions_group in ("users_with_edit", "users_with_show", "groups_with_edit", "groups_with_show"):
            users_re = re.search(r"""^\s*{0}\s*=\s*(?P<users>\S[\S\t ,]*)$""".format(permissions_group), self.config_text, re.IGNORECASE | re.MULTILINE)
            # TODO - Should users be mapped to lowercase?  Are "root" and "ROOT" different to the OS?
            setattr(self, permissions_group, frozenset([] if (users_re is None) else users_re.group("users").replace(',', ' ').split()))

        # Populate the blacklist for ifupdown.  Blacklisted terms serve to filter the available ifupdown2 commands.
        blacklist_re = re.search(r"""^\s*ifupdown_blacklist\s*=\s*(?P<blacklist>\S[\S\t ,]*)$""", self.config_text, re.IGNORECASE | re.MULTILINE)
        self.blacklist = frozenset([] if (blacklist_re is None) else blacklist_re.group("blacklist").replace(',', ' ').split())

        # Get a lowercase set of all components configured to be enabled.  The netmisc and ports components are always available.
        self.enabled_wrappers = set(["netmisc"])
        self.enabled_plugins = set(["ports"])
        self.components = map(str.lower, re.findall(r"""^\s*enable_(\S+)\s*=\s*True""", self.config_text, re.IGNORECASE | re.MULTILINE))


######################
# Auto BGP utilities
######################
# These are the official bounds, as per RFC 6996
FIRST_PRIVATE_ASN = 4200000000
LAST_PRIVATE_ASN = 4294967294

SPINE_ASN = FIRST_PRIVATE_ASN

FIRST_LEAF_ASN = SPINE_ASN + 1
LAST_LEAF_ASN = LAST_PRIVATE_ASN
# Add one, since the range is inclusive.
RANGE_LEAF_ASNS = LAST_LEAF_ASN - FIRST_LEAF_ASN + 1


BIN_SYS_EEPROM = "/usr/cumulus/bin/decode-syseeprom"


def mac_address_to_integer(mac_string):
    """
    Convert a MAC address into its integer form.

    Args:
        mac_string: The string MAC address in hexadecimal form.

    Returns:
        The integer form of the MAC address.
    """
    if isinstance(mac_string, int):
        # Do nothing if it's already an int
        return mac_string
    return int(
        mac_string.replace(":", "").replace("-", "").replace(".", ""),
        16
    )


def generate_auto_leaf_asn_from_mac_address(mac_string):
    """
    Generate a leaf ASN by converting the MAC address into an integer, and
    coercing that integer into the private ASN range.

    Args:
        mac_string: The string MAC address in hexadecimal form.

    Returns:
        An integer private ASN.
    """
    mac_int = mac_address_to_integer(mac_string)
    return FIRST_LEAF_ASN + (mac_int % RANGE_LEAF_ASNS)


def generate_auto_spine_asn():
    """
    It's always the same, but it's better to have these both in the same place.

    Returns:
        An integer private ASN.
    """
    return SPINE_ASN


def generate_auto_leaf_asn():
    """
    Generate a leaf ASN in the private range, generated from the device's base
    MAC address.

    Returns:
        An integer private ASN, or None if a valid MAC address couldn't be
        found.

    Side effects:
        Executes `decode-syseeprom --json`
    """
    mac_address = retrieve_base_mac_address()
    if mac_address is not None:
        return generate_auto_leaf_asn_from_mac_address(mac_address)
    else:
        return None


def retrieve_base_mac_address():
    '''
    Generate an asn from a hash of the mac address modulo (private asn range).

    Returns:
        A MAC address if one could be found, otherwise None.

    Side effects:
        Executes `decode-syseeprom --json`
    '''
    try:
        # Get the eeprom data
        output = check_output([BIN_SYS_EEPROM, "--json"])
        eeprom_data = json.loads(output)
        return eeprom_data["tlv"]["Base MAC Address"]["value"]
    except (subprocess.CalledProcessError, ValueError, KeyError) as e:
        logger.exception(e)

    return None


def get_platform_info():
    """
    Get platform information.  Returns a dict with platform hw/sw
    indication (sw being on a Vx).  I can see this expanding to return
    some more platform hardware (soc/ASIC) info when there is one.
    """
    info  = {}

    try:
        platform_detect = check_output("/usr/bin/platform-detect").strip()
    except CalledProcessError as e:
        log.exception(e)
        return info

    if platform_detect:
        # platform_detect looks like "dell,s6000_s1220".
        _, second = platform_detect.split(',')
        if second == "vx":
            info['platform'] = 'software'
        else:
            info['platform'] = 'hardware'

    return info
