"""
Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
Copyright (C) 2017-2020 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt

Settings managed in this module are set in accordance with this Mellanox documentation:
https://community.mellanox.com/docs/DOC-3036


Usage:
    net (add|del) forwarding ecmp hash-seed <0-4294967295>

Options:
    forwarding              : Packet forwarding parameters
    ecmp                    : Equal Cost MultiPath
    hash-seed               : Hash seed for ECMP entries
"""

from collections import OrderedDict
from nclu import (
    ConfigWrapperBase,
    ConfigFileKeyValue,
    globify_interface_commands,
    platform_vendor,
    get_interfaces,
    ifnames_to_glob,
    ifname_expand_glob,
    file_get_key_value,
    get_pending_add_del_commands,
)
from nclu.wrappers.ifupdown2 import (
    get_ifnames,
    EtcNetworkInterface,
)
from nclu.utils import match_swp
from network_docopt import sort_ifaces
import logging
import os
import shutil
import re
import json
import math
from tabulate import tabulate

QOS_FEATURES = "/etc/cumulus/datapath/qos/qos_features.conf"
ETC_TRAFFIC = "/etc/cumulus/datapath/traffic.conf"
QOS_FEATURES_SCRATCHPAD = "/run/nclu/traffic/etc_cumulus_datapath_qos_qos_features_conf_scratchpad"
QOS_INFRA = "/etc/mlx/datapath/qos/qos_infra.conf"
QOS_CONFIG = "/etc/cumulus/switchd.d/qos.conf"
ROCE_MODE_KEY = "traffic.roce_mode"
SWITCHDCTL_COMMAND = "/usr/lib/cumulus/switchdctl"
CC_DIR = "/usr/cumulus/bin/"
EGRESS_SCHED_KEY = "traffic.egress_sched"
EGR_QUEUE_LIST = [0,1,2,3,4,5,6,7]
DEFAULT_EGR_SCHED_BW = {0:12, 1:13, 2:12, 3:13, 4:12, 5:13, 6:12, 7:13}
DEFAULT_MC_SP_DYNAMIC_QUOTA = {0:'ALPHA_1_4', 1:'ALPHA_8', 2:'ALPHA_4', 3:'ALPHA_2', 4:'ALPHA_1_8', 5:'ALPHA_1', 6:'ALPHA_1_2', 7:'ALPHA_1_4'}
DEFAULT_EGR_QUEUE_UC_DYNAMIC_QUOTA = {0:'ALPHA_2', 1:'ALPHA_4', 2:'ALPHA_1', 3:'ALPHA_1_2', 4:'ALPHA_1_4', 5:'ALPHA_1_8', 6:'ALPHA_1_16', 7:'ALPHA_1_32'}
DEFAULT_EGRESS_SCHED_PROFILE_NAME = "sched_port_group1"

MAP_ROCE_MODE = {
    "disabled"  : 0,  # ROCE mode is disabled (not configured)
    "lossy"     : 1,  # ROCE lossy mode configured
                      # Note: roce mode 2 is N/A
    "lossless"  : 3,  # ROCE lossless mode configured
}

MAP_EGRESS_SCHED = {
    "disabled"      : 0,  # No Egress scheduler configured
    "egress_sched"  : 2,  # 8-egress scheduler config per port-group
}

log = logging.getLogger("netd")

def build_doc_string():
    global __doc__

    # QoS configuration CLI is only supported on MLNX platforms.
    if platform_vendor() in ("mlnx", "cumulus") :
        mlnx_doc = """
Usage:
    net add roce [lossless|lossy]
    net del roce

    net add qos egress scheduler profile default
    net add qos egress scheduler profile (<prof-name>|<text>) queue <0-7> (strict|dwrr bw-percent <1-100>)
    net del qos egress scheduler profile <prof-name> [queue <0-7>]
    net add qos egress scheduler profile (<prof-name>|<text>) interface <interface>
    net del qos egress scheduler profile <prof-name> interface [<interface>]

    net show roce config [json]

Options:
    qos                          : QoS related configurations
    egress                       : QoS Egress configurations
    scheduler                    : 8-egress scheduler configuration
    queue                        : egress queue for 8-egress scheduler config
    bw-percent                   : bandwidth percent value for 8-egress scheduler config
    strict                       : Strict priority scheduling on respective egress queue
    dwrr                         : DWRR weight scheduling on respective egress queue depending on bandwidth percent
    default                      : default profile for 8-egress scheduler config applicable for all the ports
    profile                      : profile to apply 8-egress scheduler config.
    <prof-name>                  : non-default profile for egress scheduler config
    roce                         : Enable RoCE on all interfaces, default mode is lossless
    lossy                        : RoCE lossy mode
    lossless                     : RoCE lossless mode
    config                       : RoCE configuration
"""
        __doc__ += mlnx_doc

    return __doc__

# Override keyword definitions created by other plugins/wrappers
per_command_helptext = {
    "net (add|del) qos": {
        "egress": "QoS egress direction"
    },
    "net (add|del) qos egress scheduler profile": {
        "default": "egress scheduler default profile"
    },
    "net add qos egress scheduler profile": {
        "<text>": "egress scheduler profile name"
    },
    "net add qos egress scheduler profile (<prof-name>|<text>) queue": {
        "<0-7>": "egress queue number"
    },
    "net del qos egress scheduler profile <prof-name> queue": {
        "<0-7>": "egress queue number"
    },
}

def convert_list_string_to_list(list_string, list_is_ifnames):
    """
    traffic.conf has several keywords with a list as the value but is inconsistent
    with the format it uses to store the list.  Some of them use []s while others
    do not. Parse the various formats and return a list.
    """

    if list_string is None:
        return []

    assert isinstance(list_is_ifnames, bool)
    assert isinstance(list_string, basestring)

    # Standardize the format of list_string by:
    # - remove leading and trailing whitespaces
    # - remove leading and trailing []s
    # - replace commas with whitespaces
    # - split by whitespace and then rejoin with a single whitespace
    list_string = ' '.join((list_string.strip(" []").replace(',', ' ').split()))
    result = list_string.split()

    # Remove duplicate values within the list.
    result = list(set(result))

    # If these are interface names, sort accordingly.
    return sort_ifaces(result) if list_is_ifnames else sorted(result)

def get_egress_scheduler_profile_names():
    final = []
    if os.path.isfile(QOS_FEATURES_SCRATCHPAD):
        filename = QOS_FEATURES_SCRATCHPAD
    elif os.path.isfile(QOS_FEATURES):
        filename = QOS_FEATURES
    else:
        return final

    port_group_key = "egress_sched.port_group_list"
    config = file_get_key_value(filename, port_group_key)
    port_group_str = config.get(port_group_key, "")
    if port_group_str:
        for ch in ['[',']']:
            port_group_str = port_group_str.replace(ch,'')
        final = convert_list_string_to_list(port_group_str,False)

    return final

def match_egr_sched_profile(argv_word, use_fuzzy, tab_completing):
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_egress_scheduler_profile_names():
            return True

        if tab_completing:
            for prof in get_egress_scheduler_profile_names():
                if prof.startswith(argv_word):
                    return True
    return False

def tab_egr_sched_profile(args, ended_with_space, last_argv):
    final = []

    for x in get_egress_scheduler_profile_names():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'existing configured profile for egress scheduler config'))

    if "add" in args:
        final.append(('<prof-name>', 'non-default profile for egress scheduler config'))

    return final

class ConfigWrapper(ConfigWrapperBase):

    def __init__(self, show_linux_command, color_diffs):
        super(ConfigWrapper, self).__init__("traffic", show_linux_command, color_diffs)

        self.files[QOS_CONFIG] =\
            ConfigFileKeyValue(self, QOS_CONFIG, [])

        self.files[ETC_TRAFFIC] =\
            ConfigFileKeyValue(self, ETC_TRAFFIC, [])

        self.files[QOS_FEATURES] =\
            ConfigFileKeyValue(self, QOS_FEATURES, [])

        self.files[QOS_INFRA] =\
            ConfigFileKeyValue(self, QOS_INFRA, [])

        # roce commands introduce a roce_mode in qos.conf that allows backward
        # compatibility with legacy storage-optimized commands, though they both
        # cannot be configured at the same time.  The qos.conf has a 'roce_mode'
        # that is used to keep the 'doROCE mode' as well as a setting to 'legacy'
        # to prevent the two sets of commands (doROCE mode and legacy
        # storage-optimized mode) from stepping on each other.
        # In the case the qos.conf file is not present, create the default file
        # with the default setting "roce disabled"
        if not os.path.exists(QOS_CONFIG):
            with open(QOS_CONFIG, 'a') as f:
                line = '{0} = {1}\n'.format(ROCE_MODE_KEY, MAP_ROCE_MODE['disabled'])
                line += '{0} = {1}\n'.format(EGRESS_SCHED_KEY, MAP_EGRESS_SCHED['disabled'])
                f.write(line)

    def get_service_cmd_for_file(self, filename):
        """
        This wrapper manages 3 files (traffic.conf, datapath.conf, and qos.conf) though
        the restart commands are not unique to the files. So no need to return a set
        of restart commands for each of the three files as the dups get filtered out anway.
        So just return commands on one config file. We will return handling commands for
        any file which is updated.
        """
        cmds = []

        # if RoCE mode is being configured, we can use 'reload' of switchd instead of
        # restart
        reload_cmd = "reload"

        # For non-mlx platform, cl-consistency-check and FUSE node reload is not supported - will do switchd restart
        if platform_vendor() not in ("mlnx") :
            reload_cmd = "restart"

        if reload_cmd == "reload":
            # For reload, we need to call switchd reload
            cmds = [
                ["/bin/systemctl", "reload", "switchd.service"]
            ]
        elif reload_cmd == "restart":
            cmds = [
                ["/bin/systemctl", "enable", "switchd.service"],
                ["/bin/systemctl", "reset-failed", "switchd.service"],
                ["/bin/systemctl", "restart", "switchd.service"]
            ]

        return cmds

    def get_conf_file_key_value(self, filename, key):
        return self.files[filename].get_value().get(key)


    def get_legacy_ifnames(self):
        """
        Return a 2-tuple of two lists.  The first list includes interface names that have been configured for storage-optimized.
        The second list includes associated ecn port settings.
        """

        cfg_content = self.files[QOS_FEATURES].get_value()

        pfc_port_set = cfg_content.get("pfc.ROCE_PFC.port_set")
        pfc_port_set = convert_list_string_to_list(pfc_port_set, True)

        ecn_port_set = cfg_content.get("ecn_red.ROCE_ECN.port_set")
        ecn_port_set = convert_list_string_to_list(ecn_port_set, True)

        return (pfc_port_set, ecn_port_set)


    def show_config_commands(self, user_may_edit, ifupdown2_wrapper):
        supported = []
        interface_commands = []
        unsupported = []

        roce_config_mode = self.get_roce_config_mode()
        egress_sched_config_mode = self.get_egress_sched_config_mode()

        if roce_config_mode != MAP_ROCE_MODE['disabled']:
            # build the 'net add roce' commands
            supported.append("net add roce {0}".format("lossy" if roce_config_mode == MAP_ROCE_MODE['lossy'] else "lossless"))

        if egress_sched_config_mode == MAP_EGRESS_SCHED['egress_sched']:
            # build the 'net add qos egress_sched' commands
            supported.append("net add qos egress scheduler profile default")
            egr_sched_port_group_str = self.get_conf_file_key_value(QOS_FEATURES, "egress_sched.port_group_list")
            egr_sched_port_group_list = []
            if egr_sched_port_group_str:
                for ch in ['[',']']:
                    egr_sched_port_group_str = egr_sched_port_group_str.replace(ch,'')
                egr_sched_port_group_list = convert_list_string_to_list(egr_sched_port_group_str, False)
                for egr_sched_port_group in egr_sched_port_group_list:
                    egr_sched_port_grp_port_set_key = "egress_sched.{}.port_set".format(egr_sched_port_group)
                    egr_sched_port_grp_port_set_str = self.get_conf_file_key_value(QOS_FEATURES, egr_sched_port_grp_port_set_key)
                    if egr_sched_port_grp_port_set_str:
                        supported.append("net add qos egress scheduler profile {} interface {}".format(egr_sched_port_group,egr_sched_port_grp_port_set_str))
                    for egr_queue in EGR_QUEUE_LIST:
                        egr_sched_port_grp_bw_key = "egress_sched.{}.egr_queue_{}.bw_percent".format(egr_sched_port_group,egr_queue)
                        egr_sched_port_grp_bw_val = self.get_conf_file_key_value(QOS_FEATURES, egr_sched_port_grp_bw_key)
                        if egr_sched_port_grp_bw_val is not None:
                            if egr_sched_port_grp_bw_val == 0:
                                supported.append("net add qos egress scheduler profile {} queue {} strict".format(egr_sched_port_group, egr_queue))
                            else:
                                supported.append("net add qos egress scheduler profile {} queue {} dwrr bw_percent {}".format(egr_sched_port_group, egr_queue, egr_sched_port_grp_bw_val))

        # ecmp hash is it's own beast independent of the storage-optimized add commands so look for that one also.
        ecmp_hash_seed = self.get_conf_file_key_value(ETC_TRAFFIC, "ecmp_hash_seed")
        if ecmp_hash_seed is not None:
            supported.append("net add forwarding ecmp hash-seed {0}".format(ecmp_hash_seed))

        return (supported, unsupported)

    def show_config_summary(self, user_may_edit, summary):
        (pfc_port_set, ecn_port_set) = self.get_legacy_ifnames()

        roce_mode = self.get_roce_pending_mode(self.files[QOS_CONFIG])

        for ifname in pfc_port_set + ecn_port_set:
            ifname = "INTERFACE:{0}".format(ifname)

            if ifname not in summary:
                summary[ifname] = OrderedDict()

            if "COMMAND_LIST" not in summary[ifname]:
                summary[ifname]["COMMAND_LIST"] = []

        for ifname in pfc_port_set:
            if not self.is_roce_mode(roce_mode):
                ifname = "INTERFACE:{0}".format(ifname)
                summary[ifname]["COMMAND_LIST"].append("storage-optimized pfc")

        for ifname in ecn_port_set:
            if ifname not in pfc_port_set and not self.is_roce_mode(roce_mode):
                ifname = "INTERFACE:{0}".format(ifname)
                summary[ifname]["COMMAND_LIST"].append("storage-optimized")

        ecmp_hash_seed = self.get_conf_file_key_value(ETC_TRAFFIC, "ecmp_hash_seed")

        if ecmp_hash_seed is not None:
            if "forwarding" not in summary:
                summary["forwarding"] = []

            summary["forwarding"].append("ecmp hash-seed {0}".format(ecmp_hash_seed))

    def get_pending(self):
        """
        Return the pending diff as a string.  If there are no changes, return the empty string.
        """

        result = ConfigWrapperBase.get_pending(self)

        if result:
            restart_cmd = 'systemctl reload switchd'
            result += """\n\033[91m\n
********************************************************************
********************************************************************
NOTE: "net commit" will run '{0}' to apply the
configuration changes. This will be service impacting.
********************************************************************
********************************************************************
\033[0m
""".format(restart_cmd)

        return result

    def del_all(self, cli):
        """ Remove all RoCE related configuration. """

        # get all of the scratchpad files created.
        for etc_filename, cfg in self.files.iteritems():
            if os.path.exists(cfg.filename) and not os.path.isfile(cfg.scratchpad):
                shutil.copyfile(cfg.filename, cfg.scratchpad)

        # grab the roce_mode
        is_roce_mode = self.is_roce_mode(int(self.get_conf_file_key_value(QOS_CONFIG, ROCE_MODE_KEY)))
        is_egress_sched_pending_conf = self.is_egress_sched_mode(int(self.get_egress_sched_pending_mode(self.files[QOS_CONFIG])))

        # walk through each config file and restore it.
        for etc_filename, cfg in self.files.iteritems():
            cfg_content = cfg.get_value(True)

            # restore traffic.conf and datapath.conf independently.
            if etc_filename == QOS_FEATURES:
                # in egress sched mode, restore all the egress sched config from traffic.conf
                if is_egress_sched_pending_conf:
                    egr_sched_port_group_str = cfg.get_value(True).get("egress_sched.port_group_list")
                    egr_sched_port_group_list = []
                    if egr_sched_port_group_str:
                        for ch in ['[', '[']:
                            egr_sched_port_group_str = egr_sched_port_group_str.replace(ch,'')
                        egr_sched_port_group_list = convert_list_string_to_list(egr_sched_port_group_str, False)
                        for prof_name in egr_sched_port_group_list:
                            self.configure_all_egress_queues(prof_name,"del")
            # in Roce mode, we restore all .conf files when deleting roce.
            elif etc_filename == QOS_CONFIG:
                # in Roce mode, restore all the .conf files (traffic.conf, datapath.conf, and qos.conf)
                # in one shot.
                if is_roce_mode:
                    self.del_roce_mode(cfg)

                # reset the roce mode to the default ('disabled') on del_all as all
                # configuration (legacy or roce) has been deleted.
                cfg.edit(ROCE_MODE_KEY, MAP_ROCE_MODE['disabled'], give_feedback=False)

                # configure egress sched mode as 'disabled'
                cfg.edit(EGRESS_SCHED_KEY, MAP_EGRESS_SCHED["disabled"], give_feedback=False)

    def qos_features_conf_eval_roce(self, add_del, roce_mode):
        """
        Handle all add/del for roce in the traffic.conf file.
        """

        cfg_qos_features = self.files[QOS_FEATURES]
        cfg_content = cfg_qos_features.get_value(True)
        command_ok = False

        ifnames = self.ifnames_all_swps()

        if add_del == "add":
            command_ok = True

            # =======================================================================
            # 1. Enable ECN and RED for priority 3 (By default ECN is enabled on 0)
            # =======================================================================

            cfg_qos_features.edit("default_ecn_conf.egress_queue_list", "[0,3]", give_feedback=False)

            # ======================================================
            # 2. Set trust mode to DSCP, and map DSCP values to COS.
            # ======================================================
            cfg_qos_features.edit("traffic.packet_priority_source_set", "[802.1p,dscp]", give_feedback=False)

            cfg_qos_features.edit("traffic.cos_0.priority_source.8021p", "[0]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_1.priority_source.8021p", "[1]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_2.priority_source.8021p", "[2]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_3.priority_source.8021p", "[3]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_4.priority_source.8021p", "[4]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_5.priority_source.8021p", "[5]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_6.priority_source.8021p", "[6]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_7.priority_source.8021p", "[7]", give_feedback=False)

            cfg_qos_features.edit("traffic.cos_0.priority_source.dscp", re.sub(r'\s+','',str(range(8))), give_feedback=False)
            cfg_qos_features.edit("traffic.cos_1.priority_source.dscp", re.sub(r'\s+','',str(range(8,16))), give_feedback=False)
            cfg_qos_features.edit("traffic.cos_2.priority_source.dscp", re.sub(r'\s+','',str(range(16,24))), give_feedback=False)
            cfg_qos_features.edit("traffic.cos_3.priority_source.dscp", re.sub(r'\s+','',str(range(24,32))), give_feedback=False)
            cfg_qos_features.edit("traffic.cos_4.priority_source.dscp", re.sub(r'\s+','',str(range(32,40))), give_feedback=False)
            cfg_qos_features.edit("traffic.cos_5.priority_source.dscp", re.sub(r'\s+','',str(range(40,48))), give_feedback=False)
            cfg_qos_features.edit("traffic.cos_6.priority_source.dscp", re.sub(r'\s+','',str(range(48,56))), give_feedback=False)
            cfg_qos_features.edit("traffic.cos_7.priority_source.dscp", re.sub(r'\s+','',str(range(56,64))), give_feedback=False)

            cfg_qos_features.edit("default_egress_sched.egr_queue_0.bw_percent", "50", give_feedback=False)
            cfg_qos_features.edit("default_egress_sched.egr_queue_3.bw_percent", "50", give_feedback=False)
            cfg_qos_features.edit("default_egress_sched.egr_queue_6.bw_percent", "0", give_feedback=False)

            # comment out the remaining default egress queues scheduler config
            current_value = cfg_qos_features.get_value(False)
            for queue in [1,2,4,5,7]:
                val = current_value.get("default_egress_sched.egr_queue_{}.bw_percent".format(queue))
                if val:
                    cfg_qos_features.comment_out("default_egress_sched.egr_queue_{}.bw_percent".format(queue), give_feedback=False)

            # comment out egress_sched config for each port-group
            current_value = cfg_qos_features.get_value(False)
            egr_sched_port_grp_list = current_value.get("egress_sched.port_group_list")
            if egr_sched_port_grp_list:
                for ch in ['[',']']:
                    egr_sched_port_grp_list = egr_sched_port_grp_list.replace(ch,'')
                sched_list = convert_list_string_to_list(egr_sched_port_grp_list, False)
                for egr_sched_port_grp in sched_list:
                    self.configure_all_egress_queues(egr_sched_port_grp,"del")

            if roce_mode == MAP_ROCE_MODE['lossless'] :
                # =============================
                # 3. Enable PFC for priority 3.
                # =============================

                # Add ROCE_PFC to pfc.port_group_list
                cfg_qos_features.edit("pfc.port_group_list", "[ROCE_PFC]", give_feedback=False)
                cfg_qos_features.edit("pfc.ROCE_PFC.port_set","allports", give_feedback=False)
                cfg_qos_features.edit("pfc.ROCE_PFC.cos_list", "[3]", give_feedback=False)
                cfg_qos_features.edit("pfc.ROCE_PFC.cable_length", "100", give_feedback=False)

            elif roce_mode == MAP_ROCE_MODE['lossy'] :
                #This will only have effect when there is is a previous PFC configuration
                #in case of net add roce lossless -> net add roce lossy
                #no need to restore this in delete case

                # Remove pfc.port_group_list
                cfg_qos_features.edit("pfc.port_group_list", "[pfc_port_group]", give_feedback=False, comment_out=True)
                cfg_qos_features.comment_out("pfc.port_group_list", give_feedback=False)

                # Delete all ROCE_PFC configuration.
                for key in cfg_content:
                    if "ROCE_PFC" in key:
                        cfg_qos_features.delete(key)

        elif add_del == "del":
            command_ok = True
            # Remove ROCE_PFC from pfc.port_group_list
            # Put the port_group_list back to the default that was in traffic.conf
            cfg_qos_features.edit("pfc.port_group_list", "[pfc_port_group]", give_feedback=False)
            cfg_qos_features.comment_out("pfc.port_group_list", give_feedback=False)

            # Delete all ROCE_PFC configuration.
            for key in cfg_content:
                if "ROCE_PFC" in key:
                    cfg_qos_features.delete(key)

            # Remove TC 3 from ECN configuration
            cfg_qos_features.edit("default_ecn_conf.egress_queue_list", "[0]", give_feedback=False) 

            # ======================================================
            # Put traffic.conf back to the original defaults.
            # ======================================================
            cfg_qos_features.edit("traffic.packet_priority_source_set", "[802.1p]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_0.priority_source.8021p", "[0]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_1.priority_source.8021p", "[1]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_2.priority_source.8021p", "[2]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_3.priority_source.8021p", "[3]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_4.priority_source.8021p", "[4]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_5.priority_source.8021p", "[5]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_6.priority_source.8021p", "[6]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_7.priority_source.8021p", "[7]", give_feedback=False)

            # When values of the key are same, "comment_out=True" does not work with edit hence commenting in next set of lines
            cfg_qos_features.edit("traffic.cos_0.priority_source.dscp", "[0,1,2,3,4,5,6,7]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_1.priority_source.dscp", "[8,9,10,11,12,13,14,15]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_2.priority_source.dscp", "[16,17,18,19,20,21,22,23]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_3.priority_source.dscp", "[24,25,26,27,28,29,30,31]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_4.priority_source.dscp", "[32,33,34,35,36,37,38,39]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_5.priority_source.dscp", "[40,41,42,43,44,45,46,47]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_6.priority_source.dscp", "[48,49,50,51,52,53,54,55]", give_feedback=False)
            cfg_qos_features.edit("traffic.cos_7.priority_source.dscp", "[56,57,58,59,60,61,62,63]", give_feedback=False)

            cfg_qos_features.comment_out("traffic.cos_0.priority_source.dscp", give_feedback=False)
            cfg_qos_features.comment_out("traffic.cos_1.priority_source.dscp", give_feedback=False)
            cfg_qos_features.comment_out("traffic.cos_2.priority_source.dscp", give_feedback=False)
            cfg_qos_features.comment_out("traffic.cos_3.priority_source.dscp", give_feedback=False)
            cfg_qos_features.comment_out("traffic.cos_4.priority_source.dscp", give_feedback=False)
            cfg_qos_features.comment_out("traffic.cos_5.priority_source.dscp", give_feedback=False)
            cfg_qos_features.comment_out("traffic.cos_6.priority_source.dscp", give_feedback=False)
            cfg_qos_features.comment_out("traffic.cos_7.priority_source.dscp", give_feedback=False)

            cfg_qos_features.delete("ecmp_hash_seed")

            # Restore these changes to default settings.
            for egr_queue in EGR_QUEUE_LIST:
                egr_sched_bw_key = "default_egress_sched.egr_queue_{}.bw_percent".format(egr_queue)
                cfg_qos_features.edit(egr_sched_bw_key, DEFAULT_EGR_SCHED_BW[egr_queue], give_feedback=False)

        return command_ok

    def get_roce_mode_str(self, roce_mode):
        """
        Return the roce mode string / name for the roce mode.
        """
        return dict(zip(MAP_ROCE_MODE.values(), MAP_ROCE_MODE.keys()))[roce_mode]


    def get_roce_config_mode(self):
        # Get the current configured roce mode (from the roce config file)
        roce_mode = self.get_conf_file_key_value(QOS_CONFIG, ROCE_MODE_KEY)
        if roce_mode is not None:
            roce_mode = int(roce_mode)
        else:
            roce_mode = MAP_ROCE_MODE['disabled']
        return roce_mode


    def get_egress_sched_config_mode(self):
        # Get the current configured egress-sched mode (from the egress-sched config file)
        egress_sched = self.get_conf_file_key_value(QOS_CONFIG, EGRESS_SCHED_KEY)
        if egress_sched is not None:
            egress_sched = int(egress_sched)
        else:
            egress_sched = MAP_EGRESS_SCHED['disabled']
        return egress_sched


    def get_egress_sched_pending_mode(self, cfg):
        egress_sched = cfg.get_value(True).get(EGRESS_SCHED_KEY)
        if egress_sched is not None:
            egress_sched = int(egress_sched)
        else:
            egress_sched = MAP_EGRESS_SCHED['disabled']
        return egress_sched


    def get_roce_pending_mode(self, cfg):
        roce_mode = cfg.get_value(True).get(ROCE_MODE_KEY)
        if roce_mode is not None:
            roce_mode = int(roce_mode)
        else:
            roce_mode = MAP_ROCE_MODE['disabled']
        return roce_mode


    def is_roce_mode(self, roce_mode):
        return roce_mode in (
            MAP_ROCE_MODE['lossy'],
            MAP_ROCE_MODE['lossless'],
        )

    def is_egress_sched_mode(self, egress_sched):
        return egress_sched in (
            MAP_EGRESS_SCHED['egress_sched'],
        )

    def is_roce_allowed(self, cfg_qos):
        roce_config_mode = self.get_roce_config_mode()
        roce_pending_mode = self.get_roce_pending_mode(cfg_qos)

        if roce_config_mode != roce_pending_mode:
            roce_mode = roce_pending_mode
        else:
            roce_mode = roce_config_mode

        if roce_mode in (
            MAP_ROCE_MODE['disabled'],
            MAP_ROCE_MODE['lossy'],
            MAP_ROCE_MODE['lossless'],):

            return True
        else:
            return False

    def is_egress_sched_allowed(self, cfg_qos):
        roce_mode = self.get_roce_pending_mode(cfg_qos)
        if roce_mode in (
            MAP_ROCE_MODE['disabled'],):
            return True
        else:
            return False


    def is_disabled(self, cfg):
        roce_config_mode = self.get_roce_config_mode()
        roce_pending_mode = self.get_roce_pending_mode(cfg)

        if roce_config_mode != roce_pending_mode:
            roce_mode = roce_pending_mode
        else:
            roce_mode = roce_config_mode

        if roce_mode == MAP_ROCE_MODE['disabled']:
            return True
        else:
            return False

    def ifnames_all_swps(self):
        return [ifname for ifname in get_interfaces() if match_swp(ifname)]


    def del_roce_mode(self, cfg_qos, cfg_content=None):
        # Delete the roce configuration and restore default value

        # Check for roce configured or in pending config. Without this check
        # the passthrough to the storage-optimized calls puts up a really weird
        # message. Stop it here and put up something reasonable.
        roce_mode = self.get_roce_pending_mode(cfg_qos)
        if roce_mode == MAP_ROCE_MODE['disabled']:
            roce_mode = self.get_roce_config_mode()
            if roce_mode == MAP_ROCE_MODE['disabled']:
                print("ROCE mode is not configured")
                return False

        self.qos_features_conf_eval_roce("del", roce_mode)
        self.qos_infra_conf_eval_roce("del", roce_mode)

        # Update the roce_mode in the qos.conf file
        cfg_qos.edit(ROCE_MODE_KEY, MAP_ROCE_MODE['disabled'], give_feedback=False)

        # Update the egress_sched mode in the roce.conf file
        cfg_qos.edit(EGRESS_SCHED_KEY, MAP_EGRESS_SCHED["disabled"], give_feedback=False)

        return True


    def get_roce_config_show_json(self):
        roce_config = {}

        roce_config['RoCE mode'] = 'Not configured'
        roce_config_mode = self.get_roce_config_mode()

        if self.is_roce_mode(roce_config_mode):
            roce_mode_str = self.get_roce_mode_str(roce_config_mode)
            roce_config['RoCE mode'] = roce_mode_str

            congestion_control = {}

            cos_list = self.get_conf_file_key_value(QOS_FEATURES, "default_ecn_conf.egress_queue_list")
            congestion_control['enabled TC'] = json.loads(cos_list)

            ecn_enable = self.get_conf_file_key_value(QOS_FEATURES, "default_ecn_conf.ecn_enable")
            red_enable = self.get_conf_file_key_value(QOS_FEATURES, "default_ecn_conf.red_enable")

            if ecn_enable == 'true':
                congestion_control['mode'] = 'ECN+RED' if red_enable == 'true' else 'ECN'
            else:
                congestion_control['mode'] = 'RED' if red_enable == 'true' else 'Not Enabled'

            congestion_control['min threshold'] = self.get_conf_file_key_value(QOS_FEATURES, "default_ecn_conf.min_threshold_bytes")
            congestion_control['max threshold'] = self.get_conf_file_key_value(QOS_FEATURES, "default_ecn_conf.max_threshold_bytes")

            roce_config['congestion control'] = congestion_control

            pfc = {}
            pfc_enabled = self.get_conf_file_key_value(QOS_FEATURES, "pfc.ROCE_PFC.port_set")

            if pfc_enabled is None:
                pfc['status'] = 'disabled'
            else:
                pfc['status'] = 'enabled'
                cos_list = self.get_conf_file_key_value(QOS_FEATURES, "pfc.ROCE_PFC.cos_list")
                pfc['enabled SPs'] = json.loads(cos_list)

            roce_config['PFC'] = pfc

            ecn_port_set = []

            ifnames = self.ifnames_all_swps()
            for ifname in ifnames:
                if ifname not in ecn_port_set:
                    ecn_port_set.append(ifname)

            roce_config['interfaces'] = ecn_port_set

            dscp_map = {}
            for priority in range(8):
                tc = self.get_conf_file_key_value(QOS_INFRA, "cos_egr_queue.cos_{0}.uc".format(priority))
                dscp = self.get_conf_file_key_value(QOS_FEATURES, "traffic.cos_{0}.priority_source.dscp".format(priority))

                dscp_map_priority = {}
                dscp_map_priority['tc'] = tc
                # the 'dscp' values is in the config file in what looks like a list but it's a string quoted list
                # like this... '[1,2,3,4,5,6,7]' use json.loads to get this into a proper list.
                dscp_map_priority['dscp'] = json.loads(dscp)

                dot1p = self.get_conf_file_key_value(QOS_FEATURES, "traffic.cos_{0}.priority_source.8021p".format(priority))
                dscp_map_priority['802.1p'] = json.loads(dot1p)

                dscp_map[priority] = dscp_map_priority
                ets = self.get_conf_file_key_value(QOS_FEATURES, "default_egress_sched.egr_queue_{0}.bw_percent".format(tc))
                if ets is None:
                    dscp_map_priority['ETS'] = 'DWRR 0%'
                else:
                    if int(ets) == 0:
                        dscp_map_priority['ETS'] = 'strict priority'
                    else:
                        dscp_map_priority['ETS'] = 'DWRR {0}%'.format(ets)

                dscp_map[priority] = dscp_map_priority

            roce_config['dscp map'] = dscp_map

        return roce_config


    def format_roce_config_show(self, roce_config_json):
        out = ""
        dscp_map_table = ""
        tc_map_table = ""
        try:
            out +=  "{0:.<19} {1}\n".format('RoCE mode', roce_config_json['RoCE mode'])
            out += "Congestion Control:\n    {0:.<15} {1}\n".format('Enabled TC', ' '.join([str(i) for i in roce_config_json['congestion control']['enabled TC']]))
            out += "    {0:.<15} {1}\n".format('Mode', roce_config_json['congestion control']['mode'])
            out += "    {0:.<15} {1:.0f} KB\n".format('Min Threshold', float(roce_config_json['congestion control']['min threshold'])/float(1000))
            out += "    {0:.<15} {1:.0f} KB\n".format('Max Threshold', float(roce_config_json['congestion control']['max threshold'])/float(1000))
            pfc_status = roce_config_json['PFC']['status']
            out += "PFC:\n    {0:.<15} {1}\n".format('Status', pfc_status)
            if pfc_status == 'enabled':
                out += "    {0:.<15} {1}\n".format('Enabled SPs', ' '.join([str(i) for i in roce_config_json['PFC']['enabled SPs']]))
            out += "{0:.<19} {1}\n".format('Interfaces', ifnames_to_glob(roce_config_json['interfaces']))
            out += '\n'

            dscp_map_table = []
            tc_map_dict = {}
            for priority, dscp_map in roce_config_json['dscp map'].items():
                dscp_map_entry = []
                dscp_map_entry.append(' '.join([str(i) for i in dscp_map['dscp']]))
                dscp_map_entry.append(' '.join([str(i) for i in dscp_map['802.1p']]))
                dscp_map_entry.append(priority)

                # Populate the tc_map_dict
                tc = dscp_map['tc']
                if tc in tc_map_dict:
                    tc_map_dict[tc]['SP'].append(priority)
                else:
                    tc_map_dict[tc] = {}
                    tc_map_dict[tc]['SP'] = [priority]
                    tc_map_dict[tc]['ETS'] = dscp_map['ETS']

                # append each dscp mapping entry
                dscp_map_table.append(dscp_map_entry)

            dscp_map_table = tabulate(
                dscp_map_table,
                headers=['DSCP', '802.1p', 'switch-priority'],
                numalign="right",
            )

            tc_map_table = []
            for tc, tc_map in tc_map_dict.items():
                tc_map_entry = []
                tc_map_entry.append(' '.join([str(i) for i in tc_map['SP']]))
                tc_map_entry.append(tc)
                tc_map_entry.append(tc_map['ETS'])

                # append each dscp mapping entry
                tc_map_table.append(tc_map_entry)

            tc_map_table = tabulate(
                tc_map_table,
                headers=['switch-priority', 'TC', 'ETS'],
                numalign="right",
            )

        except KeyError:
            # KeyError can occur if some of the content is not present, which is expected
            pass

        return out + dscp_map_table + '\n\n' + tc_map_table

    # prof_list can be ["default"] or ["default", "non-default"]
    def is_egress_sched_config_applied(self, prof_list):
        cfg_traffic = self.files[QOS_FEATURES]
        is_egr_sched_applied = False

        current_value = cfg_traffic.get_value(True)

        for prof in prof_list:
            if prof == "default":
                # Get default egress queues scheduler config
                for queue in EGR_QUEUE_LIST:
                    val = current_value.get("default_egress_sched.egr_queue_{}.bw_percent".format(queue))
                    if val:
                        is_egr_sched_applied = True
                        break
            else:
                # Get egress_sched config for each port-group
                egr_sched_port_grp_list = current_value.get("egress_sched.port_group_list")
                if egr_sched_port_grp_list:
                    for ch in ['[',']']:
                        egr_sched_port_grp_list = egr_sched_port_grp_list.replace(ch,'')
                    if egr_sched_port_grp_list:
                        is_egr_sched_applied = True
                        break

        return is_egr_sched_applied

    def get_total_bw_egress_queue(self, prof_name):
        cfg_traffic = self.files[QOS_FEATURES]
        sum_bw = 0
        for egr_queue in EGR_QUEUE_LIST:
            if 'default' in prof_name:
                curr_value = cfg_traffic.get_value(True).get("{}.egr_queue_{}.bw_percent".format(prof_name,egr_queue))
            else:
                curr_value = cfg_traffic.get_value(True).get("egress_sched.{}.egr_queue_{}.bw_percent".format(prof_name,egr_queue))

            if curr_value is not None:
                sum_bw += int(curr_value)

        return sum_bw

    # add_del:"del": comment the default egress_sched profile and delete the non-default port-group egress_sched profile
    # add_del:"add": add the default egress_sched profile with default values as well as for non-default port-group egress_sched profile
    def configure_all_egress_queues(self, prof_name, add_del):
        cfg_traffic = self.files[QOS_FEATURES]
        if prof_name == 'default_egress_sched':
            for egr_queue in EGR_QUEUE_LIST:
                egr_sched_bw_key = "default_egress_sched.egr_queue_{}.bw_percent".format(egr_queue)
                # Configure default value
                cfg_traffic.edit(egr_sched_bw_key, DEFAULT_EGR_SCHED_BW[egr_queue], give_feedback=False)
        else:
            current_value = cfg_traffic.get_value(True)

            port_group_key = "egress_sched.port_group_list"
            port_group_str = cfg_traffic.get_value(True).get(port_group_key)
            port_group_list = []
            if port_group_str:
                for ch in ['[',']']:
                    port_group_str = port_group_str.replace(ch,'')
                port_group_list = convert_list_string_to_list(port_group_str,False)

            if add_del == "add":
                # Configure default value for port-group egress sched profile
                for egr_queue in EGR_QUEUE_LIST:
                    egr_sched_bw_key = "egress_sched.{}.egr_queue_{}.bw_percent".format(prof_name,egr_queue)
                    cfg_traffic.edit(egr_sched_bw_key, DEFAULT_EGR_SCHED_BW[egr_queue], give_feedback=False)

                # Add the port-group profile in egress_sched port-group
                if prof_name not in port_group_list:
                    port_group_list.append(prof_name)
                    cfg_traffic.edit(port_group_key, "[{}]".format(",".join(port_group_list)), give_feedback=False)
            else:
                # Delete the egress queue for the non-default profile
                for egr_queue in EGR_QUEUE_LIST:
                    egr_sched_bw_key = "egress_sched.{}.egr_queue_{}.bw_percent".format(prof_name,egr_queue)
                    if prof_name == DEFAULT_EGRESS_SCHED_PROFILE_NAME:
                        # Do not delete the default egress sched profile in traffic.conf
                        cfg_traffic.edit(egr_sched_bw_key, DEFAULT_EGR_SCHED_BW[egr_queue], give_feedback=False, comment_out=True)
                    else:
                        cfg_traffic.delete(egr_sched_bw_key)

                # Delete the port-set from the port-group profile
                egr_sched_port_grp_port_set_key = "egress_sched.{}.port_set".format(prof_name)
                if prof_name == DEFAULT_EGRESS_SCHED_PROFILE_NAME:
                    # Do not delete the default egress sched profile in traffic.conf
                    cfg_traffic.edit(egr_sched_port_grp_port_set_key, "swp2-swp3,swp4", give_feedback=False, comment_out=True)
                else:
                    cfg_traffic.delete(egr_sched_port_grp_port_set_key)

                # Remove the port-group profile in egress_sched port-group
                if prof_name in port_group_list:
                    port_group_list[:] = [x for x in port_group_list if prof_name not in x]
                    if port_group_list:
                        cfg_traffic.edit(port_group_key, "[{}]".format(",".join(port_group_list)), give_feedback=False)
                    else:
                        cfg_traffic.edit(port_group_key, "[{}]".format(DEFAULT_EGRESS_SCHED_PROFILE_NAME), give_feedback=False)
                        cfg_traffic.comment_out(port_group_key)

    def eval_default_egress_sched_profile(self, cli):
        tokens = cli.args
        # For default profile, add default values to default egress scheduler profile
        if "add" in tokens:
            self.configure_all_egress_queues("default_egress_sched","add")
            return True
        return False

    def conf_egress_sched_port_set(self, prof_name, ifnames, add_del):
        rc = False
        cfg_traffic = self.files[QOS_FEATURES]

        if prof_name and ifnames:
            port_group_port_set_key = "egress_sched.{}.port_set".format(prof_name)
            port_set_str = cfg_traffic.get_value(True).get(port_group_port_set_key)
            port_list = []
            if port_set_str:
                port_list = convert_list_string_to_list(port_set_str,True)
            if add_del == "add":
                for ifname in ifnames:
                    port_list.append(ifname)
                # Remove duplicate ports from the list
                port_list = list(set(port_list))
                cfg_traffic.edit(port_group_port_set_key, ",".join(port_list), give_feedback=False)
                rc = True
            if add_del == "del":
                port_list = list(set(port_list) - set(ifnames))
                if port_list:
                    cfg_traffic.edit(port_group_port_set_key, ",".join(port_list), give_feedback=False)
                else:
                    # Delete all the port_group egress scheduler config as well as port_set
                    self.configure_all_egress_queues(prof_name,"del")
                rc = True

        return rc

    def eval_port_group_egress_sched_profile(self, cli):
        tokens = cli.args
        cfg_traffic = self.files[QOS_FEATURES]

        egr_queue = tokens.get('<0-7>')
        strict = tokens.get('strict')
        dwrr = tokens.get('dwrr')
        bw_percent = tokens.get('bw-percent')
        bw = None
        port_group_list = []

        prof_name = tokens.get('<prof-name>')
        if not prof_name:
            prof_name = tokens.get('<text>')

        if not prof_name:
            print "NCLU does not know how to process command \"{0}\".".format(cli.argv_expanded)
            return False

        port_set = tokens.get('interface')

        interface = tokens.get('<interface>')
        ifnames = []
        if interface:
            ifnames = ifname_expand_glob(str(interface))
            if not ifnames:
                log.debug("NCLU does not know how to process command \"{0}\".".format(cli.argv_expanded))
                return False

            invalid_ifaces = set(ifnames) - set(self.ifnames_all_swps())
            if invalid_ifaces != set():
                print "Following interfaces are not valid:{}".format(list(invalid_ifaces))
                return False

        port_group_key = "egress_sched.port_group_list"
        port_group_str = cfg_traffic.get_value(True).get(port_group_key)
        if port_group_str:
            for ch in ['[',']']:
                port_group_str = port_group_str.replace(ch,'')
            port_group_list = convert_list_string_to_list(port_group_str,False)

        rc = False
        if "add" in tokens:
            if dwrr and bw_percent:
                bw = tokens.get('<1-100>')
            elif strict:
                # strict prio queue has bw_percent configured as 0
                bw = 0
            elif port_set:
                rc = self.conf_egress_sched_port_set(prof_name, ifnames, "add")
            else:
                # should not come here - no NCLU CLI support
                bw = None
                rc = False

            if bw is not None and egr_queue is not None:
                egr_sched_key = "egress_sched.{}.egr_queue_{}.bw_percent".format(prof_name,egr_queue)
                # Check whether port-set is configured for the port_group
                port_group_port_set_key = "egress_sched.{}.port_set".format(prof_name)
                port_set_str = cfg_traffic.get_value(True).get(port_group_port_set_key)
                if not port_set_str:
                    print "Need to first configure port-set interface for egress scheduler profile '{}'".format(prof_name)
                    return False

                # Validate the total bandwidth for all queues in port_group to be less than 100
                sum_bw = self.get_total_bw_egress_queue(prof_name)
                # Adjust the value of the current queue to be configured
                conf_bw = cfg_traffic.get_value(True).get(egr_sched_key)
                if not conf_bw:
                    sum_bw += bw
                else:
                    sum_bw += (bw - int(conf_bw))
                if sum_bw > 100:
                    print "Sum total of all egress queues for egress scheduler profile {} exceeds 100".format(prof_name)
                    return False

                # Configure bw_percent per port_group egr_scheduler profile per queue
                cfg_traffic.edit(egr_sched_key, bw, give_feedback=False)

                rc = True
            if prof_name not in port_group_list:
                port_group_list.append(prof_name)
                cfg_traffic.edit(port_group_key, "[{}]".format(",".join(port_group_list)), give_feedback=False)

        elif "del" in tokens:
            bw = None
            if port_set:
                if ifnames:
                    # Delete a particular set of ifnames from the port-set
                    rc = self.conf_egress_sched_port_set(prof_name, ifnames, "del")
                else:
                    # Delete all the ports of the port-group port_set
                    port_group_port_set_key = "egress_sched.{}.port_set".format(prof_name)
                    port_set_str = cfg_traffic.get_value(True).get(port_group_port_set_key)
                    port_list = []
                    if port_set_str:
                        port_list = convert_list_string_to_list(port_set_str,True)
                    rc = self.conf_egress_sched_port_set(prof_name, port_list, "del")
            elif egr_queue is not None:
                # Make sure that the egress_sched config for the queue is deleted
                egr_sched_key = "egress_sched.{}.egr_queue_{}.bw_percent".format(prof_name,egr_queue)
                if prof_name == DEFAULT_EGRESS_SCHED_PROFILE_NAME:
                    cfg_traffic.edit(egr_sched_key, DEFAULT_EGR_SCHED_BW[egr_queue], give_feedback=False, comment_out=True)
                else:
                    cfg_traffic.delete(egr_sched_key)
                rc = True
            else:
                # Delete all the port_group egress scheduler config as well as port_set
                self.configure_all_egress_queues(prof_name,"del")
                rc = True

        return rc

    def eval_egress_sched_cmd(self,cli):
        tokens = cli.args
        cfg_traffic = self.files[QOS_FEATURES]
        cfg_qos = self.files[QOS_CONFIG]

        rc = False

        if "profile" in tokens:
            if "default" in tokens:
                rc = self.eval_default_egress_sched_profile(cli)
            else:
                rc = self.eval_port_group_egress_sched_profile(cli)

        if rc:
            if self.is_egress_sched_config_applied(["default", "non-default"]):
                cfg_qos.edit(EGRESS_SCHED_KEY, MAP_EGRESS_SCHED["egress_sched"], give_feedback=False)

        return rc

    def eval_qos_cmd(self, cli):
        tokens = cli.args
        log.debug("Processing qos commands: \"{0}\".".format(cli.argv_expanded))
        rc = False
        if 'egress' in tokens and 'scheduler' in tokens:
            # dont allow qos commands if ROCE is in place.
            if not self.is_egress_sched_allowed(self.files[QOS_CONFIG]):
                print("Egress scheduler config not allowed when ROCE commands are configured")
                return False
            rc = self.eval_egress_sched_cmd(cli)
        return rc

    def eval_roce_mode(self, cli):
        tokens = cli.args
        cfg_qos = self.files[QOS_CONFIG]

        rc = False
        if "add" in tokens:

            if "lossy" in tokens:
                roce_mode = MAP_ROCE_MODE["lossy"]
            else:
                # lossless is the default if none of the other modes are present
                roce_mode = MAP_ROCE_MODE["lossless"]

            self.qos_features_conf_eval_roce("add", roce_mode)
            self.qos_infra_conf_eval_roce("add", roce_mode)

            # Update the roce_mode in the qos.conf file
            cfg_qos.edit(ROCE_MODE_KEY, roce_mode, give_feedback=False)

            # Update the egress_sched mode in the roce.conf file
            cfg_qos.edit(EGRESS_SCHED_KEY, MAP_EGRESS_SCHED["disabled"], give_feedback=False)

            rc = True

        elif "del" in tokens:

            rc =  self.del_roce_mode(cfg_qos)

        elif "show" in tokens:
            try:
                json_output = self.get_roce_config_show_json()

                if "json" in tokens:
                    show_output = json.dumps(
                        json_output,
                        indent=4,
                        sort_keys=True,
                        separators=(",", ": "),
                        )
                else:
                    show_output = self.format_roce_config_show(json_output)

                # wrappers need show output printed and a boolean RC to show success or failure
                print(show_output)
                rc = True

            except Exception as e:
                log.exception(e)
                raise Exception("Unable to get output for show RoCE command")

        return rc


    def qos_infra_conf_eval_roce(self, add_del, roce_mode):
        """
        Handle all add/del for roce in the datapath.conf file.
        """
        cfg_qos_infra = self.files[QOS_INFRA]
        cfg_content = cfg_qos_infra.get_value(True)

        # Delete all other PG configuration.
        for pg in ['control', 'service1', 'service2', 'service3', 'service4', 'service5', 'service6']:
            key = 'priority_group.{0}.cos_list'.format(pg)
            if cfg_content.get(key):
                cfg_qos_infra.delete(key)

            key = 'priority_group.{0}.id'.format(pg)
            if cfg_content.get(key):
                cfg_qos_infra.delete(key)

            key = 'priority_group.{0}.service_pool'.format(pg)
            if cfg_content.get(key):
                cfg_qos_infra.delete(key)

            key = 'priority_group.{0}.ingress_buffer.dynamic_quota'.format(pg)
            if cfg_content.get(key):
                cfg_qos_infra.delete(key)

        if add_del == "add":

            # ============================================
            # 4. Map switch priorities to priority groups.
            # ============================================

            if roce_mode == MAP_ROCE_MODE['lossless']:
                cfg_qos_infra.edit("traffic.priority_group_list", "[bulk]", give_feedback=False)
                cfg_qos_infra.edit("priority_group.bulk.cos_list", "[0,1,2,3,4,5,6,7]", give_feedback=False)
            else:
                cfg_qos_infra.edit("traffic.priority_group_list", "[bulk,service1]", give_feedback=False)
                cfg_qos_infra.edit("priority_group.service1.cos_list", "[3]", give_feedback=False)
                cfg_qos_infra.edit("priority_group.bulk.cos_list", "[0,1,2,4,5,6,7]", give_feedback=False)

            # =============================================
            # 5. Configure scheduling algorithm
            # =============================================
            cfg_qos_infra.edit("scheduling.algorithm", "dwrr", give_feedback=False)

            # =============================================
            # 6. Assign group IDs, and create buffer pools.
            # =============================================
            cfg_qos_infra.edit("priority_group.bulk.id", '0', give_feedback=False)
            cfg_qos_infra.edit("priority_group.bulk.service_pool", '0', give_feedback=False)

            cfg_qos_infra.edit("ingress_service_pool.0.percent", "50.0", give_feedback=False)
            cfg_qos_infra.edit("ingress_service_pool.0.mode", '1', give_feedback=False)
            cfg_qos_infra.edit("egress_service_pool.0.percent", "50.0", give_feedback=False)
            cfg_qos_infra.edit("egress_service_pool.0.mode", '1', give_feedback=False)

            cfg_qos_infra.edit("ingress_service_pool.1.percent", "50.0", give_feedback=False)
            cfg_qos_infra.edit("ingress_service_pool.1.mode", '1', give_feedback=False)
            cfg_qos_infra.edit("egress_service_pool.1.mode", '1', give_feedback=False)

            cfg_qos_infra.edit("egress_buffer.cos_3.mc.service_pool", '1', give_feedback=False)
            cfg_qos_infra.edit("egress_buffer.egr_queue_3.uc.service_pool", '1', give_feedback=False)

            if roce_mode == MAP_ROCE_MODE['lossless']:
                cfg_qos_infra.edit("flow_control.ingress_service_pool", '1', give_feedback=False)
                cfg_qos_infra.edit("flow_control.egress_service_pool", '1', give_feedback=False)
                cfg_qos_infra.edit("egress_service_pool.1.infinite_flag", 'TRUE', give_feedback=False)
                cfg_qos_infra.edit("egress_service_pool.1.percent", "100.0", give_feedback=False)
            else:
                #To handle the case where user directly moves from roce lossless to roce lossy
                cfg_qos_infra.edit("flow_control.ingress_service_pool", '0', give_feedback=False)
                cfg_qos_infra.edit("flow_control.egress_service_pool", '0', give_feedback=False)
                cfg_qos_infra.comment_out('flow_control.ingress_service_pool')
                cfg_qos_infra.comment_out('flow_control.egress_service_pool')

                cfg_qos_infra.edit("priority_group.service1.id", '1', give_feedback=False)
                cfg_qos_infra.edit("priority_group.service1.service_pool", '1', give_feedback=False)

                cfg_qos_infra.edit("egress_service_pool.1.percent", "50.0", give_feedback=False)
                cfg_qos_infra.delete("egress_service_pool.1.infinite_flag")

            # ==========================
            # 7. Configure alpha values.
            # ==========================

            # Ingress buffer
            cfg_qos_infra.edit("priority_group.bulk.ingress_buffer.dynamic_quota", "ALPHA_8", give_feedback=False)
            if roce_mode == MAP_ROCE_MODE['lossless']:
                cfg_qos_infra.edit("flow_control.ingress_buffer.dynamic_quota", 'ALPHA_1', give_feedback=False)  # RoCE
                cfg_qos_infra.edit("flow_control.egress_buffer.dynamic_quota", 'ALPHA_INFINITY', give_feedback=False)  # RoCE egress INFINITY
                cfg_qos_infra.edit("port.service_pool.1.ingress_buffer.dynamic_quota", 'ALPHA_1', give_feedback=False)  # RoCE
                cfg_qos_infra.edit("port.service_pool.1.egress_buffer.uc.dynamic_quota", 'ALPHA_INFINITY', give_feedback=False)  # RoCE
            else:
                cfg_qos_infra.edit("flow_control.ingress_buffer.dynamic_quota",'ALPHA_1', give_feedback=False)
                cfg_qos_infra.comment_out("flow_control.ingress_buffer.dynamic_quota", give_feedback=False)
                cfg_qos_infra.edit("flow_control.egress_buffer.dynamic_quota",'ALPHA_1', give_feedback=False)
                cfg_qos_infra.comment_out("flow_control.egress_buffer.dynamic_quota", give_feedback=False)

                cfg_qos_infra.edit("priority_group.service1.ingress_buffer.dynamic_quota", "ALPHA_8", give_feedback=False)
                cfg_qos_infra.edit("port.service_pool.1.ingress_buffer.dynamic_quota", 'ALPHA_8', give_feedback=False)  # RoCE
                cfg_qos_infra.edit("port.service_pool.1.egress_buffer.uc.dynamic_quota", 'ALPHA_8', give_feedback=False)  # RoCE

            cfg_qos_infra.edit("port.service_pool.0.egress_buffer.uc.dynamic_quota", 'ALPHA_8', give_feedback=False) #RoCE

            for cos in range(8):
                key = "egress_buffer.cos_{}.mc.dynamic_quota".format(cos)
                cfg_qos_infra.edit(key, 'ALPHA_1_4', give_feedback=False)

            # Additions to legacy settings common to lossy and lossless modes.
            # Note for storage-optimized the default values are being used so
            # these do not appear in traffic.conf  Adding them in for ROCE modes.
            cfg_qos_infra.edit("cos_egr_queue.cos_0.uc", "0", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_0.cpu", "0", give_feedback=False)

            cfg_qos_infra.edit("cos_egr_queue.cos_1.uc", "0", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_1.cpu", "0", give_feedback=False)

            cfg_qos_infra.edit("cos_egr_queue.cos_2.uc", "0", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_2.cpu", "0", give_feedback=False)

            cfg_qos_infra.edit("cos_egr_queue.cos_3.uc", "3", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_3.cpu", "3", give_feedback=False)

            cfg_qos_infra.edit("cos_egr_queue.cos_4.uc", "0", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_4.cpu", "0", give_feedback=False)

            cfg_qos_infra.edit("cos_egr_queue.cos_5.uc", "0", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_5.cpu", "0", give_feedback=False)

            cfg_qos_infra.edit("cos_egr_queue.cos_6.uc", "6", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_6.cpu", "6", give_feedback=False)

            cfg_qos_infra.edit("cos_egr_queue.cos_7.uc", "0", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_7.cpu", "0", give_feedback=False)

            cfg_qos_infra.edit("egress_buffer.egr_queue_0.uc.dynamic_quota", "ALPHA_8", give_feedback=False)
            cfg_qos_infra.edit("egress_buffer.egr_queue_3.uc.dynamic_quota", "ALPHA_8", give_feedback=False)
            cfg_qos_infra.edit("egress_buffer.egr_queue_6.uc.dynamic_quota", "ALPHA_8", give_feedback=False)

            for queue in [1,2,4,5,7]:
                key = "egress_buffer.egr_queue_{0}.uc.reserved".format(queue)
                cfg_qos_infra.edit(key, "0", give_feedback=False)

            if roce_mode == MAP_ROCE_MODE['lossless']:
                cfg_qos_infra.edit("port.egress_buffer.mc.reserved", '0', give_feedback=False)
                cfg_qos_infra.edit("port.egress_buffer.mc.shared_size", '4294967295', give_feedback=False)

                cfg_qos_infra.edit("port.service_pool.1.ingress_buffer.reserved", '0', give_feedback=False)
                cfg_qos_infra.edit("port.service_pool.1.egress_buffer.uc.reserved", '0', give_feedback=False)

                cfg_qos_infra.edit("egress_buffer.egr_queue_0.uc.reserved", "1024", give_feedback=False)
                cfg_qos_infra.edit("egress_buffer.egr_queue_3.uc.reserved", "0", give_feedback=False)
                cfg_qos_infra.edit("egress_buffer.egr_queue_6.uc.reserved", "1024", give_feedback=False)
            else:
                cfg_qos_infra.edit("port.egress_buffer.mc.reserved", '10240', give_feedback=False)
                cfg_qos_infra.edit("port.egress_buffer.mc.shared_size", '92160', give_feedback=False)

                cfg_qos_infra.edit("port.service_pool.1.ingress_buffer.reserved", '10240', give_feedback=False)
                cfg_qos_infra.edit("port.service_pool.1.egress_buffer.uc.reserved", '10240', give_feedback=False)

                cfg_qos_infra.edit("egress_buffer.egr_queue_0.uc.reserved", "1024", give_feedback=False)
                cfg_qos_infra.edit("egress_buffer.egr_queue_3.uc.reserved", "1024", give_feedback=False)
                cfg_qos_infra.edit("egress_buffer.egr_queue_6.uc.reserved", "1024", give_feedback=False)

        elif add_del == "del":
            cfg_qos_infra.edit("traffic.priority_group_list", "[bulk]", give_feedback=False)
            cfg_qos_infra.edit("priority_group.bulk.cos_list", "[0,1,2,3,4,5,6,7]", give_feedback=False)

            cfg_qos_infra.edit("priority_group.bulk.id", '0', give_feedback=False)
            cfg_qos_infra.edit("priority_group.bulk.service_pool", '0', give_feedback=False)

            cfg_qos_infra.edit("ingress_service_pool.0.mode", '0', give_feedback=False)
            cfg_qos_infra.comment_out("ingress_service_pool.0.mode", give_feedback=False)

            cfg_qos_infra.edit("egress_service_pool.0.mode", '0', give_feedback=False)
            cfg_qos_infra.comment_out("egress_service_pool.0.mode", give_feedback=False)

            cfg_qos_infra.edit("flow_control.ingress_service_pool", '0', give_feedback=False)
            cfg_qos_infra.edit("flow_control.egress_service_pool", '0', give_feedback=False)
            cfg_qos_infra.comment_out('flow_control.ingress_service_pool')
            cfg_qos_infra.comment_out('flow_control.egress_service_pool')

            cfg_qos_infra.edit("ingress_service_pool.0.percent", "100.0", give_feedback=False)
            cfg_qos_infra.delete("ingress_service_pool.1.percent")
            cfg_qos_infra.delete("ingress_service_pool.1.mode")

            cfg_qos_infra.edit("egress_service_pool.0.percent", "100.0", give_feedback=False)
            cfg_qos_infra.delete("egress_service_pool.1.percent")
            cfg_qos_infra.delete("egress_service_pool.1.mode")
            cfg_qos_infra.delete("egress_service_pool.1.infinite_flag")

            cfg_qos_infra.edit("priority_group.bulk.ingress_buffer.dynamic_quota", 'ALPHA_8', give_feedback=False)
            cfg_qos_infra.comment_out("priority_group.bulk.ingress_buffer.dynamic_quota", give_feedback=False)

            cfg_qos_infra.edit("flow_control.ingress_buffer.dynamic_quota",'ALPHA_1', give_feedback=False)
            cfg_qos_infra.comment_out("flow_control.ingress_buffer.dynamic_quota", give_feedback=False)
            cfg_qos_infra.edit("flow_control.egress_buffer.dynamic_quota",'ALPHA_INFINITY', give_feedback=False)
            cfg_qos_infra.comment_out("flow_control.egress_buffer.dynamic_quota", give_feedback=False)

            cfg_qos_infra.delete("port.service_pool.1.ingress_buffer.dynamic_quota")
            cfg_qos_infra.delete("port.service_pool.1.egress_buffer.uc.dynamic_quota")
            cfg_qos_infra.delete("port.service_pool.1.ingress_buffer.reserved")
            cfg_qos_infra.delete("port.service_pool.1.egress_buffer.uc.reserved")

            cfg_qos_infra.edit("port.egress_buffer.mc.reserved", '10240', give_feedback=False)
            cfg_qos_infra.edit("port.egress_buffer.mc.shared_size", '92160', give_feedback=False)
            cfg_qos_infra.comment_out("port.egress_buffer.mc.reserved", give_feedback=False)
            cfg_qos_infra.comment_out("port.egress_buffer.mc.shared_size", give_feedback=False)

            cfg_qos_infra.edit("port.service_pool.0.egress_buffer.uc.dynamic_quota", "ALPHA_16", give_feedback=False)
            cfg_qos_infra.comment_out("port.service_pool.0.egress_buffer.uc.dynamic_quota", give_feedback=False)

            #Multicast Egress buffer dynamic quota
            for cos in range(8):
                key = "egress_buffer.cos_{}.mc.dynamic_quota".format(cos)
                cfg_qos_infra.edit(key, DEFAULT_MC_SP_DYNAMIC_QUOTA[cos], give_feedback=False)
                cfg_qos_infra.comment_out(key, give_feedback=False)

                key = "egress_buffer.cos_{}.mc.service_pool".format(cos)
                cfg_qos_infra.edit(key, '0', give_feedback=False)

            for queue in EGR_QUEUE_LIST:
                key = "egress_buffer.egr_queue_{0}.uc.reserved".format(queue)
                cfg_qos_infra.edit(key, "1024", give_feedback=False)
                cfg_qos_infra.comment_out(key, give_feedback=False)

                key = "egress_buffer.egr_queue_{0}.uc.dynamic_quota".format(queue)
                cfg_qos_infra.edit(key, DEFAULT_EGR_QUEUE_UC_DYNAMIC_QUOTA[queue], give_feedback=False)
                cfg_qos_infra.comment_out(key, give_feedback=False)

                key = "egress_buffer.egr_queue_{0}.uc.service_pool".format(queue)
                cfg_qos_infra.edit(key, "0", give_feedback=False)

            # These were additions to ROCE config. Return them to the default state.
            cfg_qos_infra.edit("cos_egr_queue.cos_0.uc", "0", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_0.cpu", "0", give_feedback=False)

            cfg_qos_infra.edit("cos_egr_queue.cos_1.uc", "1", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_1.cpu", "1", give_feedback=False)

            cfg_qos_infra.edit("cos_egr_queue.cos_2.uc", "2", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_2.cpu", "2", give_feedback=False)

            cfg_qos_infra.edit("cos_egr_queue.cos_3.uc", "3", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_3.cpu", "3", give_feedback=False)

            cfg_qos_infra.edit("cos_egr_queue.cos_4.uc", "4", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_4.cpu", "4", give_feedback=False)

            cfg_qos_infra.edit("cos_egr_queue.cos_5.uc", "5", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_5.cpu", "5", give_feedback=False)

            cfg_qos_infra.edit("cos_egr_queue.cos_6.uc", "6", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_6.cpu", "6", give_feedback=False)

            cfg_qos_infra.edit("cos_egr_queue.cos_7.uc", "7", give_feedback=False)
            cfg_qos_infra.edit("cos_egr_queue.cos_7.cpu", "7", give_feedback=False)

        return True

    def eval_ecmp_hash_seed(self, cli):
        """
        net (add|del) forwarding ecmp hash-seed <0-4294967295>

        Handle all add/del for the "ecmp hash-seed" macro.
        """

        # Convenience variable
        tokens = cli.args
        cfg_traffic = self.files[ETC_TRAFFIC]

        if "add" in tokens:
            assert "<0-4294967295>" in tokens, "command: {0}".format(cli.argv_expanded)
            return cfg_traffic.edit("ecmp_hash_seed", tokens["<0-4294967295>"])
        elif "del" in tokens:
            return cfg_traffic.delete("ecmp_hash_seed")

        log.debug("NCLU does not know how to process command \"{0}\".".format(cli.argv_expanded))
        return False


    def eval_command_line_args(self, cli):
        super(ConfigWrapper, self).eval_command_line_args(cli)

        # Convenience variable
        tokens = cli.args

        if "roce" in tokens:
            if self.is_roce_allowed(self.files[QOS_CONFIG]) is False:
                # Currently there is no case when roce will not be allowed
                print("ROCE commands not allowed to be configured")
                return False

            return self.eval_roce_mode(cli)

        elif "qos" in tokens:
            return self.eval_qos_cmd(cli)

        elif "forwarding" in tokens and "ecmp" in tokens and "hash-seed" in tokens:
            return self.eval_ecmp_hash_seed(cli)

        log.debug("NCLU does not know how to process command \"{0}\".".format(cli.argv_expanded))
        return False
